import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.util.ArrayList;

public class Client {
	private String serverAddress;
	private int serverPort;
	private ArrayList<Work> workList;
	
	public Client(String serverAddress, int serverPort) {
		this.serverAddress = serverAddress;
		this.serverPort = serverPort;
		this.workList = new ArrayList<Work>();
	}
	
	public void addWorkList(String id, String inference, int[] sample, int ds) {
		Work work = new Work(id, inference, sample, ds);
		workList.add(work);
	}
	
	public void addWork(String id) throws IOException {
		Message msg = new Message(id, Message.ADD_WORK);
		msg.setWorkList(workList);
		
		// add work
		sendMessage(msg);
	}
	
	public void clearQueues(String id) throws IOException {
		Message msg = new Message(id, Message.CLEAR_QUEUES);
		sendMessage(msg);
	}
	
	public WorkResult getWorkResult(String id) throws IOException {
		Message msg = new Message(id, Message.GET_GRADIENT);
		Message resp = null;
		
		// send request
		Socket socket;
		try {
			socket = new Socket(serverAddress, serverPort);
		} catch (IOException e1) {
			System.out.println("Couldn't connect to server or other IO error.");
			return null;
		}
		try {
			ObjectOutputStream objectOutput = new ObjectOutputStream(socket.getOutputStream());
			objectOutput.writeObject(msg);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		// read back answer
		try {
			ObjectInputStream objectInput = new ObjectInputStream(socket.getInputStream());
			resp = (Message) objectInput.readObject();
			
			System.out.println(String.format("[Client] Received answer. Type = %d", resp.getType()));
			System.out.println("Work result: " + resp.getWorkResult());
			
			if (resp.getType() == Message.GRADIENT) {
				return resp.getWorkResult();
			} else {
				System.out.println(String.format("[Client] Invalid type. Type = %d", resp.getType()));
				throw new RuntimeException("Invalid message received.");
			}
		} catch (java.net.SocketTimeoutException e) {
			System.out.println("[Client] Reading response timed out. It seems to be time to shutdown.");
		} catch (IOException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} finally {
			socket.close();
		}
		
		return null;
	}
	
	public void sendGradient(String id, int ds, int[] yt, int[] yht, double[] grad) throws IOException {
		Message msg = new Message(id, Message.WORK_DONE);
		WorkResult workResult = new WorkResult(id, ds, yt, yht, grad);
		msg.setWorkResult(workResult);
		
		// send result
		sendMessage(msg);
	}
	
	public void setWeights(String id, double[] weights) throws IOException {
		Message msg = new Message(id, Message.SET_WEIGHTS);
		msg.setWeights(weights);
		
		// send result
		sendMessage(msg);
	}
	
	private void sendMessage(Message msg) throws IOException {
		Socket socket;
		try {
			socket = new Socket(serverAddress, serverPort);
		} catch (IOException e1) {
			System.out.println("Couldn't connect to server or other IO error.");
			throw e1;
		}
		try {
			ObjectOutputStream objectOutput = new ObjectOutputStream(socket.getOutputStream());
			objectOutput.writeObject(msg);
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			socket.close();
		}
	}
	
	public boolean isWorkDone(String id) throws IOException {
		Message msg = new Message(id, Message.IS_WORK_DONE);
		Message resp = null;
		
		// send request
		Socket socket;
		try {
			socket = new Socket(serverAddress, serverPort);
		} catch (IOException e1) {
			System.out.println("Couldn't connect to server or other IO error.");
			return false;
		}
		try {
			ObjectOutputStream objectOutput = new ObjectOutputStream(socket.getOutputStream());
			objectOutput.writeObject(msg);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		// read back answer
		try {
			ObjectInputStream objectInput = new ObjectInputStream(socket.getInputStream());
			resp = (Message) objectInput.readObject();
			
			System.out.println(String.format("[Client] Received answer. Type = %d", resp.getType()));
			System.out.println("Status: " + resp.isWorkDone());
			
			if (resp.getType() == Message.WORK_STATUS) {
				return resp.isWorkDone();
			} else {
				System.out.println(String.format("[Client] Invalid type. Type = %d", resp.getType()));
				throw new RuntimeException("Invalid message received.");
			}
		} catch (java.net.SocketTimeoutException e) {
			System.out.println("[Client] Reading response timed out. It seems to be time to shutdown.");
		} catch (IOException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} finally {
			socket.close();
		}
		
		return false;
		
	}
	
	/**
	 * Asks the server if there is anything to do
	 * 
	 * @throws IOException
	 * @throws InterruptedException
	 */
	public Work getWork() throws InterruptedException, IOException {
		Message msg = new Message("Client", Message.GET_WORK);
		Message resp = null;
		
		// send request
		Socket socket;
		try {
			socket = new Socket(serverAddress, serverPort);
		} catch (IOException e1) {
			System.out.println("Couldn't connect to server or other IO error.");
			return null;
		}
		try {
			ObjectOutputStream objectOutput = new ObjectOutputStream(socket.getOutputStream());
			objectOutput.writeObject(msg);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		// read back answer
		try {
			ObjectInputStream objectInput = new ObjectInputStream(socket.getInputStream());
			resp = (Message) objectInput.readObject();
			
			System.out.println(String.format("[Client] Received answer. Type = %d", resp.getType()));
			
			if (resp.getType() == Message.WORK) {
				return resp.getWork();
			} else if (resp.getType() == Message.WAIT) {
				return null;
			} else {
				System.out.println(String.format("[Client] Invalid type. Type = %d", resp.getType()));
				throw new RuntimeException("Invalid message received.");
			}
		} catch (java.net.SocketTimeoutException e) {
			System.out.println("[Client] Reading response timed out. It seems to be time to shutdown.");
		} catch (IOException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} finally {
			socket.close();
		}
		
		return null;
	}
	
	public static void main(String[] args) throws IOException, InterruptedException {
		Client client = new Client("bayes.spsc.tugraz.at", 9999);
		String id = "1034f7b35a3f5fb78ec4fa5a2e2668a8";
		
		client.clearQueues(id);
		// System.out.println("Work done: " + client.isWorkDone(id));
		// client.addWork(id);
		System.out.println("Work done: " + client.isWorkDone(id));
		
	}
	
}
