public class ClusteredFacilityLocation implements Function {
	private final int nrClusters;
	private final int[] clusterAssignment;
	private final double[] rewards;
	
	public ClusteredFacilityLocation(int[] clusterAssignment, double[] rewards) {
		this.clusterAssignment = clusterAssignment;
		this.rewards = rewards;
		
		// number of clusters
		int nrClusters = 0;
		for (int i = 0; i < clusterAssignment.length; i++) {
			if (clusterAssignment[i] > nrClusters) {
				nrClusters = clusterAssignment[i];
			}
		}
		this.nrClusters = nrClusters;
		
		// sanity check
		for (int i = 0; i < clusterAssignment.length; i++) {
			if (clusterAssignment[i] <= 0 || clusterAssignment[i] > nrClusters) {
				throw new RuntimeException("Invalid cluster assignments.");
			}
		}
	}
	
	public double eval(int[] S) {
		if (S == null) {
			return 0.0;
		}
		
		double[] maxReward = new double[nrClusters];
		for (int i : S) {
			int cluster = clusterAssignment[i - 1] - 1;
			if (rewards[i - 1] > maxReward[cluster]) {
				maxReward[cluster] = rewards[i - 1];
			}
		}
		
		double val = 0.0;
		for (int i = 0; i < nrClusters; i++) {
			val += maxReward[i];
		}
		val /= nrClusters;
		
		return val;
	}
	
	public double[] eval(int[] B, int[] candidates) {
		if (candidates == null)
			candidates = new int[] {};
		
		int nrCandidates = candidates.length;
		double[] values = new double[nrCandidates];
		
		double[] maxReward = new double[nrClusters];
		for (int i : B) {
			int cluster = clusterAssignment[i - 1] - 1;
			if (rewards[i - 1] > maxReward[cluster]) {
				maxReward[cluster] = rewards[i - 1];
			}
		}
		
		for (int c = 0; c < nrCandidates; c++) {
			int candidate = candidates[c];
			int cluster = clusterAssignment[candidate - 1] - 1;
			double reward = rewards[candidate - 1];
			for (int i = 0; i < nrClusters; i++) {
				if (i != cluster) {
					values[c] += maxReward[i];
				} else {
					values[c] += (maxReward[i] > reward ? maxReward[i] : reward);
				}
			}
			values[c] /= nrClusters;
		}
		
		return values;
	}
	
	// >> For octave compatibility >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
	
	public double eval(int St) {
		int[] S = new int[] { St };
		return eval(S);
	}
	
	public double[] eval(int St, int[] candidates) {
		int[] S = new int[] { St };
		return eval(S, candidates);
	}
	
	public double[] eval(int[] S, int candidatest) {
		int[] candidates = new int[] { candidatest };
		return eval(S, candidates);
	}
}
