public class CosineSimilarity implements Function {
	private final double[][] sim;
	private final double threshold;
	private final double[] normalization;
	private final int nrImages;
	
	public CosineSimilarity(Matrix sim, double threshold) {
		this(sim.getMatrix(), threshold);
	}
	
	public CosineSimilarity(double[][] sim, double threshold) {
		this.sim = sim;
		this.threshold = threshold;
		this.nrImages = sim.length;
		
		// sanity checks
		if (sim.length == 0) {
			throw new RuntimeException("Similarity matrix must contain at least one row.");
		}
		if (sim.length != sim[0].length) {
			throw new RuntimeException("Similarity matrix must be square");
		}
		
		double[] normalization = new double[nrImages];
		for (int i = 0; i < nrImages; i++) {
			for (int j = 0; j < nrImages; j++) {
				normalization[i] += sim[j][i];
			}
			normalization[i] *= threshold;
		}
		this.normalization = normalization;
	}
	
	public double eval(int[] S) {
		if (S == null) {
			return 0.0;
		}
		
		double val = 0.0;
		
		double[] tmp = new double[nrImages];
		for (int i = 0; i < nrImages; i++) {
			for (Integer j : S) {
				tmp[i] += sim[j - 1][i];
			}
		}
		
		// sum and threshold
		for (int i = 0; i < nrImages; i++) {
			double tval = tmp[i] / normalization[i];
			if (tval < 1.0) {
				val += tval;
			} else {
				val += 1.0;
			}
		}
		
		return val / 100.0;
	}
	
	public double[] eval(int[] B, int[] candidates) {
		if (candidates == null)
			candidates = new int[] {};
		
		int nrCandidates = candidates.length;
		double[] values = new double[nrCandidates];
		
		double[] tmp = new double[nrImages];
		for (int i = 0; i < nrImages; i++) {
			for (Integer j : B) {
				tmp[i] += sim[j - 1][i];
			}
		}
		
		// sum and threshold
		for (int c = 0; c < nrCandidates; c++) {
			int candidate = candidates[c];
			
			double val = 0.0;
			for (int i = 0; i < nrImages; i++) {
				double tval = (tmp[i] + sim[candidate - 1][i]) / normalization[i];
				if (tval < 1.0) {
					val += tval;
				} else {
					val += 1.0;
				}
			}
			
			values[c] = val / 100.0;
		}
		
		return values;
	}
	
	// >> For octave compatibility >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
	
	public double eval(int St) {
		int[] S = new int[] { St };
		return eval(S);
	}
	
	public double[] eval(int St, int[] candidates) {
		int[] S = new int[] { St };
		return eval(S, candidates);
	}
	
	public double[] eval(int[] S, int candidatest) {
		int[] candidates = new int[] { candidatest };
		return eval(S, candidates);
	}
	
}
