import static jcuda.driver.JCudaDriver.cuCtxCreate;
import static jcuda.driver.JCudaDriver.cuCtxSynchronize;
import static jcuda.driver.JCudaDriver.cuDeviceGet;
import static jcuda.driver.JCudaDriver.cuInit;
import static jcuda.driver.JCudaDriver.cuLaunchKernel;
import static jcuda.driver.JCudaDriver.cuMemAlloc;
import static jcuda.driver.JCudaDriver.cuMemFree;
import static jcuda.driver.JCudaDriver.cuMemcpyDtoH;
import static jcuda.driver.JCudaDriver.cuMemcpyHtoD;
import static jcuda.driver.JCudaDriver.cuModuleGetFunction;
import static jcuda.driver.JCudaDriver.cuModuleLoad;

import java.util.Random;

import jcuda.Pointer;
import jcuda.Sizeof;
import jcuda.driver.CUcontext;
import jcuda.driver.CUdevice;
import jcuda.driver.CUdeviceptr;
import jcuda.driver.CUfunction;
import jcuda.driver.CUmodule;
import jcuda.driver.JCudaDriver;

public class FacilityLocationCuda implements Function {
	private final double[][] sim;
	private final CUdeviceptr deviceSimilarity;
	private final CUdeviceptr deviceOutput;
	// private final CUdeviceptr deviceSet;
	private final CUfunction function;
	
	public static void main(String args[]) {
		int nrDim = 100;
		int nrEntries = 10;
		
		System.out.println("Working.");
		Random rg = new Random(124);
		
		// Create random similiarity matrix
		double[][] sim = UtilsTest.generateSimiliarityMatrix(nrDim, rg);
		
		// random set
		int[] S = UtilsTest.randomSet(nrEntries, 100, rg);
		
		FacilityLocationCuda fc = new FacilityLocationCuda(sim);
		
		// evaluate traditionally
		double val1 = fc.eval(S);
		double val2 = fc.evalCUDA(S);
		System.out.println("Classic code: " + val1);
		System.out.println("CUDA code: " + val2);
		
		// perform timing
		System.out.println("Performing timing...");
		int nrTrials = 100000;
		long startTime = System.nanoTime();
		for (int i = 0; i < nrTrials; i++) {
			fc.eval(S);
		}
		long endTime = System.nanoTime();
		System.out.println(String.format("Classic evaluation took %f seconds (%f us per call).", (endTime - startTime) / 1e9, (endTime - startTime) / (1e9 * nrTrials / 1000000)));
		
		startTime = System.nanoTime();
		for (int i = 0; i < nrTrials; i++) {
			fc.evalCUDA(S);
		}
		endTime = System.nanoTime();
		System.out.println(String.format("CUDA evaluation took %f seconds (%f us per call).", (endTime - startTime) / 1e9, (endTime - startTime) / (1e9 * nrTrials / 1000000)));
	}
	
	public FacilityLocationCuda(double[][] sim) {
		this.sim = sim;
		
		// ----
		// Enable exceptions and omit all subsequent error checks
		JCudaDriver.setExceptionsEnabled(true);
		
		// Create the PTX file by calling the NVCC
		String ptxFileName = "/afs/spsc.tugraz.at/user/stschia/ECCV2014/Codes/Java/SubmodularFunctions/src/FacilityLocation.ptx";
		
		// Initialize the driver and create a context for the first device.
		cuInit(0);
		CUdevice device = new CUdevice();
		cuDeviceGet(device, 0);
		CUcontext context = new CUcontext();
		cuCtxCreate(context, 0, device);
		
		// Load the ptx file.
		CUmodule module = new CUmodule();
		cuModuleLoad(module, ptxFileName);
		
		// Obtain a function pointer to the "add" function.
		function = new CUfunction();
		cuModuleGetFunction(function, module, "FacilityLocation");
		
		int numElements = sim.length * sim.length;
		
		// Allocate and fill the host input data
		float hostSimilarity[] = new float[numElements];
		
		int idx = 0;
		for (int i = 0; i < sim.length; i++) {
			for (int j = 0; j < sim.length; j++) {
				hostSimilarity[idx] = (float) sim[i][j];
				idx++;
			}
		}
		
		// Allocate the device input data, and copy the
		// host input data to the device
		deviceSimilarity = new CUdeviceptr();
		cuMemAlloc(deviceSimilarity, numElements * Sizeof.FLOAT);
		cuMemcpyHtoD(deviceSimilarity, Pointer.to(hostSimilarity), numElements * Sizeof.FLOAT);
		
		// Allocate device output memory
		deviceOutput = new CUdeviceptr();
		cuMemAlloc(deviceOutput, Sizeof.FLOAT);
		
		// TODO: REMOVE
		// deviceSet = new CUdeviceptr();
		// cuMemAlloc(deviceSet, sim.length * Sizeof.INT);
	}
	
	public double evalCUDA(int[] S) {
		
		CUdeviceptr deviceSet = new CUdeviceptr();
		cuMemAlloc(deviceSet, S.length * Sizeof.INT);
		cuMemcpyHtoD(deviceSet, Pointer.to(S), S.length * Sizeof.INT);
		
		// Set up the kernel parameters: A pointer to an array
		// of pointers which point to the actual values.
		Pointer kernelParameters = Pointer.to(Pointer.to(new int[] { S.length }), Pointer.to(deviceSet), Pointer.to(new int[] { sim.length }), Pointer.to(deviceSimilarity), Pointer.to(deviceOutput));
		
		// Call the kernel function.
		int blockSizeX = sim.length;
		int gridSizeX = 1;
		cuLaunchKernel(function, gridSizeX, 1, 1, // Grid dimension
		        blockSizeX, 1, 1, // Block dimension
		        500, null, // Shared memory size and stream
		        kernelParameters, null // Kernel- and extra parameters
		);
		cuCtxSynchronize();
		
		// Allocate host output memory and copy the device output
		// to the host.
		float hostOutput[] = new float[1];
		cuMemcpyDtoH(Pointer.to(hostOutput), deviceOutput, Sizeof.FLOAT);
		
		// Verify the result
		// boolean passed = true;
		// float expected = (float) eval(S);
		// System.out.println("Host output: " + hostOutput[0]);
		// if (Double.isNaN(hostOutput[0]) || Math.abs(hostOutput[0] - expected) > 1e-5) {
		// System.out.println("At index " + 0 + " found " + hostOutput[0] + " but expected " + expected);
		// passed = false;
		// }
		// System.out.println("Test " + (passed ? "PASSED" : "FAILED"));
		cuMemFree(deviceSet);
		
		return hostOutput[0];
	}
	
	protected void finalize() {
		// Clean up.
		cuMemFree(deviceSimilarity);
		cuMemFree(deviceOutput);
	}
	
	public FacilityLocationCuda(Matrix sim) {
		this(sim.getMatrix());
	}
	
	// For octave compatibility
	public double eval(int St) {
		int[] S = new int[1];
		S[0] = St;
		return eval(S);
	}
	
	public double eval(int[] S) {
		if (S == null) {
			return 0.0;
		}
		
		double val = 0.0;
		for (int j = 0; j < sim.length; j++) {
			double maxVal = Double.NEGATIVE_INFINITY;
			for (int i : S) {
				if (sim[j][i - 1] > maxVal) {
					maxVal = sim[j][i - 1];
				}
			}
			val += maxVal;
		}
		return val / 100.0;
	}
	
	public double[] eval(int[] B, int[] candidates) {
		return null;
	}
	
}
