import static jcuda.driver.JCudaDriver.cuCtxPushCurrent;
import static jcuda.driver.JCudaDriver.cuCtxPopCurrent;
import static jcuda.driver.JCudaDriver.cuCtxSynchronize;
import static jcuda.driver.JCudaDriver.cuLaunchKernel;
import static jcuda.driver.JCudaDriver.cuMemAlloc;
import static jcuda.driver.JCudaDriver.cuMemFree;
import static jcuda.driver.JCudaDriver.cuMemcpyDtoH;
import static jcuda.driver.JCudaDriver.cuMemcpyHtoD;
import static jcuda.driver.JCudaDriver.cuModuleGetFunction;
import static jcuda.driver.JCudaDriver.cuModuleLoad;

import java.util.Random;

import jcuda.Pointer;
import jcuda.Sizeof;
import jcuda.driver.CUcontext;
import jcuda.driver.CUdeviceptr;
import jcuda.driver.CUfunction;
import jcuda.driver.CUmodule;
import jcuda.driver.JCudaDriver;

public class FidelityCuda implements Function {
	public static String ptxFileName = "/afs/spsc.tugraz.at/user/stschia/ECCV2014/Codes/Java/SubmodularFunctions/src/Fidelity.ptx";
	
	private double[][] bow;
	private double threshold;
	private int nrWords;
	private int nrImages;
	
	// Cuda
	private CUcontext context;
	private CUfunction function;
	private CUmodule module;
	private final float[] hostBOW;
	
	public void setContext(CUcontext context) {
		this.context = context;
		
		cuCtxPushCurrent(context);
		
		// Load the ptx file.
		CUmodule module = new CUmodule();
		cuModuleLoad(module, ptxFileName);
		
		// Obtain a function pointer to the "add" function.
		function = new CUfunction();
		cuModuleGetFunction(function, module, "Fidelity");
		
		cuCtxPopCurrent(context);
	}
	
	public static void main(String args[]) {
		int nrElements = 100;
		int nrWords = 1500;
		int nrEntries = 10;
		int nrTrials = 100;
		double threshold = 900;
		
		System.out.println("Working.");
		Random rg = new Random(123);
		
		// Fire up CUDA
		InitCuda cuda = new InitCuda();
		
		// Create random similiarity matrix
		double[][] bow = UtilsTest.generateBOW(nrElements, nrWords, rg);
		
		// random set
		int[] S = UtilsTest.randomSet(nrEntries, nrElements, rg);
		int[] candidates = UtilsTest.generateCandidatesSet(S, nrElements, rg);
		
		FidelityCuda fidelity = new FidelityCuda(bow, threshold);
		fidelity.setContext(cuda.getContext());
		
		// evaluate traditionally
		S = new int[] { 1 };
		double[] val1 = fidelity.evalClassic(S, UtilsTest.generateCandidatesSet(S, nrElements, rg));
		double[] val2 = fidelity.eval(S, UtilsTest.generateCandidatesSet(S, nrElements, rg));
		for (int i = 0; i < val1.length; i++) {
			System.out.println("CANDIDATE " + i);
			System.out.println("  Classic code: " + val1[i]);
			System.out.println("  CUDA code: " + val2[i]);
			System.out.println("OK: " + (Math.abs(val1[i] - val2[i]) < 1e-4));
		}
		
		// perform timing
		System.out.println("Performing timing...");
		long startTime = System.nanoTime();
		for (int i = 0; i < nrTrials; i++) {
			// candidates = UtilsTest.generateCandidatesSet(S, nrElements, rg);
			fidelity.evalClassic(S, candidates);
		}
		long endTime = System.nanoTime();
		System.out.println(String.format("Classic evaluation took %f seconds (%f us per call).", (endTime - startTime) / 1e9, (endTime - startTime) / (1e9 * nrTrials / 1000000)));
		
		startTime = System.nanoTime();
		for (int i = 0; i < nrTrials; i++) {
			// candidates = UtilsTest.generateCandidatesSet(S, nrElements, rg);
			fidelity.eval(S, candidates);
		}
		endTime = System.nanoTime();
		System.out.println(String.format("CUDA evaluation took %f seconds (%f us per call).", (endTime - startTime) / 1e9, (endTime - startTime) / (1e9 * nrTrials / 1000000)));
	}
	
	public FidelityCuda(Matrix bow, double threshold) {
		this(bow.getMatrix(), threshold);
	}
	
	public FidelityCuda(double[][] bow, double threshold) {
		this.bow = bow;
		this.threshold = threshold;
		this.nrWords = bow.length;
		this.nrImages = bow[0].length;
		
		// ---- // Enable exceptions and omit all subsequent error checks
		
		// JCudaDriver.setExceptionsEnabled(true);
		// Initialize the driver and create a context for the first device.
		// int errorFlag = cudaSetDeviceFlags(cudaDeviceMapHost);
		// System.out.println("Error when setting flag: " + errorFlag);
		
		// cuInit(0);
		// CUdevice device = new CUdevice();
		// cuDeviceGet(device, 0);
		
		// errorFlag = cudaSetDeviceFlags(cudaDeviceMapHost);
		// System.out.println("Error when setting flag: " + errorFlag);
		
		// CUcontext context = new CUcontext();
		
		// cuCtxCreate(context, cudaDeviceMapHost, device); // Check if the device supports mapped host memory
		// cuCtxCreate(context, 0, device); // Check if the device supports mapped host memory
		
		// cudaDeviceProp deviceProperties = new cudaDeviceProp();
		// cudaGetDeviceProperties(deviceProperties, 0);
		// if (deviceProperties.canMapHostMemory == 0) {
		// System.err.println("This device can not map host memory");
		// System.err.println(deviceProperties.toFormattedString());
		// System.exit(-1);
		// }
		// System.out.println("PROPS: " + deviceProperties);
		// System.out.println("We can use mapped memory.");
		// Set the flag indicating that mapped memory will be used
		// cudaSetDeviceFlags(cudaDeviceMapHost);
		
		// JCudaDriver.setExceptionsEnabled(true);
		// cuInit(0);
		// CUcontext context = new CUcontext();
		// CUdevice device = new CUdevice();
		// try {
		// cuCtxPopCurrent(context);
		// cuCtxPushCurrent(context);
		// } catch (jcuda.CudaException e) {
		// cuDeviceGet(device, 0);
		// cuCtxCreate(context, 0, device);
		// }
		// System.out.println("Context: " + context);
		context = null;
		
		int numElements = bow.length * bow[0].length;
		
		// Allocate and fill the host input data
		float hostBOW[] = new float[numElements];
		
		int idx = 0;
		for (int i = 0; i < bow.length; i++) {
			for (int j = 0; j < bow[0].length; j++) {
				hostBOW[idx] = (float) bow[i][j];
				idx++;
			}
		}
		
		this.hostBOW = hostBOW;
		
		// // Allocate the device input data, and copy the
		// // host input data to the device
		// deviceBOW = new CUdeviceptr();
		// cuMemAlloc(deviceBOW, numElements * Sizeof.FLOAT);
		// cuMemcpyHtoD(deviceBOW, Pointer.to(hostBOW), numElements * Sizeof.FLOAT);
	}
	
	public double eval(int[] S) {
		if (S == null)
			S = new int[] {};
		
		// sumBow = sum(bow(:, S), 2);
		double[] sumBow = new double[nrWords];
		for (int j = 0; j < nrWords; j++) {
			double[] imBow = bow[j];
			for (int i : S) {
				sumBow[j] += imBow[i - 1];
			}
			
		}
		
		// sim = min(repmat(sumBow, 1, 100), bow);
		// val = min(sum(sim, 1), alpha*ones(1,100));
		double[] sum = new double[nrImages];
		for (int j = 0; j < bow.length; j++) {
			double sumBowVal = sumBow[j];
			double[] imBow = bow[j];
			for (int i = 0; i < nrImages; i++) {
				if (sumBowVal < imBow[i])
					sum[i] += sumBowVal;
				else
					sum[i] += imBow[i];
			}
		}
		// System.out.println("SUM: " + sum[0]);
		
		// val = sum(val) / (100*alpha);
		double val = 0;
		for (int j = 0; j < nrImages; j++) {
			if (sum[j] < threshold)
				val += sum[j];
			else
				val += threshold;
		}
		return val / (100.0 * threshold);
	}
	
	public double[] evalClassic(int[] B, int[] cand) {
		if (cand == null)
			cand = new int[] {};
		
		int nrCandidates = cand.length;
		double[] values = new double[nrCandidates];
		
		// sumBow = sum(bow(:, S), 2);
		double[] sumBow = new double[nrWords];
		for (int j = 0; j < nrWords; j++) {
			double[] imBow = bow[j];
			for (int i : B) {
				sumBow[j] += imBow[i - 1];
			}
		}
		
		// go on
		
		// sim = min(repmat(sumBow, 1, 100), bow);
		// val = min(sum(sim, 1), alpha*ones(1,100));
		double[][] sum = new double[nrImages][nrCandidates];
		for (int c = 0; c < nrCandidates; c++) {
			int candidate = cand[c];
			for (int j = 0; j < bow.length; j++) {
				double sumBowVal = sumBow[j] + bow[j][candidate - 1];
				double[] imBow = bow[j];
				for (int i = 0; i < nrImages; i++) {
					if (sumBowVal < imBow[i])
						sum[i][c] += sumBowVal;
					else
						sum[i][c] += imBow[i];
				}
			}
		}
		
		// val = sum(val) / (100*alpha);
		for (int c = 0; c < nrCandidates; c++) {
			double val = 0.0;
			for (int j = 0; j < nrImages; j++) {
				if (sum[j][c] < threshold)
					val += sum[j][c];
				else
					val += threshold;
			}
			values[c] = val / (100.0 * threshold);
		}
		return values;
	}
	
	public double[] eval(int[] S, int[] candidates) {
		if (candidates == null) {
			candidates = new int[] {};
			return new double[] {};
		}
		
		int error;
		
		JCudaDriver.setExceptionsEnabled(true);
		
		CUdeviceptr deviceOutput;
		CUdeviceptr deviceBOW;
		
		if (context == null)
			throw new RuntimeException("CUDA context not set.");
		cuCtxPushCurrent(context);
		
		// System.out.println("USING CONTEXT: " + context);
		
		// try {
		// // System.out.println("Using existing context.");
		// cuCtxPopCurrent(context);
		// cuCtxPushCurrent(context);
		//
		// // Load the ptx file.
		// CUmodule module = new CUmodule();
		// cuModuleLoad(module, ptxFileName);
		//
		// // Obtain a function pointer to the "add" function.
		// function = new CUfunction();
		// cuModuleGetFunction(function, module, "Fidelity");
		// } catch (jcuda.CudaException e) {
		// // System.out.println("Setting up new context.");
		// cuDeviceGet(device, 0);
		// cuCtxCreate(context, 0, device);
		//
		// // Load the ptx file.
		// CUmodule module = new CUmodule();
		// cuModuleLoad(module, ptxFileName);
		//
		// // Obtain a function pointer to the "add" function.
		// function = new CUfunction();
		// cuModuleGetFunction(function, module, "Fidelity");
		// }
		// System.out.println("EVAL. CONTEXT: " + context);
		
		// Allocate device output memory
		deviceOutput = new CUdeviceptr();
		error = cuMemAlloc(deviceOutput, Sizeof.INT * candidates.length);
		if (error != 0)
			throw new RuntimeException("Error while reserving memory on the GPU. " + error);
		
		CUdeviceptr deviceSet = new CUdeviceptr();
		if (S.length == 0) {
		  cuMemAlloc(deviceSet, 1 * Sizeof.INT);
		  cuMemcpyHtoD(deviceSet, Pointer.to(new int[] {-1}), 1 * Sizeof.INT);
		} else {
		  cuMemAlloc(deviceSet, S.length * Sizeof.INT);
		  cuMemcpyHtoD(deviceSet, Pointer.to(S), S.length * Sizeof.INT);
		}
		
		CUdeviceptr deviceCandidates = new CUdeviceptr();
		cuMemAlloc(deviceCandidates, candidates.length * Sizeof.INT);
		cuMemcpyHtoD(deviceCandidates, Pointer.to(candidates), candidates.length * Sizeof.INT);
		
		// Allocate the device input data, and copy the
		// host input data to the device
		int numElements = hostBOW.length;
		
		/*
		 * Pointer hostBOWt = new Pointer(); cudaHostAlloc(hostBOWt, Sizeof.FLOAT*numElements, cudaHostAllocMapped); FloatBuffer fb = hostBOWt.getByteBuffer(0, numElements *
		 * Sizeof.FLOAT).order(ByteOrder.nativeOrder()).asFloatBuffer(); fb.position(0); fb.put(hostBOW); fb.position(0); Pointer deviceBOW = new Pointer(); cudaHostGetDevicePointer(deviceBOW,
		 * hostBOWt, 0);
		 */
		
		deviceBOW = new CUdeviceptr();
		error = cuMemAlloc(deviceBOW, numElements * Sizeof.FLOAT);
		if (error != 0)
			throw new RuntimeException("Could not reserve memory for bag-of-words.");
		error = cuMemcpyHtoD(deviceBOW, Pointer.to(hostBOW), numElements * Sizeof.FLOAT);
		if (error != 0)
			throw new RuntimeException("Could not copy bag-of-words to device.");
		
		// Set up the kernel parameters: A pointer to an array
		// of pointers which point to the actual values.
		Pointer kernelParameters = Pointer.to(Pointer.to(new int[] { S.length }), Pointer.to(deviceSet), Pointer.to(new int[] { candidates.length }), Pointer.to(deviceCandidates),
		        Pointer.to(new int[] { bow[0].length }), Pointer.to(new int[] { bow.length }), Pointer.to(deviceBOW), Pointer.to(new float[] { (float) threshold }), Pointer.to(deviceOutput));
		
		// Call the kernel function.
		int blockSizeX = bow[0].length;
		int gridSizeX = candidates.length;
		error = cuLaunchKernel(function, gridSizeX, 1, 1, // Grid dimension
		        blockSizeX, 1, 1, // Block dimension
		        Sizeof.INT * (S.length + 2) + Sizeof.FLOAT * nrWords + Sizeof.FLOAT * nrImages, null, // Shared memory size and stream
		        kernelParameters, null // Kernel- and extra parameters
		);
		if (error != 0)
		  throw new RuntimeException("Error when launching kernel.");
		error = cuCtxSynchronize();
		if (error != 0)
		  throw new RuntimeException("Error synchronizing.");
		
		
		// Allocate host output memory and copy the device output
		// to the host.
		float hostOutput[] = new float[candidates.length];
		error = cuMemcpyDtoH(Pointer.to(hostOutput), deviceOutput, candidates.length * Sizeof.FLOAT);
		if (error != 0)
		  throw new RuntimeException("Copying back results failed.");
		
		cuMemFree(deviceSet);
		cuMemFree(deviceCandidates);
		cuMemFree(deviceOutput);
		cuMemFree(deviceBOW);
		// cudaFreeHost(hostBOWt);
		
		double output[] = new double[candidates.length];
		for (int i = 0; i < output.length; i++) {
			output[i] = hostOutput[i];
		}
		
		cuCtxPopCurrent(context);
		
		return output;
	}
	
	protected void finalize() {
		// Clean up.
		// cuMemFree(deviceBOW);
		// cuMemFree(deviceOutput);
	}
	
	// >> For octave compatibility >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
	
	public double eval(int St) {
		int[] S = new int[] { St };
		return eval(S);
	}
	
	public double[] eval(int St, int[] candidates) {
		int[] S = new int[] { St };
		return eval(S, candidates);
	}
	
	public double[] eval(int[] S, int candidatest) {
		int[] candidates = new int[] { candidatest };
		return eval(S, candidates);
	}
}
