import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class OneMinusRouge implements Function {
	private static double tol = 1e-6;
	
	private final int[][] summaries;
	private final double[][] bow;
	private final double[][] bowSummaries;
	private final int nrWords;
	private final int nrSummaries;
	private final double normalization;
	
	private double rougeMin;
	private double rougeMax;
	
	private final double[][] thresholds; // one threshold vector per visual word
	private final double[][] offsets;
	private final int[][] counts;
	
	// TODO: Fix first argument
	public OneMinusRouge(Matrix summaries, Matrix bow, Matrix bowSummaries) {
		this(null, bow.getMatrix(), bowSummaries.getMatrix());
	}
	
	public OneMinusRouge(int[][] summaries, double[][] bow, double[][] bowSummaries) {
		this.summaries = summaries;
		this.bow = bow;
		this.bowSummaries = bowSummaries;
		this.nrWords = bow.length;
		this.nrSummaries = bowSummaries[0].length;
		
		this.rougeMin = 0.0;
		this.rougeMax = 1.0;
		
		// compute normalization
		double normalization = 0.0;
		for (int i = 0; i < nrWords; i++) {
			for (int j = 0; j < nrSummaries; j++) {
				normalization += bowSummaries[i][j];
			}
		}
		this.normalization = normalization;
		
		// compute thresholds and counts
		double[][] thresholds = new double[nrWords][];
		double[][] offsets = new double[nrWords][];
		int[][] counts = new int[nrWords][];
		
		for (int w = 0; w < nrWords; w++) {
			double[] sThresholds = bowSummaries[w].clone();
			Arrays.sort(sThresholds);
			
			List<Double> tThresholds = new ArrayList<Double>();
			List<Double> tOffsets = new ArrayList<Double>();
			List<Integer> tCounts = new ArrayList<Integer>();
			tOffsets.add(0.0);
			double lastVal = 0.0;
			int count = 0;
			for (int s = 0; s < nrSummaries; s++) {
				double curVal = sThresholds[s];
				if (Math.abs(curVal - lastVal) < tol) {
					
				} else {
					tCounts.add(nrSummaries - count);
					tThresholds.add(lastVal);
					tOffsets.add(tOffsets.get(tOffsets.size() - 1) + (nrSummaries - count) * (curVal - lastVal));
					lastVal = curVal;
					// count = 1;
				}
				count++;
			}
			tCounts.add(nrSummaries - count);
			tThresholds.add(lastVal);
			
			// for sanity
			tCounts.add(0);
			tThresholds.add(Double.POSITIVE_INFINITY);
			tOffsets.add(Double.POSITIVE_INFINITY);
			
			// convert to array
			thresholds[w] = new double[tThresholds.size()];
			offsets[w] = new double[tOffsets.size()];
			counts[w] = new int[tCounts.size()];
			for (int i = 0; i < tThresholds.size(); i++) {
				thresholds[w][i] = tThresholds.get(i);
				offsets[w][i] = tOffsets.get(i);
				counts[w][i] = tCounts.get(i);
			}
		}
		
		this.thresholds = thresholds;
		this.offsets = offsets;
		this.counts = counts;
	}
	
	public void setRougeLimits(double rougeMin, double rougeMax) {
		this.rougeMin = rougeMin;
		this.rougeMax = rougeMax;
	}
	
	public double evalClassic(int[] S) {
		if (S == null)
			S = new int[] {};
		
		// sumBow = sum(bow(:, S), 2);
		double[] sumBow = new double[nrWords];
		for (int j = 0; j < nrWords; j++) {
			double[] imBow = bow[j];
			for (int i : S) {
				sumBow[j] += imBow[i - 1];
			}
		}
		
		double val = 0.0;
		for (int j = 0; j < nrWords; j++) {
			double sumBowVal = sumBow[j];
			double[] imBow = bowSummaries[j];
			for (int i = 0; i < nrSummaries; i++) {
				if (sumBowVal < imBow[i])
					val += sumBowVal;
				else
					val += imBow[i];
			}
		}
		
		return 1.0 - ((val / normalization) - rougeMin) / (rougeMax - rougeMin);
	}
	
	public double[] evalClassic(int[] B, int[] candidates) {
		if (candidates == null)
			candidates = new int[] {};
		
		int nrCandidates = candidates.length;
		double[] values = new double[nrCandidates];
		
		// sumBow = sum(bow(:, S), 2);
		double[] sumBow = new double[nrWords];
		for (int j = 0; j < nrWords; j++) {
			double[] imBow = bow[j];
			for (int i : B) {
				sumBow[j] += imBow[i - 1];
			}
		}
		
		for (int c = 0; c < nrCandidates; c++) {
			int candidate = candidates[c];
			
			double val = 0.0;
			for (int j = 0; j < nrWords; j++) {
				double sumBowVal = sumBow[j] + bow[j][candidate - 1];
				double[] imBow = bowSummaries[j];
				for (int i = 0; i < nrSummaries; i++) {
					if (sumBowVal < imBow[i])
						val += sumBowVal;
					else
						val += imBow[i];
				}
			}
			
			values[c] = 1.0 - ((val / normalization) - rougeMin) / (rougeMax - rougeMin);
		}
		
		return values;
	}
	
	// uses precomputed lists for lookup
	public double eval(int[] S) {
		if (S == null)
			S = new int[] {};
		
		// sumBow = sum(bow(:, S), 2);
		double[] sumBow = new double[nrWords];
		for (int j = 0; j < nrWords; j++) {
			double[] imBow = bow[j];
			for (int i : S) {
				sumBow[j] += imBow[i - 1];
			}
		}
		
		double val = 0.0;
		for (int w = 0; w < nrWords; w++) {
			double sumBowVal = sumBow[w];
			double[] thresholds = this.thresholds[w];
			double[] offsets = this.offsets[w];
			int[] counts = this.counts[w];
			int nr = counts.length;
			int idx = 0;
			for (int k = 0; k < nr; k++) {
				if (sumBowVal < thresholds[k]) {
					idx = k - 1;
					break;
				}
			}
			val += offsets[idx] + (sumBowVal - thresholds[idx]) * counts[idx];
		}
		
		return 1.0 - ((val / normalization) - rougeMin) / (rougeMax - rougeMin);
	}
	
	public double[] eval(int[] B, int[] candidates) {
		if (candidates == null)
			candidates = new int[] {};
		
		int nrCandidates = candidates.length;
		double[] values = new double[nrCandidates];
		
		// sumBow = sum(bow(:, S), 2);
		double[] sumBow = new double[nrWords];
		for (int j = 0; j < nrWords; j++) {
			double[] imBow = bow[j];
			for (int i : B) {
				sumBow[j] += imBow[i - 1];
			}
		}
		
		for (int c = 0; c < nrCandidates; c++) {
			int candidate = candidates[c];
			
			double val = 0.0;
			for (int j = 0; j < nrWords; j++) {
				double sumBowVal = sumBow[j] + bow[j][candidate - 1];
				double[] thresholds = this.thresholds[j];
				double[] offsets = this.offsets[j];
				int[] counts = this.counts[j];
				int nr = counts.length;
				int idx = 0;
				for (int k = 0; k < nr; k++) {
					if (sumBowVal < thresholds[k]) {
						idx = k - 1;
						break;
					}
				}
				val += offsets[idx] + (sumBowVal - thresholds[idx]) * counts[idx];
			}
			
			values[c] = 1.0 - ((val / normalization) - rougeMin) / (rougeMax - rougeMin);
		}
		
		return values;
	}
	
	// >> For octave compatibility >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
	
	public double eval(int St) {
		int[] S = new int[] { St };
		return eval(S);
	}
	
	public double[] eval(int St, int[] candidates) {
		int[] S = new int[] { St };
		return eval(S, candidates);
	}
	
	public double[] eval(int[] S, int candidatest) {
		int[] candidates = new int[] { candidatest };
		return eval(S, candidates);
	}
}
