import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class Server {
	public static int port = 9999;
	private ServerSocket serverSocket;
	private Thread serverThread;
	private Thread serverStatusThread;
	private Map<String, Queue<Work>> workQueueMap;
	private Map<String, double[]> weightsMap;
	private Lock weightsLock = new ReentrantLock();
	private Map<String, Queue<WorkResult>> workResultQueueMap;
	private Map<String, Integer> inProcessMap;
	private boolean shutdown;
	private static int inProcessThreshold = 10;
	
	public Server() throws IOException {
		this.shutdown = false;
		
		// workQueue = new LinkedList<Work>();
		workQueueMap = new HashMap<String, Queue<Work>>();
		workResultQueueMap = new HashMap<String, Queue<WorkResult>>();
		weightsMap = new HashMap<String, double[]>();
		inProcessMap = new HashMap<String, Integer>();
		// workResultQueue = new LinkedList<WorkResult>();
	}
	
	public void setWeights(String id, double[] weights) {
		weightsLock.lock();
		this.weightsMap.put(id, weights);
		weightsLock.unlock();
	}
	
	public WorkResult getWorkResult(String id) {
		if (!workResultQueueMap.containsKey(id) || workResultQueueMap.get(id).isEmpty()) {
			return null;
		} else {
			WorkResult workResult = workResultQueueMap.get(id).poll();
			return workResult;
		}
	}
	
	private double[] getGradient(String id) {
		if (!workResultQueueMap.containsKey(id) || workResultQueueMap.get(id).isEmpty()) {
			return null;
		} else {
			WorkResult workResult = workResultQueueMap.get(id).poll();
			return workResult.getGrad();
		}
	}
	
	// TODO: This doesn't mean that the work is actually done.
	public boolean isWorkDone(String id) {
		if (workQueueMap.get(id) != null && workQueueMap.get(id).isEmpty()) {
			return true;
		} else {
			return false;
		}
	}
	
	public double[] getWeights(String id) {
		return this.weightsMap.get(id);
	}
	
	private void addWork(String id, String inference, int[] sample, int dataset) {
		Work work = new Work(id, inference, sample, dataset);
		
		if (!workQueueMap.containsKey(id)) {
			workQueueMap.put(id, new LinkedList<Work>());
		}
		workQueueMap.get(id).add(work);
	}
	
	public void startServer() throws IOException {
		serverSocket = new ServerSocket(port);
		
		serverThread = new ServerThread();
		serverThread.setDaemon(true);
		serverThread.start();
		
		serverStatusThread = new ServerStatusThread();
		serverStatusThread.start();
	}
	
	public void stopServer() {
		try {
			shutdown = true;
			serverSocket.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	private boolean allWorkQueuesEmpty() {
		for (String id : workQueueMap.keySet()) {
			if (workQueueMap.get(id).size() > 0) {
				return false;
			}
		}
		
		return true;
	}
	
	private boolean allResultsQueuesFull(int threshold) {
		if (workResultQueueMap.size() == 0) {
			return false;
		}
		
		for (String id : workResultQueueMap.keySet()) {
			if (workResultQueueMap.get(id).size() < threshold) {
				return false;
			}
		}
		
		return true;
	}
	
	public class ServerStatusThread extends Thread {
		public void run() {
			System.out.println("Starting server status thread...");
			
			while (!shutdown) {
				System.out.println("Queue status:");
				for (Entry<String, Queue<Work>> entry : workQueueMap.entrySet()) {
					int nrResults = 0;
					if (workResultQueueMap.get(entry.getKey()) != null) {
						nrResults = workResultQueueMap.get(entry.getKey()).size();
					}
					int inProcess = inProcessMap.get(entry.getKey());
					System.out.println(String.format("Queue \"%s\": #work=%4d, #results=%4d [in process=%2d/%2d]", entry.getKey(), entry.getValue().size(), nrResults, inProcess, inProcessThreshold));
				}
				
				try {
					Thread.sleep(10000);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				System.out.println();
			}
			
			System.out.println("Server shutting down.");
		}
	}
	
	public class ServerThread extends Thread {
		public void run() {
			Message resp = null;
			System.out.println("Starting server...");
			
			int id = 0;
			while (true) {
				try {
					Socket socket = serverSocket.accept();
					
					// System.out.println("[Server] Accepted request.");
					
					ObjectInputStream objectInput = new ObjectInputStream(socket.getInputStream());
					try {
						resp = (Message) objectInput.readObject();
					} catch (ClassNotFoundException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					
					switch (resp.getType()) {
					case Message.GET_WORK:
						// System.out.println("Full " + allResultsQueuesFull(10) + " / empty " + allWorkQueuesEmpty());
						if (allWorkQueuesEmpty() || allResultsQueuesFull(10)) {
							// System.out.println("[Server] Currently there is no work available.");
							Message msg = new Message("Server", Message.WAIT);
							ObjectOutputStream objectOutput = new ObjectOutputStream(socket.getOutputStream());
							objectOutput.writeObject(msg);
						} else {
							// TODO: Sort length of queues by number of work in reverse order
							
							Work work = null;
							for (String lid : workQueueMap.keySet()) {
								if (workResultQueueMap.get(lid) != null && workResultQueueMap.get(lid).size() > 10)
									continue;
								if (inProcessMap.get(lid) >= inProcessThreshold)
									continue;
								work = workQueueMap.get(lid).poll();
								if (work != null) {
									break;
								}
							}
							if (work == null) {
								Message msg = new Message("Server", Message.WAIT);
								ObjectOutputStream objectOutput = new ObjectOutputStream(socket.getOutputStream());
								objectOutput.writeObject(msg);
								break;
							}
							String lid3 = work.getId();
							
							// update weights
							weightsLock.lock();
							work.setWeights(getWeights(work.getId()));
							inProcessMap.put(lid3, inProcessMap.get(lid3) + 1);
							weightsLock.unlock();
							
							Message msg = new Message(work.getId(), Message.WORK);
							msg.setWork(work);
							ObjectOutputStream objectOutput = new ObjectOutputStream(socket.getOutputStream());
							objectOutput.writeObject(msg);
						}
						break;
					
					case Message.GET_GRADIENT:
						// System.out.println("Full " + allResultsQueuesFull(10) + " / empty " + allWorkQueuesEmpty());
						String lid = resp.getId();
						// System.out.println("Checking queue " + lid);
						if (!workResultQueueMap.containsKey(lid) || workResultQueueMap.get(lid).size() == 0) {
							// System.out.println("[Server] Currently there is no work available.");
							Message msg = new Message("Server", Message.GRADIENT);
							msg.setWorkResult(null);
							ObjectOutputStream objectOutput = new ObjectOutputStream(socket.getOutputStream());
							objectOutput.writeObject(msg);
						} else {
							WorkResult workResult = workResultQueueMap.get(lid).poll();
							if (workResult == null) {
								throw new RuntimeException("There should be a work result available.");
							}
							
							Message msg = new Message(workResult.getId(), Message.GRADIENT);
							msg.setWorkResult(workResult);
							ObjectOutputStream objectOutput = new ObjectOutputStream(socket.getOutputStream());
							objectOutput.writeObject(msg);
						}
						break;
					
					case Message.IS_WORK_DONE:
						Message msg = new Message("Server", Message.WORK_STATUS);
						if (workQueueMap.get(resp.getId()) == null) {
							msg.setWorkDone(true);
						} else {
							if (workQueueMap.get(resp.getId()).size() == 0) {
								msg.setWorkDone(true);
							} else {
								msg.setWorkDone(false);
							}
						}
						ObjectOutputStream objectOutput = new ObjectOutputStream(socket.getOutputStream());
						objectOutput.writeObject(msg);
						break;
					
					case Message.WORK_DONE:
						// System.err.println("[Server] Received work result. Result will be added to queue.");
						if (!workResultQueueMap.containsKey(resp.getId())) {
							workResultQueueMap.put(resp.getId(), new LinkedList<WorkResult>());
						}
						workResultQueueMap.get(resp.getId()).add(resp.getWorkResult());
						
						weightsLock.lock();
						String lid4 = resp.getId();
						if (inProcessMap.get(lid4) != null) {
							inProcessMap.put(lid4, inProcessMap.get(lid4) - 1);
						}
						weightsLock.unlock();
						
						// System.out.println("WorkResultQueue " + resp.getId() + " containts " + workResultQueueMap.get(resp.getId()).size() + " results.");
						break;
					
					case Message.SET_WEIGHTS:
						weightsLock.lock();
						setWeights(resp.getId(), resp.getWeights());
						weightsLock.unlock();
						break;
					
					case Message.CLEAR_QUEUES:
						String lid2 = resp.getId();
						workQueueMap.remove(lid2);
						workResultQueueMap.remove(lid2);
						inProcessMap.remove(lid2);
						break;
					
					case Message.ADD_WORK:
						// System.out.println("Adding workList: " + resp.getWorkList());
						// System.out.println("Number of items: " + resp.getWorkList().size());
						// + "/b: " + resp.getWork().getSample());
						for (Work work : resp.getWorkList()) {
							addWork(resp.getId(), work.getInference(), work.getSample(), work.getDataset());
						}
						inProcessMap.put(resp.getId(), 0);
						break;
					
					default:
						System.err.println("[Server] Invalid message received.");
					}
					
					socket.close();
				} catch (java.net.SocketException e) {
					if (shutdown) {
						System.out.println("It is time to shutdown.");
						break;
					} else {
						System.err.println("Something strange happened.");
					}
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
					// System.exit(-1);
				}
			}
		}
	}
	
	public static void main(String[] args) throws IOException, InterruptedException {
		// Setup server
		Server server = new Server();
		
		// Add work
		// server.addWork("abcd", "greedy", new int[] { 1, 2, 3, 4 }, 1);
		// server.addWork("cdef", "greedy", new int[] { 5, 6, 7, 8 }, 2);
		
		// Start server
		server.startServer();
		
		// Start client
		// Client client = new Client("127.0.0.1", port);
		// client.getWork();
		// client.getWork();
		// /client.getWork();
		
		while (true) {
			Thread.sleep(200000);
		}
		
		// Thread.sleep(20000000);
		
		// Stop server
		// System.out.println("Time is up.");
		// server.stopServer();
	}
}
