%% Learns the objective as discussed in the meeting on October, 1st

% given a set of features (color histograms, bag of features), learn 
% weights that assign importance to each of the words automatically. We
% do this by solving:
%
%   max_w sum_{k=1}^K f_w(A_k),
%
% where 
%
%   f_w(A) := sum_{u \in U} w_u \sqrt{ \sum_{a \in A} c_u(a) },
%
% U is the set of available words, D = {A_1, ..., A_K} is the set of 
% human summaries, c_u(a) counts the occurences of word u in image a (maybe
% using some kind of TF-IDF scaling), and w is the weight vector to be 
% tuned (non-negative).

clc;
clear all;
rand('seed', 123456);

%% settings
nrAverage = 1000;
nrImages = 100;
summaryLength = 10;

%% compute random summaries
randomSummaries = randomSummary(nrImages, summaryLength, nrAverage);

for dataset = 1:14,
    fprintf(1, '>> Processing dataset %d...\n', dataset);
    for k = 1:10,
    fprintf(1, 'k=%d\n', k);
    
    %% load summaries and compute scores
    summaryfile = sprintf('../Data/set%d/summary.txt', dataset);
    summaries = load(summaryfile);
    nrSummaries = size(summaries, 1);

    %% Compute scores
    optionsEvaluation = struct();
    optionsEvaluation.k = k;
    
    [randScore, randScores] = summaryEvaluation(randomSummaries, summaries, 'combinatorial', optionsEvaluation);
    [sumScore, scores] = summaryEvaluation(summaries, summaries, 'combinatorial', optionsEvaluation);
    
    scoresStr = sprintf('Human summaries/Random summaries (overlap): %f/%f (%f)\n', sumScore, randScore, computeGaussianOverlap(mean(scores), mean(randScores), std(scores), std(randScores)));
    fprintf(1, scoresStr);    
    
    %% Plot
    xmin = min(min(scores), min(randScores)) - 0.1;
    xmax = max(max(scores), max(randScores)) + 0.1;
    bins = linspace(xmin, xmax, 40);
    finebins = linspace(xmin, xmax, 100);
    %ymax = 2; %max(max(histc(scores, bins)), max(histc(randScores, bins)));

    clf
    hold all;
    % h11 = histfit(scores, 30);
    % set(h11(1), 'FaceColor', 'r');
    % set(h11(1), 'FaceAlpha', 0.2);
    % set(h11(2), 'Color', 'r');
    a1 = histc(scores, bins);
    a1 = a1 / (bins(2)-bins(1)) / sum(a1);
    m = mean(scores);
    s = std(scores);
    b1 = normpdf(finebins, m, s);
    
    a2 = histc(randScores, bins);
    a2 = a2 / (bins(2)-bins(1)) / sum(a2);
    m = mean(randScores);
    s = std(randScores);
    b2 = normpdf(finebins, m, s);
    
    ymax = max([max(a1), max(a2), max(b1), max(b2)])*1.1;
    
    hb = bar(bins,a1,'r');
    set(get(hb, 'Children'), 'FaceAlpha', 0.2);
    xlim([xmin, xmax]);
    ylim([0, ymax]);
    h = plot(finebins, b1, 'Linewidth', 2);
    set(h, 'Color', 'r');



%             h11 = histfit(randScores, 30);
%             set(h11(1), 'FaceColor', 'b');
%             set(h11(1), 'FaceAlpha', 0.2);
%             set(h11(2), 'Color', 'b');

    hb = bar(bins,a2,'b');
    set(get(hb, 'Children'), 'FaceAlpha', 0.2);
    xlim([xmin, xmax]);
    ylim([0, ymax]);
    h = plot(finebins, b2, 'Linewidth', 2);
    set(h, 'Color', 'b');
    xlim([xmin, xmax]);
    ylim([0, ymax]);
    scoresStr = sprintf('Human summaries/Random summaries (overlap): %f/%f (%f)\n', sumScore, randScore, computeGaussianOverlap(mean(scores), mean(randScores), std(scores), std(randScores)));
    title(sprintf('k=%d: %s\n', k, scoresStr));

    % save to file
    fnameImage = sprintf('html/images/SummaryOverlap/set%d/k_%d.png', dataset, k);
    saveImage(fnameImage, 800, 400);
    end
end
