function [ enc, encmap, err ] = encode( data, centers, method, options )
%ENCODE   Encodes data.
%   ENCODE encodes data using the provided centers. Multiple methods for 
%   encoding and options are available. "data" corresponds to the data to
%   be encoded, e.g. SIFT descriptors extracted from the image. "centers"
%   corresponds to reference data used for encoding, e.g. SIFT descriptors
%   used for vector quantization. Both "data" and "centers" are provided as
%   column vectors, i.e. every column corresponds to the descirption of a
%   feature. "method" can be either
%       'nn' ... nearest neighbour in Euclidean distance sence, i.e. every 
%                feature is encoded as the nearest neighbour in centers
%       'nnHistIntersection' ... same as nearest neighbour, but with
%                histogram intersection as distance measure (this is
%                useful when encoding color histograms)
%       TODO: Describe other methods.
%   "options" can be a numeric value that describes either the threshold of
%   an encoding algorithm or the variance when using en exp-based
%   similarity measure. Algorithms that require neither of these, simply
%   ignore this value.


%Method can be either 'nn' 
%   (nearest neighbor), 'stIP' (soft threshold, where options corresponds
%   to the threshold parameter and the threshold is applied to the inner
%   product of data/centers), 'stEUC (soft threshold, where options
%   corresponds to the threshhold parameter and threshold is applied
%   to (1 - norm(data -centers)).

%% settings
nOptArg = 4;
if nargin < nOptArg || isempty(options) || ~isfield(options, 'logSpaceScaling'),
    options.logSpaceScaling = false;
    warning('You didn''t specify if you want log-space scaling or not, so it is deactiviated.');
end

if nargin < nOptArg || isempty(options) || ~isfield(options, 'descriptors'),
    if options.logSpaceScaling,
        error('Log-space scaling is activated, but you did not provide descriptors.');
    end
end

%% run

if nargin > 3,
    th = options;
    sigma = options;
else
    th = 0.5;
    sigma =0.5;
end

if strcmp(method, 'nn'),
    % nearest neighbor encoding
    dist = vl_alldist(centers, data, 'l1');
    %dist = pdist2(centers', data', 'L1');
    
    [mindist, encmap] = min(dist, [], 1);
    weights = ones(1,length(encmap));
    if options.logSpaceScaling, % should we perform log-space scaling?
        % add logspace style weighting
        distToCenter = sqrt(sum((options.descriptors-50).^2,1));
        closenessToCenter = max(distToCenter) + 1 - distToCenter;
        wf = log(closenessToCenter);
        weights = weights .* wf;
    end
    tmp = accumarray(encmap(:), weights);
    tmp = vertcat(tmp, zeros(size(centers,2)-length(tmp),1));
    enc = tmp;
    err = min(dist, [], 1);
elseif strcmp(method, 'sparseCoding'),
    %fprintf(1, 'Performing sparse coding...\n');
    enc = zeros(size(centers,2), 1);
    idxStart = 1;
    nr = 100;
    for k = 1:ceil(size(centers,2) / nr),
        fprintf(1,':');
        idxEnd = idxStart + nr - 1;
        [~,encT] = sparseCoding(data(:,idxStart:idxEnd), centers, 1e-2);
        enc = enc + sum(encT,2);
        enc(enc < 1e-3) = 0;
        idxStart = idxStart + nr;
        sum(enc > 0)
    end
    enc(enc < 1e-3) = 0;
    enc = sum(enc, 2);
elseif strcmp(method, 'nnSoft'),
    % nearest neighbor encoding, soft weighting
    dist = vl_alldist(centers, data, 'l1');
    %dist = pdist2(centers', data', 'L1');
    
    [mindist, encmap] = min(dist, [], 1);
    if options.logSpaceScaling, % should we perform log-space scaling?
        error('Not implemented.');
    end
    tmp = accumarray(encmap(:), max(zeros(size(mindist)), options.encoderThreshold - mindist));
    tmp = vertcat(tmp, zeros(size(centers,2)-length(tmp),1));
    enc = tmp;
elseif strcmp(method, 'nnSoftSum'),
    % nearest neighbor encoding, soft weighting
    dist = vl_alldist(centers, data, 'l1');
    %dist = pdist2(centers', data', 'L1');
    
    if options.logSpaceScaling, % should we perform log-space scaling?
        error('Not implemented.');
    end
    enc = sum(max(zeros(size(dist)), options.encoderThreshold - dist), 2);
elseif strcmp(method, 'nnTh'),
    % nearest neighbor encoding, soft weighting
    dist = vl_alldist(centers, data, 'l2');
    %dist = pdist2(centers', data', 'L1');
    
    %enc = sum(max(zeros(size(dist)), options.encoderThreshold - dist), 2);
    
    enc = zeros(size(centers,2),1);
    %for k = 1:5,
        [mindist, encmap] = min(dist, [], 1);
        tmp = accumarray(encmap(:), max(zeros(size(mindist)), options.encoderThreshold - mindist));
        tmp = vertcat(tmp, zeros(size(centers,2)-length(tmp),1));
        enc = enc + tmp;
        dist(sub2ind(size(dist), encmap, 1:size(dist,2))) = inf;
    %end
    
%     tmp = vertcat(tmp, zeros(size(centers,2)-length(tmp),1));
%     enc = tmp;
%     [mindist, encmap] = min(dist, [], 1);
%     if options.logSpaceScaling, % should we perform log-space scaling?
%         error('Not implemented.');
%     end
%     tmp = accumarray(encmap(:), max(zeros(size(mindist)), 1 - mindist));
%     tmp = vertcat(tmp, zeros(size(centers,2)-length(tmp),1));
%     enc = tmp;
elseif strcmp(method, 'stIP'),
    % soft threshold with inner product
    tmp = centers'*data;
    tmp = tmp - options.encoderThreshold;
    tmp = max(tmp, zeros(size(tmp)));
    enc = sum(tmp, 2);
elseif strcmp(method, 'exp'),
    % exp
    enc = zeros(size(centers,2), 1);
    for c = 1:size(centers, 2),
        for d = 1:size(data,2),
            enc(c) = enc(c) + exp(-options.encoderThreshold * norm(centers(:,c) - data(:,d)));
        end
    end
elseif strcmp(method, 'nnHistIntersection'),
    % nearest neighbor encoding with histogram intersection as distance
    % measure
    dist = zeros(size(centers, 2), size(data,2));
    for l = 1:size(data,2),
        a = min(centers, repmat(data(:,l), 1, size(centers,2)));
        dist(:,l) = 1 - sum(a,1);
    end
    [~, binsa] = min(dist, [], 1);
    encmap = binsa;
    tmp = accumarray(binsa(:), ones(length(binsa), 1));
    tmp = vertcat(tmp, zeros(size(centers,2)-length(tmp),1));
    enc = tmp;
elseif strcmp(method, 'nnHistIntersectionSoft'),
    % warning('nearest neighbour with histogram intersection distance -- experimental.');
    dist = zeros(size(centers, 2), size(data,2));
    sigma = 0.1;

    enc = zeros(size(centers, 2), 1);
    for l = 1:size(data,2),
        a = min(centers, repmat(data(:,l), 1, size(centers,2)));
        dist(:,l) = 1 - sum(a,1);
        tmp = exp(-dist(:,l) / (2*sigma^2));
        enc = enc + tmp / sum(tmp);
    end
elseif strcmp(method, 'nnEMD'),
    % warning('nearest neighbour with EMD distance -- experimental.');
    dist = zeros(size(centers, 2), size(data,2));
    
    if ~isfield(options, 'bins'),
        error('Bins must be provided.');
    end
    
    if ~isfield(options, 'quantizerType') || ~strcmp(options.quantizerType, 'SQ'),
        error('Quantizer-Type not specified or invalid (only SQ supported)');
    end

    dist = zeros(size(centers,2), size(data,2));
    binStart = 1;
    for k = 1:3, % cycle through bins
        binEnd = binStart + options.bins(k) - 1;
        for l = 1:size(data,2),
            % dist(:,l) = pdist2(centers', data(:,l)', 'emd') / size(centers, 1);
            dist(:,l) = dist(:,l) + pdist2((centers(binStart:binEnd,:))', data(binStart:binEnd,l)', 'emd') / size(centers, 1);
        end
        binStart = binStart + options.bins(k);
    end
    
    enc = sum(max(zeros(size(dist)), options.encoderThreshold - dist), 2);
    
    
%     [~, binsa] = min(dist, [], 1);
%     weights = ones(1,length(binsa));
%     if options.logSpaceScaling, % should we perform log-space scaling?
%         % add logspace style weighting
%         error('Needs to be rechecked.');
%         maxDist = norm([50,50]);
%         distToCenter = sqrt(sum((options.descriptors-50).^2,1));
%         closenessToCenter = maxDist + 1.1 - distToCenter;
%         wf = log(closenessToCenter);
%         weights = weights .* wf;
%     end
%     
%     tmp = accumarray(binsa(:), weights);
%     tmp = vertcat(tmp, zeros(size(centers,2)-length(tmp),1));
%     enc = tmp;
elseif strcmp(method, 'nnEMDSoft'),
    % warning('nearest neighbour with EMD distance -- experimental.');
    dist = zeros(size(centers, 2), size(data,2));
    
    weights = ones(1,size(data,2));
    if options.logSpaceScaling, % should we perform log-space scaling?
        % add logspace style weighting
        distToCenter = sqrt(sum((options.descriptors-50).^2,1));
        closenessToCenter = max(distToCenter) + 1.1 - distToCenter;
        wf = log(closenessToCenter);
        weights = weights .* wf;
    end

    enc = zeros(size(centers, 2), 1);
    for l = 1:size(data,2),
        dist(:,l) = pdist2(centers', data(:,l)', 'emd') / size(centers, 1) / weights(l);
       
        tmp = exp(-dist(:,l) / (2*options.sigma^2));
        stmp = sum(tmp);
        if stmp < 1e-6,
            stmp = 1;
        end
        enc = enc + tmp / stmp;
    end
else
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% NO INTENSIVE CODE CHECK PERFORMED BEYOND THIS LINE %%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 warning('Using an experimental encoding variant.');
 if strcmp(method, 'nnHistIntersectionSoftSum'),
    % warning('nearest neighbour with histogram intersection distance -- experimental.');
    dist = zeros(size(centers, 2), size(data,2));

    %for i = 1:size(centers,2),
    for l = 1:size(data,2),
        a = min(centers, repmat(data(:,l), 1, size(centers,2)));
        dist(:,l) = 1 - sum(a,1);
    end
    %end
    
    dist = exp(-dist/(2*sigma^2));
    for k = 1:size(dist,2),
        nrm = norm(dist(:,k), 1);
        if nrm > 0,
            % dist(:,k) = max(dist(:,k)) * dist(:,k) / nrm; % activate by a total of 1
        end
    end

    enc = sum(dist,2);
    elseif strcmp(method, 'nnHistIntersectionSoft'),
        % warning('nearest neighbour with histogram intersection distance -- experimental.');
        dist = zeros(size(centers, 2), size(data,2));
    %     for i = 1:size(centers,2),
    %         for l = 1:size(data,2),
    %             dist(i,l) = 1 - sum(min(centers(:,i), data(:,l)));
    %         end
    %     end

        %for i = 1:size(centers,2),
        for l = 1:size(data,2),
            a = min(centers, repmat(data(:,l), 1, size(centers,2)));
            dist(:,l) = 1 - sum(a,1);
        end
        %end

        [vals, binsa] = min(dist, [], 1);
        encmap = binsa;
        tmp = accumarray(binsa(:), 1-vals(:));
        tmp = vertcat(tmp, zeros(size(centers,2)-length(tmp),1));
        enc = tmp;
    elseif strcmp(method, 'nnSoft'),
        % make soft assignment to clusters using a exponential simiarity
        % measure
        sigma = 0.5;
        dist = vl_alldist(single(centers), single(data));
        [~, binsa] = min(dist, [], 1);

        idcs = sub2ind(size(dist), binsa, 1:length(binsa));
        tmp = accumarray(binsa(:), exp(-dist(idcs)/(2*sigma^2)));
        tmp = vertcat(tmp, zeros(size(centers,2)-length(tmp),1));
        enc = tmp;
    elseif strcmp(method, 'nnSoftSum'),
        % make soft assignment to clusters using a exponential simiarity
        % measure
        dist = vl_alldist(single(centers), single(data));

        dist = exp(-dist/(2*sigma^2));
        for k = 1:size(dist,2),
            nrm = norm(dist(:,k), 1);
            if nrm > 0,
                % dist(:,k) = max(dist(:,k)) * dist(:,k) / nrm; % activate by a total of 1
            end
        end

        enc = sum(dist, 2);
    elseif strcmp(method, 'nnRobust'),
        % make soft assignment and assign only if there are no other strong
        % competitor matches.
        sigma = 0.2;
        dist = vl_alldist(single(centers), single(data));
        dist = exp(-dist/(2*sigma^2));
        enc = zeros(size(dist,1), 1);
        for l = 1:size(dist, 2),
            [a,b] = sort(dist(:,l), 'ascend');
            if a(1)/a(2) < 0.8, % "clear" match
                enc(b(1)) = enc(b(1)) + 1;
            end
            %enc = enc + dist(:,l) / norm(dist(:,l), 1);
        end

        %[~, binsa] = min(dist, [], 1);

        %idcs = sub2ind(size(dist), binsa, 1:length(binsa));
        %tmp = accumarray(binsa(:), exp(-dist(idcs)/(2*sigma^2)));
        %tmp = vertcat(tmp, zeros(size(centers,2)-length(tmp),1));
        %enc = tmp;
    elseif strcmp(method, 'stIP'),
        % soft threshold with inner product
        % every vector can cause at most (1-th) activation -> divide by number
        % of nonzero activations
        tmp = centers'*data;
        tmp = tmp - th;
        tmp = max(tmp, zeros(size(tmp)));
        cnt = 1./sum(tmp > 0, 1);
        cnt(isinf(cnt)) = 1;
        enc = sum(tmp .* repmat(cnt, size(tmp,1), 1), 2);
    elseif strcmp(method, 'stIPsum'),
        % sum all activations caused by a center (maybe not too meaningful)
        tmp = centers'*data;
        tmp = tmp - th;
        tmp = max(tmp, zeros(size(tmp)));
        enc = sum(tmp, 2);
    elseif strcmp(method, 'stIPmax'),
        % assign to best matching center
        tmp = centers'*data; % format: centers x datapoints
        tmp = tmp - th;
        tmp = max(tmp, zeros(size(tmp)));
        [a,b] = max(tmp, [], 1);
        enc = accumarray(b(:), a(:)');
        enc = vertcat(enc, zeros(size(centers,2) - length(enc),1));
    elseif strcmp(method, 'sc'),
        % sparse coding
        error('Not implemented yet.');
    else
        error('Not implemented yet.');
 end
end


end

