function [] = experiment_losses()

% Learn weights in large margin framework

clc

%% settings
overlap = 0.05;

%% setup pathes
javaaddpath(fullfile(pwd, 'Java/SubmodularFunctions/bin'));
addpath('external/sfo');
addpath('SubmodularComponents');
addpath('Java/MatlabInterface');

if vl_isoctave(),
    page_output_immediately(1);
    page_screen_output(0);
    pkg load java
end

rand('seed', 12345);

meanOverlap = 0;
datasetRange = 1:14;
for ds = datasetRange,
    fprintf(1, '>> DS %2d <<<<<<\n', ds);
    
    % load summaries and generate random summaries
    summaries = pruneSummaries(ds, get_pruning_info(ds, overlap));
    nrSummaries = size(summaries, 1);
    nrRandomSummaries = nrSummaries;
    randomSummaries = randomSummary(100, 10, nrRandomSummaries);
    
    % load bag-of-words
    bowLoss = get_BOW_evaluation(ds);
    
    % setup evaluation function
    optionsEvaluationLoss = struct();
    optionsEvaluationLoss.bow = bowLoss;
    optionsEvaluationLoss.summariesAsBow = computeBOWSummary(summaries, optionsEvaluationLoss.bow);
    [~,humanScores] = summaryEvaluation(summaries, summaries, 'rouge', optionsEvaluationLoss);
    [~,randomScores] = summaryEvaluation(randomSummaries, summaries, 'rouge', optionsEvaluationLoss);
    
    % overlap of one-minus-rouge
    randomLosses = 1-randomScores;
    humanLosses = 1-humanScores;
    overlapEmp = computeOverlap(humanLosses, randomLosses, 'loss');
    allLosses = [humanLosses; randomLosses];
    allSummaries = [summaries; randomSummaries];
    fprintf(1, 'Overlap One-Minus-Rouge: %2.2f\n', overlapEmp);
    meanHumanLossOrig = mean(humanLosses)
    meanRandomLossOrig = mean(randomLosses)
    
    % for development only, prune words
    nrWordsT = size(bowLoss, 1);
    t = randperm(nrWordsT);
    bowLoss = bowLoss(t(1:end), :);
    nrWords = size(bowLoss,1);
    

    sumAsBowHuman = computeBOWSummary(summaries, bowLoss);
    sumAsBowRandom = computeBOWSummary(randomSummaries, bowLoss);
    
    fprintf(1, 'Mean Human: %f, mean random: %f\n', mean(humanLosses), mean(randomLosses));
    
    weights = zeros(2*nrWords, 1);
    
    tweights = [log(0.5); 0*ones(size(weights))];
    
    sl = SurrogateLossJ(MatrixJ(bowLoss), MatrixJ(optionsEvaluationLoss.summariesAsBow));

    initTh = -rand(size(weights));
    initialWeights = [log(mean(allLosses)); initTh];
    fct = @(w) obj4(sl, w, humanLosses, randomLosses, sumAsBowHuman, sumAsBowRandom, nrWords, meanHumanLossOrig, meanRandomLossOrig);
    fct(initialWeights);
    
    % perform gradient descent
%     weights = initialWeights;
%     delta = 1e-2;
%     stepsize = 1e-1;
%     for i = 1:50,
%         obj = fct(weights);
%         
%         % compute gradient
%         grad = zeros(size(weights));
%         for k = 1:length(grad),
%             w = weights;
%             w(k) = w(k) + delta;
%             
%             obj2 = fct(w);
%             
%             grad(k) = (obj2 - obj) / delta;
%         end
%         
%         weights = weights - stepsize * grad;
%         
%         fprintf(1, '[%2d] Objective=%f\n', i, obj);
%     end
%     weightsNew = weights;

    options = optimset('display', 'iter');
    options = optimset(options, 'Algorithm', 'interior-point');
    options = optimset(options, 'Hessian', {'lbfgs', 3});
    options = optimset(options, 'MaxFunEvals', 50000);
    weightsNew = fmincon(@(w) fct(w),initialWeights,[],[],[],[],-10*ones(1*length(weights)+1,1),inf*ones(1*length(weights)+1,1),[],options);
    
    %% final evaluation
    thMin = 0.5*mean(sumAsBowHuman, 1)';
    sl.setOffset(exp(weightsNew(1)));
    sl.setThresholds(thMin+exp(weightsNew(2:(1+nrWords))));
    sl.setWeights(1e-1+exp(weightsNew((2+nrWords):end)));
    humanLosses = sl.evalSumBow(computeBOWSummary(summaries, bowLoss));
    randomLosses = sl.evalSumBow(computeBOWSummary(randomSummaries, bowLoss));
    
    overlapEmp = computeOverlap(humanLosses, randomLosses, 'loss');
    fprintf(1, 'Overlap Surrogate Rouge Loss: %2.2f\n', overlapEmp);
    fprintf(1, 'Mean Human: %f, mean random: %f\n', mean(humanLosses), mean(randomLosses));
    
    % store weighs
    allWeights = weightsNew(2:end);
    allOffset =  weightsNew(1);
    
    meanOverlap = meanOverlap + overlapEmp;
    
    
    fprintf(1, '\n');
    save(sprintf('weights_%d.mat', ds), 'allOffset', 'allWeights', 'weightsNew');
    
    %% plot
    clf
    bins = linspace(min(min(randomLosses), min(humanLosses))-0.1, max(max(randomLosses), max(humanLosses))+0.1, 100);
    bar(bins, hist(randomLosses, bins), 'r');
    hold on
    bar(bins, hist(humanLosses, bins), 'g');
    
    figure
    hist(exp(allWeights));
    
    %%
    keyboard
end

fprintf(1, 'Mean overlap over all datasets: %f\n', meanOverlap / length(datasetRange));    

end

%% compute margin score
function [val] = obj4(sl, w, humanSummaries, randomSummaries, summariesAsBowHuman, summariesAsBowRandom, nrWords, meanHumanLossOrig, meanRandomLossOrig)
    lambda = 1e-3;
    threshold = -0.1;

    % update parameters
    sl.setOffset(exp(w(1)));
    thMin = 0.5*mean(summariesAsBowHuman, 1)';
    sl.setThresholds(thMin+exp(w(2:(1+nrWords))));
    sl.setWeights(1e-1+exp(w((2+nrWords):end)));
    
    % compute losses of human and random summaries
    lossesHuman = sl.evalSumBow(summariesAsBowHuman);
    lossesRandom = sl.evalSumBow(summariesAsBowRandom);
    
    % compute objective
    mH = mean(lossesHuman);
    mR = mean(lossesRandom);
    % val = -(mH - mR)^2 / (var(lossesHuman) + var(lossesRandom)) + lambda*norm(w);
    val = -(mH - mR)^2 / (var(lossesHuman) + var(lossesRandom)) + lambda*norm(w);
    %val = sum(max(threshold, lossesHuman - mR)) + lambda * mean(abs(w)) + lambda * mean(abs(exp(w))) + lambda*abs(mH - meanHumanLossOrig) + lambda*abs(mR - meanRandomLossOrig);
end

function [finalVal, grad] = obj3(sl, w, humanSummaries, randomSummaries, summariesAsBowHuman, summariesAsBowRandom, nrWords)
    sl.setOffset(exp(w(1)));
    sl.setThresholds(exp(w(2:(1+nrWords))));
    %t = exp(w((2+nrWords):end));
    %sl.setWeights(t);
    
    val = sl.evalSumBow(summariesAsBow);
    
    finalVal = norm(allLosses - val, 1);
end

function [finalVal, grad] = obj2(sl, w, allLosses, allSummaries, nrWords)
    sl.setOffset(exp(w(1)));
    sl.setThresholds(exp(w(2:(1+nrWords))));
    t = exp(w((2+nrWords):end));
    sl.setWeights(t);
    
    val = zeros(size(allSummaries,1),1);
    for s = 1:size(allSummaries, 1),
        val(s) = sl.eval(allSummaries(s,:));
    end
    
    finalVal = norm(allLosses - val, 1);
    
    if nargout == 2,
        % compute gradient
        grad = zeros(length(w),1);
        sgn = -sign(allLosses - val);
        
        for s = 1:size(allSummaries, 1),
            t = java2matArray(sl.computeGradient(allSummaries(s,:)));
            grad = grad(:) + sgn(s)*t(:);
        end
        
        grad = grad .* exp(w)
    end
end

function [valFinal, grad] = obj(w, allLosses, sumAsBow, bow, summariesAsBow)
    nrSamples = length(allLosses);
    nrSummaries = size(summariesAsBow, 1);
    nrWords = size(summariesAsBow, 2);

    offset = exp(w(1));
    expw = exp(w(1:(1+nrWords)));
    weights = exp(w((2+nrWords):end));
    %weights = ones(nrWords,1);
    reg = 0;

    % compute thresholds; size: nrSummaries x nrWords
    normalization = nrSummaries*sum(expw(2:end) .* weights);
    thresholds = repmat(expw(2:end)', nrSummaries, 1) - summariesAsBow;
    thresholds(thresholds < 0) = 0;
    %weightsMatrix = repmat(weights(:)', nrSummaries, 1);
    % normalization = sum(sum(thresholds));

    val = zeros(nrSamples, 1);
    nom = zeros(nrSamples, 1);
    denom = zeros(nrSamples, 1);

    for i = 1:nrSamples,
        asBow = sumAsBow(i,:);
        % nom(i) = sum(weights(:)' .* sum(min(thresholds, repmat(asBow, nrSummaries, 1)),1));
        nom(i) = sum(weights(:)' .* sum(min(thresholds, repmat(asBow, nrSummaries, 1)),1));
        denom(i) = normalization;
        val(i) = offset + nom(i) / denom(i);
    end

    partVal = norm(allLosses - val,1);
    % valFinal = partVal^2 + 0.5+reg*norm(w)^2;
    valFinal = partVal + 0.5*reg*norm(w(2:end))^2;

    %% gradient
    if nargout == 2,
        grad = zeros(size(w));
        for i = 1:nrSamples,
            sgn = sign(allLosses(i) - val(i));
            grad(1) = grad(1) + sgn*(-offset);
        end

        for i = 1:nrSamples,
            asBow = sumAsBow(i,:); % sum(bow(:,S(i,:)), 2)';
            sgn = -sign(allLosses(i) - val(i));

            for j = 1:nrWords,

                grd = -nrSummaries * nom(i) * weights(j) * expw(1+j) / (denom(i)^2);
                grad(1+j) = grad(1+j) + sgn*(grd);

                for k = 1:nrSummaries,
                    if asBow(j) > thresholds(k,j),
                        grd = weights(j)*expw(1+j) / denom(i);
                    else
                        grd = 0;
                    end
                    %grd
                    grad(1+j) = grad(1+j) + sgn*(grd);
                end
            end
        end
        
        % derivative with respect to weights
        idxOffset = 1+nrWords;
        for i = 1:nrSamples,
            asBow = sumAsBow(i,:); % sum(bow(:,S(i,:)), 2)';
            sgn = -sign(allLosses(i) - val(i));

            for j = 1:nrWords,

                grd = -nrSummaries * nom(i) * weights(j) * expw(1+j) / (denom(i)^2);
                grad(idxOffset+j) = grad(idxOffset+j) + sgn*(grd);

                for k = 1:nrSummaries,
                    if asBow(j) > thresholds(k,j),
                        tval = thresholds(k,j);
                    else
                        tval = asBow(j);
                    end
                    grd = weights(j)*tval/ denom(i);
                    %grd
                    grad(idxOffset+j) = grad(idxOffset+j) + sgn*(grd);
                end
            end
        end
        grad(2:end) = grad(2:end) + reg*w(2:end);
    end
    
    
    %% gradient
%     allsign = sign(allLosses - val);
%     if nargout == 2,
%         grad = zeros(size(w));
%         grad(1) = grad(1) - sum(allsign.*offset);
% 
%         for i = 1:nrSamples,
%             asBow = sumAsBow(i,:);
%             sgn = -allsign(i);
% 
%             mask = repmat(asBow(:)', nrSummaries, 1) >= thresholds;
%             grad(2:end) = grad(2:end) -  sgn.*nrSummaries * nom(i) * expw(2:end) / (denom(i)^2);
%             grad(2:end) = grad(2:end) + sum(mask,1)' .* expw(2:end) / denom(i);
%         end
% 
%         %grad = grad * partVal;
%         grad = grad + reg*w;
%     end
end
