#include "Helper.h"
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <algorithm>
#include <stdio.h>
#include <time.h>

using namespace std;

#include "Config/ConfigFile.h"

#include "Functions/constantFunction.h"
#include "Functions/clusteredDiversity.h"
#include "Functions/clusteredFacilityLocation.h"
#include "Functions/rougeFunctions.h"
#include "Functions/fidelityFunctions.h"
#include "Functions/cosineSimilarity.h"
#include "Functions/compressedWordCount.h"
#include "Functions/hingeLoss.h"
#include "Functions/surrogateRougeFunctions.h"

const std::string Helper::preDir = "../../../precomputed/";

std::list<ComponentDescriptor> Helper::readComponentsFile(std::string filename) {
    std::cout << "Trying to load components description..." << std::endl;

    std::ifstream componentsFile;
    componentsFile.open(filename.c_str());

    if (!componentsFile.is_open()) {
        throw std::runtime_error("Could not open file: " + filename);
    }

    std::string line;
    std::string type;
    std::string resource;
    std::list<std::string> resources;
    std::list<ComponentDescriptor> componentDescriptors;
    while (getline(componentsFile, line)) {
        // trim line
        line = Helper::trim(line);

        if (line.compare(0, 9, "Component") == 0) {
            // do we have to add a previous read component?
            if (type.length() != 0) {
                componentDescriptors.push_back(ComponentDescriptor(type, resources));
                type.clear();
                resources.clear();
            }

            // line starts with component
            type = line.substr(11, std::string::npos);
        } else if (line.compare(0, 8, "Resource") == 0) {
            // line starts with resource => add to list of resources
            resource = line.substr(10, std::string::npos);
            resources.push_back(resource);
        }
    }
    
    // do we have to add a previous read component?
    if (type.length() != 0) {
	componentDescriptors.push_back(ComponentDescriptor(type, resources));
	type.clear();
	resources.clear();
    } 

    componentsFile.close();

    return componentDescriptors;
}

std::vector<std::vector<double> > Helper::readMatrixBinary(std::string filename, int x, int y) {
	std::vector<std::vector<double> > matrix = readMatrixBinary(filename, x);
	
	if (matrix.size() != y) {
		throw std::runtime_error("Could not read desired number of rows.");
	}
	
	return matrix;
}

std::vector<std::vector<double> > Helper::readMatrixBinary(std::string filename, int x) {
	std::vector<std::vector<double> > matrix;
	
	int unitsize = sizeof(float);
	
	ifstream iFile;	
	FILE* fp;
	if (!(fp=fopen(filename.c_str(),"rb"))){
		throw std::runtime_error("ERROR: cannot open file.\n");
	}
	
	float cache[2*x];
	int i = 0;
	while( fread(cache, unitsize, x, fp) == x) {
		std::vector<double> tvector(x);
		for (int j = 0; j < x; j++){
			tvector[j] = (double) cache[j];			
		}
		matrix.push_back(tvector);
		i++;
	}
	fclose(fp);
	
	return matrix;
}

std::vector<std::vector<int> > Helper::readMatrixBinaryInteger(std::string filename, int x)
{
	std::vector<std::vector<int> > matrix;
	
	int unitsize = sizeof(int);
	
	ifstream iFile;	
	FILE* fp;
	if (!(fp=fopen(filename.c_str(),"rb"))){
		throw std::runtime_error("ERROR: cannot open file.\n");
	}
	
	int cache[2*x];
	int i = 0;
	while( fread(cache, unitsize, x, fp) == x) {
		std::vector<int> tvector(x);
		for (int j = 0; j < x; j++){
			tvector[j] = cache[j];			
		}
		matrix.push_back(tvector);
		i++;
	}
	fclose(fp);
	
	return matrix;
}

void Helper::dumpComponentDescriptorList(std::list<ComponentDescriptor> componentDescriptorList) {
    std::list<ComponentDescriptor>::iterator it;
    int idx = 0;
    for (it = componentDescriptorList.begin(); it != componentDescriptorList.end(); it++) {
        std::cout << "[" << idx << "] Component: " << it->getType() << std::endl;

        std::list<std::string> resources = it->getResources();
        std::list<std::string>::iterator itr;
        for (itr = resources.begin(); itr != resources.end(); itr++) {
            std::cout << "     Resource: " << *itr << std::endl;
        }

        idx++;
    }
}

facilityLocation* Helper::createFacilityLocationComponent(ComponentDescriptor descriptor, std::unordered_set<int> &groundSet, int ds) {
    std::string type = descriptor.getType();
    std::list<std::string> resources = descriptor.getResources();
    
    //std::cout << "Constructing facility location component." << std::endl;
    //std::cout << "There are " << resources.size() << " resources." << std::endl;
    if (resources.size() != 1) {
        throw std::runtime_error("One resource must be supplied to generate a facility location component.");
    }

    char* similarityFileStr = Helper::getFilenameForResource(resources.front(), ds);
    facilityLocation* fL = new facilityLocation(groundSet, similarityFileStr);

    return fL;
}

modularFunctions* Helper::createSumCoverageComponent(ComponentDescriptor descriptor, std::unordered_set<int> &groundSet, int ds) {
    std::string type = descriptor.getType();
    std::list<std::string> resources = descriptor.getResources();
    
    //std::cout << "Constructing sum-coverage component." << std::endl;
    //std::cout << "There are " << resources.size() << " resources." << std::endl;
    if (resources.size() != 1) {
        throw std::runtime_error("One resource must be supplied to generate a sum-coverage component.");
    }
    
    // compute singletons (this could/should be shifted to a precomputation step)
    int n = groundSet.size();
    std::vector<std::vector<double> > matrix = Helper::readMatrixBinary(Helper::getFilenameForResource(resources.front(), ds), n, n);
    std::vector<double> *scores = new std::vector<double>(n);
    double sum = 0.0;
    for (int i = 0; i < n; i++) {
    	    (*scores)[i] = 0.0;
    	    for (int j = 0; j < n; j++) {
    	    	    (*scores)[i] += matrix[j][i];
    	    }
    	    sum += (*scores)[i];
    }
    for (int i = 0; i < n; i++) {
    	    (*scores)[i] /= sum;
    }

    modularFunctions* mod = new modularFunctions(groundSet, *scores, 0.0); 
    
    // old version
    //char* similarityFileStr = Helper::getFilenameForResource(resources.front(), ds);
    //modularFunctions* mod = new modularFunctions(groundSet, similarityFileStr);

    return mod;
}

graphCutFunctions* Helper::createCutCoverageComponent(ComponentDescriptor descriptor, std::unordered_set<int> &groundSet, int ds) {
    std::string type = descriptor.getType();
    std::list<std::string> resources = descriptor.getResources();
    
    //std::cout << "Constructing cut-coverage component." << std::endl;
    //std::cout << "There are " << resources.size() << " resources." << std::endl;
    if (resources.size() != 1) {
        throw std::runtime_error("One resource must be supplied to generate a cut-coverage component.");
    }
    
    int n = groundSet.size();
    std::vector<std::vector<double> > matrix = Helper::readMatrixBinary(Helper::getFilenameForResource(resources.front(), ds), n, n);
    double sum = 0.0;
    for (int i = 0; i < n; i++) {
    	    for (int j = 0; j < n; j++) {
    	    	    sum += matrix[j][i];
    	    }
    }
    std::vector<std::vector<float> > *kernel = new std::vector<std::vector<float> >(n);
    for (int i = 0; i < n; i++) {
    	    (*kernel)[i].resize(n);
    	    for (int j = 0; j < n; j++) {
    	    	    (*kernel)[i][j] = (float) (matrix[i][j] / sum);
    	    }
    }
    
    // TODO: How to normalize this to zero/one properly?

    graphCutFunctions* graph = new graphCutFunctions(groundSet, *kernel, 1.0);
    //char* similarityFileStr = Helper::getFilenameForResource(resources.front(), ds);
    //graphCutFunctions* graph = new graphCutFunctions(groundSet, similarityFileStr, 1.0);

    return graph;
}

submodularFunctions* Helper::createCosineSimilarityComponent(ComponentDescriptor descriptor, std::unordered_set<int> &groundSet, int ds) {
    std::string type = descriptor.getType();
    std::list<std::string> resources = descriptor.getResources();
    
    //std::cout << "Constructing cut-coverage component." << std::endl;
    //std::cout << "There are " << resources.size() << " resources." << std::endl;
    if (resources.size() != 2) {
        throw std::runtime_error("Two resources must be supplied to generate a cosine-similarity component.");
    }
    
    std::list<std::string>::iterator it = resources.begin();
    it++;
    double threshold = Helper::getValueForResource(*it);
    
    int n = groundSet.size();
    std::vector<std::vector<double> > matrix = Helper::readMatrixBinary(Helper::getFilenameForResource(resources.front(), ds), n, n);
    std::vector<double> normalization(n);
    for (int i = 0; i < n; i++) {
    	    normalization[i] = 0.0;
    	    for (int j = 0; j < n; j++) {
    	    	    normalization[i] += matrix[i][j];
    	    }
    	    normalization[i] *= threshold;
    }
    std::vector<std::vector<double> > kernel = std::vector<std::vector<double> >(n);
    for (int i = 0; i < n; i++) {
        kernel[i].resize(n);
    }
    for (int i = 0; i < n; i++) {
    	    for (int j = 0; j < n; j++) {
    	    	    kernel[i][j] = (matrix[i][j] / (normalization[i]));
    	    }
    }
    
    /*// setup features
    std::vector<double> featureWeights(n);
    std::vector<struct Feature> features;
    int *featureIndices = new int[n];
    for (int i = 0; i < n; i++) {
    	    featureWeights[i] = 1.0;
    	    featureIndices[i] = i;
    	    
    	    float* featureVec = new float[n];
    	    for (int j = 0; j < n; j++) {
    	    	    featureVec[j] = kernel[i][j];
    	    }
    	    
    	    struct Feature feat;
    	    feat.index = i;
    	    feat.num_uniq_wrds = n;
    	    feat.featureIndex = featureIndices;
    	    feat.featureVec = featureVec;
    	    
    	    features.push_back(feat);
    }*/
    
    // TODO: How to normalize this to zero/one properly
    //featureBasedFunctions *feat = new featureBasedFunctions(groundSet, 4, features, featureWeights, new concaveFunctionMin());
    cosineSimilarity *cos = new cosineSimilarity(groundSet, kernel);
    
    double scaling = 100.0;
    scaleSubmodularFunctions* fct = new scaleSubmodularFunctions(groundSet, cos, 1.0/scaling);

    return fct;
}

submodularFunctions* Helper::createCompressedWordCountComponent(ComponentDescriptor descriptor, std::unordered_set<int> &groundSet, int ds) {
    std::string type = descriptor.getType();
    std::list<std::string> resources = descriptor.getResources();
    
    //std::cout << "Constructing compressed word count component." << std::endl;
    //std::cout << "There are " << resources.size() << " resources." << std::endl;
    if (resources.size() != 1) {
        throw std::runtime_error("One resource must be supplied to generate a compressed word count component.");
    }

    //char* featureFileStr = Helper::getFilenameForResource(resources.front(), ds);
    //featureBasedFunctions* feat = new featureBasedFunctions(groundSet, 1, featureFileStr);
    
    int n = groundSet.size();
    std::vector<std::vector<double> > matrix = Helper::readMatrixBinary(Helper::getFilenameForResource(resources.front(), ds), n);
    int nfeat = matrix.size();
    //std::cout << "Number of features: " << nfeat << std::endl;
    
    // setup features
    /*std::vector<double> featureWeights(nfeat);
    for (int i = 0; i < nfeat; i++) {
    	    featureWeights[i] = 1.0;
    }
    std::vector<struct Feature> features;
    for (int i = 0; i < n; i++) {    
    	    // count non-zero elements
    	    int nrNonZero = 0;
    	    for (int j = 0; j < nfeat; j++) {
    	    	    if (matrix[j][i] > 1e-12)
    	    	    	    nrNonZero++;
    	    }
    	    
    	    // setup feature
    	    float* featureVec = new float[nrNonZero];
    	    int *featureIndices = new int[nrNonZero];
    	    int fidx = 0;
    	    for (int j = 0; j < nfeat; j++) {
    	    	    if (matrix[j][i] > 1e-12) {
    	    	    	    featureVec[fidx] = matrix[j][i];
    	    	    	    featureIndices[fidx] = j;
    	    	    	    fidx++;
    	    	    }
    	    }
    	    
    	    struct Feature feat;
    	    feat.index = i;
    	    feat.num_uniq_wrds = nrNonZero;
    	    feat.featureIndex = featureIndices;
    	    feat.featureVec = featureVec;
    	    
    	    features.push_back(feat);
    } */
    
    // TODO: How to normalize this to zero/one properly
    // featureBasedFunctions *feat = new featureBasedFunctions(groundSet, 1, features, featureWeights, new concaveFunctionSquareRoot);
    compressedWordCount *feat = new compressedWordCount(groundSet, matrix);
    
    // determine scaling
    double scaling = feat->eval(groundSet);
    //std::cout << "Scaling: " << scaling << std::endl;
    scaleSubmodularFunctions* fct = new scaleSubmodularFunctions(groundSet, feat, 1.0/scaling);

    return fct;
}

submodularFunctions* Helper::createThresholdedFeatureCountsComponent(ComponentDescriptor descriptor, std::unordered_set<int> &groundSet, int ds) {
    std::string type = descriptor.getType();
    std::list<std::string> resources = descriptor.getResources();
    
    //std::cout << "Constructing cut-coverage component." << std::endl;
    //std::cout << "There are " << resources.size() << " resources." << std::endl;
    if (resources.size() != 2) {
        throw std::runtime_error("Two resources must be supplied to generate a cosine-similarity component.");
    }
    
    std::list<std::string>::iterator it = resources.begin();
    it++;
    double threshold = Helper::getValueForResource(*it);
    
    int n = groundSet.size();
    std::vector<std::vector<double> > matrix = Helper::readMatrixBinary(Helper::getFilenameForResource(resources.front(), ds), n);
    int nfeat = matrix.size();
    //std::cout << "Number of features: " << nfeat << std::endl;
    std::vector<double> normalization(nfeat);
    for (int i = 0; i < nfeat; i++) {
    	    normalization[i] = 0.0;
    	    for (int j = 0; j < n; j++) {
    	    	    normalization[i] += matrix[i][j];
    	    }
    	    normalization[i] *= threshold;
    	    //if (normalization[i] < 1e-12)
    	    //	    normalization[i] = 1e-12;
    	    //std::cout << "Norm: " << normalization[i] << std::endl;
    }
    std::vector<std::vector<double> > kernel = std::vector<std::vector<double> >(nfeat);
    for (int i = 0; i < nfeat; i++) {
        kernel[i].resize(n);
    }
    for (int i = 0; i < nfeat; i++) {
    	    for (int j = 0; j < n; j++) {
    	    	    kernel[i][j] = (matrix[i][j] / (normalization[i]));
    	    }
    }
    
    // setup features
    /*std::vector<double> featureWeights(nfeat);
    for (int i = 0; i < nfeat; i++) {
    	    featureWeights[i] = 1.0;
    }
    std::vector<struct Feature> features(n);
    for (int i = 0; i < n; i++) {    	        	    
    	    // count non-zero elements
    	    int nrNonZero = 0;
    	    for (int j = 0; j < nfeat; j++) {
    	    	    if (kernel[j][i] > 1e-12)
    	    	    	    nrNonZero++;
    	    }
    	    
    	    // setup feature
    	    float* featureVec = new float[nrNonZero];
    	    int *featureIndices = new int[nrNonZero];
    	    int fidx = 0;
    	    for (int j = 0; j < nfeat; j++) {
    	    	    if (kernel[j][i] > 1e-12) {
    	    	    	    featureVec[fidx] = kernel[j][i];
    	    	    	    featureIndices[fidx] = j;
    	    	    	    fidx++;
    	    	    }
    	    }
    	    
    	    struct Feature feat;
    	    feat.index = i;
    	    feat.num_uniq_wrds = nrNonZero;
    	    feat.featureIndex = featureIndices;
    	    feat.featureVec = featureVec;
    	    
    	    features[i] = feat;
    } */
    
    // TODO: The implementation of min allows to deal with NaNs => solve this properly
    // TODO: How to normalize this to zero/one properly
    //featureBasedFunctions *feat = new featureBasedFunctions(groundSet, 4, features, featureWeights, new concaveFunctionMin());
    cosineSimilarity *feat = new cosineSimilarity(groundSet, kernel);
    
    double scaling = ((double) nfeat);
    scaleSubmodularFunctions* fct = new scaleSubmodularFunctions(groundSet, feat, 1.0/scaling);

    return fct;
}

submodularFunctions* Helper::createFidelityComponent(ComponentDescriptor descriptor, std::unordered_set<int> &groundSet, int ds) {
    std::string type = descriptor.getType();
    std::list<std::string> resources = descriptor.getResources();
    
    int n = groundSet.size();
    
    //std::cout << "Constructing fidelity component." << std::endl;
    //std::cout << "There are " << resources.size() << " resources." << std::endl;
    if (resources.size() != 2) {
        throw std::runtime_error("One resource must be supplied to generate a fidelity component.");
    }
    
    std::list<std::string>::iterator it = resources.begin();
    std::string bowFileStr = *it;
    it++;
    std::string valueStr = *it;
    
    double threshold = Helper::getValueForResource(valueStr);
    
    std::vector<std::vector<double> > bow = Helper::readMatrixBinary(Helper::getFilenameForResource(bowFileStr, ds), n);

    fidelityFunctions* feat = new fidelityFunctions(groundSet, bow, threshold);
    
    if (bow.size() > 3000) {
    	    std::cout << "File: " << bowFileStr << std::endl;
    	    throw std::runtime_error("Too many words...");
    }
    
    // determine scaling
    double scaling = feat->eval(groundSet);
    scaleSubmodularFunctions* fct = new scaleSubmodularFunctions(groundSet, feat, 1.0/(100.0 * threshold));

    return fct;
}

submodularFunctions* Helper::createClusteredFacilityLocationComponent(ComponentDescriptor descriptor, std::unordered_set<int> &groundSet, int ds) {
    std::string type = descriptor.getType();
    std::list<std::string> resources = descriptor.getResources();
    
    //std::cout << "Constructing clustered facility location component." << std::endl;
    //std::cout << "There are " << resources.size() << " resources." << std::endl;
    if (resources.size() != 2) {
        throw std::runtime_error("One resource must be supplied to generate a clustered facility location component.");
    }
    
    /*int n = groundSet.size();
    std::list<std::string>::iterator it = resources.begin();
    std::string clusterResource = *it;
    it++;
    std::string rewardResource = *it;
    
    std::vector<std::vector<int>> clusterDataM = Helper::readMatrixBinaryInteger(Helper::getFilenameForResource(clusterResource, ds), n);
    std::vector<int> clusterData = clusterDataM[0];
    //std::cout << "Read cluster data." << std::endl;
    
    // compose clusters
    int nrClusters = 0;
    for (int i = 0; i < n; i++) {
    	    if (clusterData[i] > nrClusters)
    	    	    nrClusters = clusterData[i];
    }
    //std::cout << "#clusters: " << nrClusters << std::endl;
    std::vector<std::unordered_set<int>> *clusters = new std::vector<std::unordered_set<int>>(nrClusters);
    for (int i = 0; i < n; i++) {
    	    (*clusters)[clusterData[i]-1].insert(i);
    }

    char* rewardFileStr = Helper::getFilenameForResource(rewardResource, ds);
    facilityLocation* feat = new facilityLocation(groundSet, rewardFileStr);
    
    combineSubmodularFunctions *clsf = clusteredSubmodularFunctionsPointer(*feat, *clusters);
    
    // determine scaling
    double scaling = (double) nrClusters;
    scaleSubmodularFunctions* fct = new scaleSubmodularFunctions(groundSet, clsf, 1/scaling); */
    
    int n = groundSet.size();
    std::list<std::string>::iterator it = resources.begin();
    std::string clusterResource = *it;
    it++;
    std::string rewardResource = *it;
    
    std::vector<std::vector<int>> clusterDataM = Helper::readMatrixBinaryInteger(Helper::getFilenameForResource(clusterResource, ds), n);
    std::vector<int> clusterData = clusterDataM[0];

    char* rewardFileStr = Helper::getFilenameForResource(rewardResource, ds);
    std::vector<std::vector<double>> rewards = Helper::readMatrixBinary(rewardFileStr, n);
    
    clusteredFacilityLocation *fct = new clusteredFacilityLocation(groundSet, clusterData, rewards[0]);

    return fct;
}

submodularFunctions* Helper::createClusteredDiversityComponent(ComponentDescriptor descriptor, std::unordered_set<int> &groundSet, int ds) {
    std::string type = descriptor.getType();
    std::list<std::string> resources = descriptor.getResources();
    
    //std::cout << "Constructing clustered diversity component." << std::endl;
    //std::cout << "There are " << resources.size() << " resources." << std::endl;
    if (resources.size() != 3) {
        throw std::runtime_error("One resource must be supplied to generate a clustered diversity component.");
    }
    
    int n = groundSet.size();
    std::list<std::string>::iterator it = resources.begin();
    std::string clusterResource = *it;
    it++;
    std::string rewardResource = *it;
    it++;
    std::string valueResource = *it;
    double value = Helper::getValueForResource(valueResource);
    //std::cout << "Value: " << value << std::endl;
    
    std::vector<std::vector<int>> clusterDataM = Helper::readMatrixBinaryInteger(Helper::getFilenameForResource(clusterResource, ds), n);
    std::vector<int> clusterData = clusterDataM[0];
    //std::cout << "Read cluster data." << std::endl;
    


    char* rewardFileStr = Helper::getFilenameForResource(rewardResource, ds);
    std::vector<std::vector<double>> rewards = Helper::readMatrixBinary(rewardFileStr, n);
    
    // setup features for feature based function
    /*    // compose clusters
    int nrClusters = 0;
    for (int i = 0; i < n; i++) {
    	    if (clusterData[i] > nrClusters)
    	    	    nrClusters = clusterData[i];
    }
    std::cout << "#clusters: " << nrClusters << std::endl;
    std::vector<std::unordered_set<int>> *clusters = new std::vector<std::unordered_set<int>>(nrClusters);
    for (int i = 0; i < n; i++) {
    	    (*clusters)[clusterData[i]-1].insert(i);
    }
    
    std::vector<double> featureWeights(nrClusters);
    for (int i = 0; i < nrClusters; i++) {
    	    featureWeights[i] = 1.0;
    }
    std::vector<struct Feature> features(n);
    for (int i = 0; i < n; i++) {   
            int *featureIndices = new int[1] {clusterData[i]-1}; 	    
    	    float* featureVec = new float[1] { (float) rewards[0][i]};
    	    
    	    struct Feature feat;
    	    feat.index = i;
    	    feat.num_uniq_wrds = 1;
    	    feat.featureIndex = featureIndices;
    	    feat.featureVec = featureVec;
    	    
    	    features[i] = feat;
    }
    featureBasedFunctions* feat = new featureBasedFunctions(groundSet, 1, features, featureWeights, new concaveFunctionPower(value));
    
    combineSubmodularFunctions *clsf = clusteredSubmodularFunctionsPointer(*feat, *clusters);
    
    // determine scaling
    double accRewards[nrClusters];
    for (int i = 0; i < n; i++) {
    	    accRewards[clusterData[i]-1] = 0.0;
    }
    for (int i = 0; i < n; i++) {
    	    accRewards[clusterData[i]-1] += rewards[0][i];
    }
    double scaling = 0.0;
    for (int i = 0; i < nrClusters; i++) {
    	    scaling += pow(accRewards[i], value);
    }    
    std::cout << "Scaling: " << scaling << std::endl;
    scaleSubmodularFunctions* fct = new scaleSubmodularFunctions(groundSet, clsf, 1.0/scaling); */
    
    clusteredDiversity *fct2 = new clusteredDiversity(groundSet, clusterData, rewards[0],  value);

    return fct2;
}

submodularFunctions* Helper::createHingeLossComponent(ComponentDescriptor descriptor, std::unordered_set<int> &groundSet, int ds) {
	submodularFunctions *fct = new hingeLoss(groundSet, 10);
	
	return fct;
}

submodularFunctions* Helper::createSurrogateLoss(std::unordered_set<int> &groundSet, int ds, std::vector<std::unordered_set<int>> summaries) {
        int n = groundSet.size();

        std::vector<std::vector<double> > bow = Helper::readMatrixBinary(Helper::getFilenameForResource("bow:evaluation", ds), n);
        
        surrogateRougeFunctions *rouge = new surrogateRougeFunctions(groundSet, bow, summaries);
        
        // determine scaling
        double scaling = rouge->eval(groundSet);
        if (std::isnan(scaling))
        	throw std::runtime_error("Surrogate rouge loss returned NaN.");
        //std::cout << "Scaling: " << scaling << std::endl;
        // scaleSubmodularFunctions* fct = new scaleSubmodularFunctions(groundSet, rouge, 1.0/scaling);
        scaleSubmodularFunctions* fct = new scaleSubmodularFunctions(groundSet, rouge, 1.0);
	return fct;
}

submodularFunctions* Helper::createRouge(std::unordered_set<int> &groundSet, int ds, std::vector<std::unordered_set<int>> summaries) {
        int n = groundSet.size();

        std::vector<std::vector<double> > bow = Helper::readMatrixBinary(Helper::getFilenameForResource("bow:evaluation", ds), n);
        
        /*std::cout << "First 5 summaries." << std::endl;
        for (int i = 0; i < 5; i++) {
           std::unordered_set<int>::iterator it;
           for (it = summaries[i].begin(); it != summaries[i].end(); it++) {
           	   std::cout << *it << " ";
           }
           std::cout << std::endl;
        }*/
        
        rougeFunctions *rouge = new rougeFunctions(groundSet, bow, summaries);
        
       // determine scaling
       double scaling = rouge->eval(groundSet);
        scaleSubmodularFunctions* fct = new scaleSubmodularFunctions(groundSet, rouge, 1/scaling);
	return fct;
}

submodularFunctions* Helper::createComplementRouge(std::unordered_set<int> &groundSet, int ds, std::vector<std::unordered_set<int>> summaries) {
        int n = groundSet.size();

        std::vector<std::vector<double> > bow = Helper::readMatrixBinary(Helper::getFilenameForResource("bow:evaluation", ds), n);
        rougeFunctions *rouge = new rougeFunctions(groundSet, bow, summaries);
        double scaling = rouge->eval(groundSet);
        scaleSubmodularFunctions* fct = new scaleSubmodularFunctions(groundSet, rouge, 1.0/scaling);
        submodularFunctions *complement = new complementSubmodularFunctions(fct);
	return complement;
}

submodularFunctions* Helper::createOneMinusRouge(std::unordered_set<int> &groundSet, int ds, std::vector<std::unordered_set<int>> summaries) {
        int n = groundSet.size();

        std::vector<std::vector<double> > bow = Helper::readMatrixBinary(Helper::getFilenameForResource("bow:evaluation", ds), n);
        
        rougeFunctions *rouge = new rougeFunctions(groundSet, bow, summaries);
        
        // determine scaling
        double scaling = rouge->eval(groundSet);
        scaleSubmodularFunctions* fct = new scaleSubmodularFunctions(groundSet, rouge, 1.0/scaling);
        
        constantFunction *constFct = new constantFunction(groundSet, 1.0);
        std::vector<submodularFunctions*> fcts(2);
	fcts[0] = constFct;
	fcts[1] = fct;
	std::vector<double> weights(2, 0.0);
	weights[0] = 1;
	weights[1] = -1;
    
        combineSubmodularFunctions* combinedFunction = new combineSubmodularFunctions(groundSet, fcts, weights);
        if (combinedFunction == NULL) 
        	throw std::runtime_error("Could not create object.");
        
	return combinedFunction;
}
 

char* Helper::getFilenameForResource(std::string resource, int ds) {
    int separatorIndex = resource.find(":");
    std::string type = resource.substr(0, separatorIndex);
    std::string url = resource.substr(separatorIndex + 1, std::string::npos);

    //std::cout << "Type: " << type << std::endl;
    //std::cout << "URL: " << url << std::endl;

    std::string path = "../../../precomputed/resources/";
    if (type.compare("Similarity") == 0) {
        path += "similarity/";
    } else if(type.compare("bow") == 0) {
    	path += "bow/";
    } else if(type.compare("Cluster") == 0) {
    	path += "cluster/";
    } else if(type.compare("QualityMeasure") == 0) {
    	path += "quality/";
    } else {
        throw std::runtime_error("Invalid resource.");
    }

    path += url;
    path += "/";
    path += std::to_string(ds+1);
    path += ".dat";

    char* file = new char[path.size() + 1];
    strcpy(file, path.c_str());

    return file;
}

void Helper::timeFunctions(std::list<ComponentDescriptor> componentDescriptorList, std::vector<submodularFunctions*> componentsList) {
    std::cout << "Timing functinos..." << std::endl;
    std::unordered_set<int> testSet = unordered_set<int>({0, 1, 2, 3, 4, 5, 6, 7, 8, 9});
    
    int nrTries = 100;
    std::list<ComponentDescriptor>::iterator it = componentDescriptorList.begin();
    for (int i = 0; i < componentsList.size(); i++) {
    	    clock_t start;
    	    clock_t end;

    	    start = clock();
    	    double valStd;
    	    for (int n = 0; n < nrTries; n++) {
    	    	    valStd = componentsList[i]->eval(testSet);
    	    }
    	    end = clock();
    	    double timeStd = (end - start) / (double)(CLOCKS_PER_SEC / 1000) ;
    	    
    	    start = clock();
    	    double valFast;
    	    for (int n = 0; n < nrTries; n++) {
    	    	    valFast = componentsList[i]->evalFast(testSet, 1);
    	    }
    	    end = clock();
    	    double timeFast = (end - start) / (double)(CLOCKS_PER_SEC / 1000) ;
    	    
    	    std::cout << it->getType() << " => Time: " << timeStd << " ms (fast: " << timeFast << "; valOk: " << (abs(valStd - valFast) < 1e-6) << ")" << std::endl;
   
    	    it++;
    }
}

void Helper::printSummary(std::unordered_set<int> &sset) {
	for (std::unordered_set<int>::iterator sit = sset.begin(); sit != sset.end(); sit++) {
		std::cout << *sit << " ";
	}
}

double Helper::getValueForResource(std::string resource) {
	std::string valueStr = resource.substr(6, std::string::npos);
	return std::stod(valueStr, NULL);
}

configT Helper::parseConfig(std::string configFileName) {
	std::cout << "[Config] Loaded from " << configFileName << std::endl;
	
	configT config;
	ConfigFile cf(configFileName);
	
	// int testDataset;	
	int testDataset = cf.Value("general", "testDataset");
	config.testDataset = testDataset;
	std::cout << "[Config] testDataset=" << testDataset << std::endl;
	
	// // scenarioType scenario = CHEATING;
	std::string scenarioTypeStr = cf.Value("general", "scenario");
	if (scenarioTypeStr.compare("CHEATING") == 0)
		config.scenario = CHEATING;
	else if (scenarioTypeStr.compare("REAL") == 0)
		config.scenario = REAL;
	else
		throw std::runtime_error("Read invalid scenario from config: " + scenarioTypeStr);
	std::cout << "[Config] scenarioType=" << config.scenario << std::endl;
	
	// type of projection
	std::string projectionTypeStr = cf.Value("general", "projectionType");
	if (projectionTypeStr.compare("SIMPLEX") == 0) 
		config.projectionType = SIMPLEX;
	else if (projectionTypeStr.compare("POSITIVE") == 0) 
		config.projectionType = POSITIVE;
	else if (projectionTypeStr.compare("NONE") == 0)
		config.projectionType = NONE;
	else
		throw std::runtime_error("Read invalid projectino type from config: " + projectionTypeStr);
	std::cout << "[Config] Projection=" << config.projectionType << std::endl;
	
	// optimizer
	std::string optimizerStr = cf.Value("general", "optimizer");
	if (optimizerStr.compare("SGD") == 0)
		config.optimizer = SGD;
	else if (optimizerStr.compare("GRADIENTSCALING") == 0)
		config.optimizer = GRADIENTSCALING;
	else
		throw std::runtime_error("Read invalid optimizer type from config: " + optimizerStr);
	
	// inference
	std::string inferenceStr = cf.Value("general", "inference");
	if (inferenceStr.compare("GREEDY") == 0)
		config.inference = GREEDY;
	else if(inferenceStr.compare("RANDOMIZEDGREEDY") == 0) 
		config.inference = RANDOMIZEDGREEDY;
	else if(inferenceStr.compare("SUBSUP") == 0)
		config.inference = SUBSUP;
	else if(inferenceStr.compare("SUPSUB") == 0)
		config.inference = SUPSUB;
	else 
		throw std::runtime_error("Read invalid inference type from config: " + inferenceStr);
	std::cout << "[Config] inference=" << config.inference << std::endl;
	
	// loss
	std::string lossStr = cf.Value("general", "loss");
	if (lossStr.compare("SURROGATE") == 0)
		config.loss = SURROGATE;
	else if(lossStr.compare("ONEMINUSROUGE") == 0)
		config.loss = ONEMINUSROUGE;
	else if(lossStr.compare("COMPLEMENTROUGE") == 0)
		config.loss = COMPLEMENTROUGE;
	else
		throw std::runtime_error("Read invalid loss type from config: " + lossStr);
	std::cout << "[Config] loss=" << config.loss << std::endl;
	
	// l2-regularizer
	double l2reg = cf.Value("general", "l2reg");
	config.l2reg = l2reg;
	std::cout << "[Config] l2-reg=" << config.l2reg << std::endl;
	
	// initial eta
	double etaInitial = cf.Value("general", "etaInitial");
	config.etaInitial = etaInitial;
	std::cout << "[Config] etaInitial=" << config.etaInitial << std::endl;
	
	int nrIter = cf.Value("general", "nrIter");
	config.nrIter = nrIter;
	std::cout << "[Config] nrIter=" << config.nrIter << std::endl;
	
	// logEvery
	int logEvery = cf.Value("general", "logEvery");
	config.logEvery = logEvery;
	std::cout << "[Config] logEvery=" << config.logEvery << std::endl;
	
	// initializeWithGreedy
	std::string initializeWithGreedyStr = cf.Value("general", "initializeWithGreedy");
	bool initializeWithGreedy = false;
	if (initializeWithGreedyStr.compare("true") == 0) 
		initializeWithGreedy = true;
	else if (initializeWithGreedyStr.compare("false") == 0)
		initializeWithGreedy = false;
	else
		throw std::runtime_error("Read invalid initialization for subsup from config: " + initializeWithGreedyStr);
	config.initializeWithGreedy = initializeWithGreedy;
	std::cout << "[Config] initializeWithGreedy=" << config.initializeWithGreedy << std::endl;
	
	std::string componentsConfigurationFileStr = cf.Value("general", "components");
	config.componentsConfigurationFile = componentsConfigurationFileStr;
	std::cout << "[Config] components=" << config.componentsConfigurationFile << std::endl;

	std::string resultFilenameStr = cf.Value("general", "resultFilename");
	config.resultFilename = resultFilenameStr;
	std::cout << "[Config] resultFilename=" << config.resultFilename << std::endl;
	
	return config;
}

std::vector<std::vector<unordered_set<int> > > Helper::loadSummaries() {
	std::cout << "Loading summaries..." << std::endl;
	
	std::vector<std::vector<unordered_set<int> > > summaries;
	summaries.resize(14);
	int images[10000];
	char line[300000];
	for (int ds = 0; ds < 14; ds++) {
		std::string summaryFile = preDir + "Data/set" + std::to_string(ds+1) + "/summary_0.10.txt";
		
		ifstream iFile;
		FILE *fp = NULL; 
		if ((fp = fopen(summaryFile.c_str(), "rt")) == NULL){
			throw std::runtime_error("Couldn't load summaries.");
	    	}
		while (fgets(line,sizeof(line),fp) != NULL){
			long int s = 0;
			int pos = 0;
			int a = 0;
			while (sscanf(line+s, "%d %n", &images[a], &pos) == 1) {
			 s += pos;
			 a++;
			}
			
			unordered_set<int> summary = unordered_set<int>();
			for (int i=0; i<a; i++){
			  summary.insert(images[i]-1); // correct for indexing used by matlab
			}
			summaries[ds].push_back(summary);
		}
		fclose(fp);
	}

	return summaries;
}