#ifndef OPTIMIZER_H
#define OPTIMIZER_H

class Optimizer {
public:
	virtual void updateWeights(std::vector<double> &weights, std::vector<double> &grad) = 0;
};

class OptimizerSGD : public Optimizer {
private:
	double etaInitial;
	double eta;
	double l2reg;
	int t;
public:
	OptimizerSGD(double etaInitial, double l2reg) : etaInitial(etaInitial), eta(etaInitial), l2reg(l2reg), t(1) {} ;
	void updateWeights(std::vector<double> &weights, std::vector<double> &grad);
};

class OptimizerGradientScaling : public Optimizer {
private:
	double etaInitial;
	double eta;
	double l2reg;
	double delta = 1e-1;
	int t;
	std::vector<double> normGrad;
public:
	OptimizerGradientScaling(double etaInitial, double l2reg) : etaInitial(etaInitial), eta(etaInitial), l2reg(l2reg), t(1) {} ;
	void updateWeights(std::vector<double> &weights, std::vector<double> &grad);
};

void OptimizerSGD::updateWeights(std::vector<double> &weights, std::vector<double> &grad) {
	// add regularizer to gradient
	int n = grad.size();
	for (int i = 0; i < n; i++) {
		grad[i] += l2reg * weights[i];
	}
	
	// normalize gradient
	double norm = 0.0;
	for (int i = 0; i < n; i++) {
		norm += grad[i] * grad[i];
	}
	norm = sqrt(norm);
	if (norm > 1e-9) {
		for (int i = 0; i < n; i++) {
			grad[i] /= norm;
		}
	}
	
	// update learning rate
	eta = etaInitial / sqrt(t);
	std::cout << "[Optimizer] Current learning rate = " << eta << std::endl;
	t++;
	
	// update weights
	for (int i = 0; i < n; i++) {
		weights[i] -= eta*grad[i];
	}
	
            /* grad = grad + lambda * currWeights;
            allGradients(:,t) = grad;
            if norm(grad) > 1e-9,
                grad = grad / norm(grad);
                fprintf(1, 'Gradient normalized.\n');
            end
            
            eta = options.etaInitial/sqrt(t);
            newWeights = currWeights - eta*grad; */	
}

void OptimizerGradientScaling::updateWeights(std::vector<double> &weights, std::vector<double> &grad) {
	/*             grad = grad + lambda * currWeights;
            allGradients(:,t) = grad;
            delta = 1e-1;
            h = zeros(nrComponents, 1);
            for i = 1:nrComponents,
                h(i) = delta + norm(allGradients(i,1:t));
            end
            
            ngrad = eta*grad./h;
            fprintf(1, 'HINGE GRAD: %f (eta=%f, h = %f)\n', ngrad(end), eta, h(end));
            newWeights = currWeights - ngrad; */
            
	// add regularizer to gradient
	int n = grad.size();
	for (int i = 0; i < n; i++) {
		grad[i] += l2reg * weights[i];
	}
	
	// compute h
	if (normGrad.size() == 0) {
		normGrad.resize(grad.size());
		normGrad.assign(grad.size(), 0.0);
	}
	for (int i = 0; i < grad.size(); i++) {
		normGrad[i] += grad[i]*grad[i];
	}
	std::vector<double> h(grad.size());
	for (int i = 0; i < grad.size(); i++) {
		h[i] = delta + sqrt(normGrad[i]);
	}
	
	// compute "new" gradient
	std::vector<double> ngrad(grad.size());
	for (int i = 0; i < grad.size(); i++) {
		ngrad[i] = grad[i] / h[i];
	}
	
	// update weights
	for (int i = 0; i < n; i++) {
		weights[i] -= eta*ngrad[i];
	}
	
}

#endif
 
 /*       %% SGD update
        if strcmp(options.SGDalgorithm, 'classic'),
            grad = grad + lambda * currWeights;
            allGradients(:,t) = grad;
            if norm(grad) > 1e-9,
                grad = grad / norm(grad);
                fprintf(1, 'Gradient normalized.\n');
            end
            
            eta = options.etaInitial/sqrt(t);
            newWeights = currWeights - eta*grad;
        elseif strcmp(options.SGDalgorithm, 'gradientScaling'),
            fprintf(1, 'Gradient scaling.\n');
            
            grad = grad + lambda * currWeights;
            allGradients(:,t) = grad;
            delta = 1e-1;
            h = zeros(nrComponents, 1);
            for i = 1:nrComponents,
                h(i) = delta + norm(allGradients(i,1:t));
            end
            
            ngrad = eta*grad./h;
            fprintf(1, 'HINGE GRAD: %f (eta=%f, h = %f)\n', ngrad(end), eta, h(end));
            newWeights = currWeights - ngrad;
        else
            error('Invalid SGD method.');
        end */
