/*
 *	A tutorial for SOTK.
 *	Author: Rishabh Iyer
 *	Email: rkiyer@uw.edu
 *	Melodi Lab, University of Washington, Seattle
 *
 *	Here, we run medium scale experiments using the TIMIT corpus, for speech recognition.
 *	The size of the dataset is n = 4620.
 *	Our algorithms are tested on variants of datasubset selection problem for speech recognition
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <sys/stat.h>
#include <sys/types.h>
#include <set>
#include <unordered_set>
#include <algorithm>
#include <stdexcept>

#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "math.h"
#include "time.h"
#include <stdarg.h>
#include <assert.h>
#include <queue>
#include <stdexcept>
using namespace std;

// Utilities
#include "Utils/error.h"
#include "Utils/sort.h"
#include "Utils/utils.h"
#include "Utils/fileHandling.h"
#include "Utils/totalOrder.h"

//Submodular Functions
#include "Functions/setFunctions.h"
#include "Functions/submodularFunctions.h"
#include "Functions/modularFunctions.h"
#include "Functions/featureBasedFunctions.h"
#include "Functions/facilityLocation.h"
#include "Functions/graphCutFunctions.h"
#include "Functions/saturatedCoverage.h"
#include "Functions/complexityFunctions.h"
#include "Functions/confusabilityFunctions.h"
#include "Functions/combineSubmodularFunctions.h"
#include "Functions/scaleSubmodularFunctions.h"
#include "Functions/combineSubmodularModularFunctions.h"
#include "Functions/differenceSubmodularFunctions.h"
#include "Functions/complementSubmodularFunctions.h"
#include "Functions/truncateSubmodularFunctions.h"
#include "Functions/clusteredSubmodularFunctions.h"


#include "ImageSummarization/Helper.h"
#include "Projection/projection.h"
#include "Projection/projectionOntoSimplex.h"
#include "Projection/projectionPositive.h"
#include "Projection/projectionNone.h"
#include "Optimizer/Optimizer.h"

//Algorithms
#include "Algorithms/lazyGreedyMax.h"
#include "Algorithms/lazyGreedySC.h"
#include "Algorithms/revlazyGreedyMax.h"
#include "Algorithms/randomizedGreedyMax.h"
#include "Algorithms/bidirGreedyMax.h"
#include "Algorithms/localSearchMax.h"
#include "Algorithms/minNormPoint.h"
#include "Algorithms/naiveGreedyMax.h"
#include "Algorithms/naiveGreedyMaxFixedSize.h"
#include "Algorithms/mminCC.h"
#include "Algorithms/mminSCSC.h"
#include "Algorithms/mminSCSK.h"
#include "Algorithms/mminUC.h"
#include "Algorithms/mminSubSup.h"
#include "Algorithms/mminModMod.h"
#include "Algorithms/mminSupSub.h"
#define SMALLEST_NUMBER -1e70
#define LARGEST_NUMBER 1e70

using namespace std;

int main(int argc, char** argv) {
    clock_t start;
    clock_t end;
    
    if (argc < 2) 
    	    throw std::runtime_error("You must supply a config file.");
    
    // try to parse config
    configT config = Helper::parseConfig(argv[1]);
    std::string filename(config.componentsConfigurationFile);
	
    // set up training and test dataset
    std::unordered_set<int> trainingDatasets;
    std::unordered_set<int> testDatasets;
    if (config.scenario == CHEATING) {
    	    trainingDatasets.insert(config.testDataset);
    	    testDatasets.insert(config.testDataset);
    } else {
    	    testDatasets.insert(config.testDataset);
    	    for (int ds = 0; ds < 14; ds++) {
    	    	    if ((ds + 1) != config.testDataset)
    	    	    	    trainingDatasets.insert(ds + 1);
    	    }
    }
	
    int n; // GroundSet size
    unordered_set<int> groundSet = unordered_set<int>(); // GroundSet.
    unordered_set<int> emptyset = unordered_set<int>();
    unordered_set<int> testSet = unordered_set<int>({0, 1, 2, 3, 4, 5, 6, 7, 8, 9});
    int numWords;

    cout << "Loading component information from file." << std::endl;
    std::list<ComponentDescriptor> componentDescriptorList = Helper::readComponentsFile(filename);

    // Helper::dumpComponentDescriptorList(componentDescriptorList);

    // set up ground set
    n = 100;
    for (int i = 0; i < n; ++i)
        groundSet.insert(i);

    // construct components
    cout << "Constructing components." << std::endl;
    std::list<ComponentDescriptor>::iterator it;
    int idx = 0;
    std::vector<std::vector<submodularFunctions*>> componentsList(14);
    for (it = componentDescriptorList.begin(); it != componentDescriptorList.end(); it++) {
	    for (int ds = 0; ds < 14; ds++){
		string type = it->getType();
		std::cout << "[" << idx << "] Constructing component: " << it->getType() << std::endl;
	
		std::list<std::string> resources = it->getResources();
		std::list<std::string>::iterator itr;
		//for (itr = resources.begin(); itr != resources.end(); itr++) {
		//    std::cout << "     Resource: " << *itr << std::endl;
		//}
	
		submodularFunctions *fct;
		if (type.compare("FacilityLocation") == 0) {
		    fct = Helper::createFacilityLocationComponent(*it, groundSet, ds);
		} else if (type.compare("SumCoverage") == 0) {
		    fct = Helper::createSumCoverageComponent(*it, groundSet, ds);
		} else if (type.compare("CutCoverage") == 0) {
		    fct = Helper::createCutCoverageComponent(*it, groundSet, ds);
		} else if (type.compare("CosineSimilarity") == 0) {
		    fct = Helper::createCosineSimilarityComponent(*it, groundSet, ds);
		} else if (type.compare("CompressedWordCount") == 0) {
		    fct = Helper::createCompressedWordCountComponent(*it, groundSet, ds);
		} else if (type.compare("Fidelity") == 0) {
		    fct = Helper::createFidelityComponent(*it, groundSet, ds);	
		} else if (type.compare("ThresholdedFeatureCounts") == 0) {
		    fct = Helper::createThresholdedFeatureCountsComponent(*it, groundSet, ds);
		} else if (type.compare("ClusteredFacilityLocation") == 0) {
		    fct = Helper::createClusteredFacilityLocationComponent(*it, groundSet, ds);
		} else if (type.compare("ClusteredDiversity") == 0) {
		    fct = Helper::createClusteredDiversityComponent(*it, groundSet, ds);
		} else if (type.compare("HingeLoss") == 0) {
		    fct = Helper::createHingeLossComponent(*it, groundSet, ds);
		} else {
		    throw std::runtime_error("Invalid component.");
		}
		
		// sanity check
		if (fct != NULL) {
			double valEmpty = fct->eval(emptyset);
			double valUniverse = fct->eval(groundSet);
			
			//std::cout << "Range: " << valEmpty << " -- " << valUniverse << std::endl;
			//std::cout << "Val for testset: " << fct->eval(testSet) << "." << std::endl;
			
			componentsList[ds].push_back(fct);
		} else {
			throw std::runtime_error("Encountered NULL function pointer.");
		}
		
		// std::cout << std::endl;
	    }
	    idx++;
    }
    
    // throw std::runtime_error("Stopped after creating components.");
    
    // load summaries
    std::cout << "Loading summaries..." << std::endl;
    std::vector<int> nrSamplesInDataset(14);
    std::vector<std::vector<unordered_set<int> > > summaries = Helper::loadSummaries();
    for (int ds = 0; ds < 14; ds++) {
    	    std::cout << "Dataset " << (ds+1) << " contains " << summaries[ds].size() << " summaries." << std::endl;
    	    nrSamplesInDataset[ds] = summaries[ds].size();
    }

    cout << "Constructing rouge: " ;
    std::vector<submodularFunctions*> rouges(14);
    for (int ds = 0; ds < 14; ds++) {
    	std::cout << "*";
	submodularFunctions *fct = Helper::createRouge(groundSet, ds, summaries[ds]);
	rouges[ds] = fct;
    }
    std::cout << std::endl;
    
    cout << "Constructing one minus rouge: ";
    std::vector<submodularFunctions*> oneMinusRouge(14);
    for (int ds = 0; ds < 14; ds++) {
    	 std::cout << "*";
	submodularFunctions *fct = Helper::createOneMinusRouge(groundSet, ds, summaries[ds]);
	double valEmpty = fct->eval(emptyset);
	double valUniverse = fct->eval(groundSet);
	
	//std::cout << "[" << ds << "] " << "Range: " << valEmpty << " -- " << valUniverse << std::endl;
	//std::cout << "[" << ds << "] " << "Val for testset: " << fct->eval(testSet) << "." << std::endl;
	
	oneMinusRouge[ds] = fct;
    }
    std::cout << std::endl;
    
    // time rouge
     /*start = clock();
    for (int i = 0; i < 100; i++) {
    	    // rouges[0]->evalFast(testSet, 1);
    	    rouges[0]->evalFast(testSet, 1);
    }
    end = clock();
	double timeMs = (end - start) / (double)(CLOCKS_PER_SEC / 1000);
	std::cout << " (took " << timeMs << " ms)" << std::endl;
    throw std::runtime_error("STOP HERE"); */
    
    // Time functions
    // Helper::timeFunctions(componentDescriptorList, componentsList[0]);
 
    /* PERFORM MAX MARGIN TRAINING ********************************************/
    int nrIter = config.nrIter;
    int nrComponents = componentsList[0].size();
    std::cout << "Length of vector: " << nrComponents << std::endl;
    
    std::vector<double> currWeights(nrComponents, 1.0 / ((double) nrComponents));
    
    // setup projection
    projection* proj = NULL;
    switch (config.projectionType) {
        case SIMPLEX:
        	std::cout << "[Projection] Onto simplex" << std::endl;
        	proj = new projectionOntoSimplex(nrComponents);
        	break;
        	
        case POSITIVE:
        	std::cout << "[Projection] Onto positive orthant" << std::endl;
        	proj = new projectionPositive(nrComponents);
        	break;
        	
        case NONE:
        	std::cout << "[Projection] None" << std::endl;
        	proj = new projectionNone(nrComponents);
        	break;
        	
        default:
        	throw std::runtime_error("Invalid projection.");
    }
    
    // setup optimizer
    Optimizer *optimizer = NULL;
    switch (config.optimizer) {
	case SGD:
		std::cout << "[Optimizer] SGD" << std::endl;
		optimizer = new OptimizerSGD(config.etaInitial, config.l2reg);
		break;
	case GRADIENTSCALING:
		std::cout << "[Optimizer] Gradient Scaling" << std::endl;
		optimizer = new OptimizerGradientScaling(config.etaInitial, config.l2reg);
		break;
	default:
		throw std::runtime_error("Invalid optimizer.");
    }
    
    // setup functions for loss augmented inference
    std::vector<combineSubmodularFunctions*> scoreFunctions(14);
    for (int ds = 0; ds < 14; ds++) {
    	    std::vector<submodularFunctions*> scoreFunction(nrComponents);
    	    for (int c = 0; c < nrComponents; c++) {
    	    	    scoreFunction[c] = componentsList[ds][c];
    	    }
    	    combineSubmodularFunctions* combinedFunction = new combineSubmodularFunctions(groundSet, scoreFunction, currWeights);
    	    scoreFunctions[ds] = combinedFunction;
    }
    
    // specify losses
    std::cout << "Creating losses: ";
    std::vector<submodularFunctions*> losses(14);
    for (int ds = 0; ds < 14; ds++) {
    	    switch (config.loss) {
		    case SURROGATE:
			    std::cout << "S";
			    losses[ds] = Helper::createSurrogateLoss(groundSet, ds, summaries[ds]);
			    break;
		    case ONEMINUSROUGE:
		    	    std::cout << "O";
			    losses[ds] = oneMinusRouge[ds];
			    break;
		    case COMPLEMENTROUGE:
		    	    std::cout << "C";
		    	    losses[ds] = Helper::createComplementRouge(groundSet, ds, summaries[ds]);
		    	    break;
		    default:
			    throw std::runtime_error("Invalid loss selected.");
    	    }
    }
    std::cout << std::endl;
    
    // compute average loss on dataset
    std::vector<double> avgLossHuman(14, 0.0);
    std::vector<std::vector<double>> lossHuman(14);
    for (int ds = 0; ds < 14; ds++) {
    	    lossHuman[ds].resize(nrSamplesInDataset[ds]);
    	    
    	    for (int s = 0; s < nrSamplesInDataset[ds]; s++) {
    	    	    lossHuman[ds][s] = losses[ds]->eval(summaries[ds][s]);
    	    	    avgLossHuman[ds] += lossHuman[ds][s];
    	    }
    	    avgLossHuman[ds] /= (double) nrSamplesInDataset[ds];
    	    std::cout << "DS " << ds << ": avg loss = " << avgLossHuman[ds] << std::endl; 
    }
    std::cout << std::endl;
    
    // setup functions for loss augmented inference using greedy
    std::vector<combineSubmodularFunctions*> laiFunctions(14);
    std::vector<double> laiWeights(nrComponents+1, 0.0);
    laiWeights[nrComponents] = 1.0;
    for (int ds = 0; ds < 14; ds++) {
    	    std::vector<submodularFunctions*> laiFunction(nrComponents+1);
    	    for (int c = 0; c < nrComponents; c++) {
    	    	    laiFunction[c] = componentsList[ds][c];
    	    }
    	    laiFunction[nrComponents] = losses[ds];
    	    combineSubmodularFunctions* combinedFunction = new combineSubmodularFunctions(groundSet, laiFunction, laiWeights);
    	    laiFunctions[ds] = combinedFunction;
    }
    
    // TODO: Compute stats of human/random summaries
    std::cout << "Computing dataset statistics: ";
    std::vector<double> avgRougeHuman(14, 0.0);
    std::vector<std::vector<double>> rougeHuman(14);
    for (int ds = 0; ds < 14; ds++) {
    	    std::cout << "*";
    	    rougeHuman[ds].resize(nrSamplesInDataset[ds]);
    	    
    	    for (int s = 0; s < nrSamplesInDataset[ds]; s++) {
    	    	    rougeHuman[ds][s] = rouges[ds]->eval(summaries[ds][s]);
    	    	    avgRougeHuman[ds] += rougeHuman[ds][s];
    	    }
    	    avgRougeHuman[ds] /= (double) nrSamplesInDataset[ds];
    }
    std::cout << std::endl;
    
    /* std::cout << "Rouge test: " << rouges[0]->eval(testSet) << std::endl;
    std::cout << "Rouge test fast: " << rouges[0]->evalFast(testSet) << std::endl;
    std::cout << "Average rouge: " << avgRougeHuman[0] << std::endl;
    
    start = clock();
    for (int i = 0; i < 100; i++) {
    	    laiFunctions[0]->evalFast(testSet);
    }
    end = clock();
    double timeMs = (end - start) / (double)(CLOCKS_PER_SEC / 1000);
    std::cout << "Took " << timeMs << " ms " << std::endl;
    
    throw std::runtime_error("STOP HERE."); */
    
    std::vector<std::vector<double> > allGradients;
    std::vector<std::vector<double> > allWeights;
    std::vector<double> avgWeights(nrComponents, 0.0);
    
    std::unordered_set<int>::iterator trSetIt;
    // compute initial parameters
    /*for (trSetIt = trainingDatasets.begin(); trSetIt != trainingDatasets.end(); trSetIt++) {
    	    int ds = *trSetIt - 1;
    	    for (int s = 0; s < nrSamplesInDataset[ds]; s++) {
    	    	    for (int f = 0; f < nrComponents; f++) {
    	    	    	    currWeights[f] += componentsList[f]->eval(summaries[ds][s]);
    	    	    }
    	    }
    }
    proj->project(currWeights);*/
    
    /* std::cout << "Initial weights:" << std::endl;
    for (int c = 0; c < nrComponents; c++) {
    	    std::cout << "  [" << c << "] w = " << currWeights[c] << std::endl;
    }*/
    
    /*std::cout << "Summaries: " << std::endl;
    for (int i = 0; i < 10; i++) {
    	 Helper::printSummary(summaries[0][i]);
    	 std::cout << std::endl;
    }*/
    
    std::vector<double> costList(100, 1.0);
    int t = 0;
    for (int k = 0; k < nrIter; k++) {
    	  for (trSetIt = trainingDatasets.begin(); trSetIt != trainingDatasets.end(); trSetIt++) {
    	        int ds = *trSetIt - 1;
    	        
    	        // get next sample [sampleIndex = 1 + mod(k, nrSamples(ds));]
    	        int sampleIndex = k % nrSamplesInDataset[ds];
		std::unordered_set<int> yt = summaries[ds][sampleIndex];
    	        
    	        std::cout << "-- Iteration " << k << " (sample #" << sampleIndex << " from dataset " << (ds + 1) << "; score " << rouges[ds]->eval(yt) << ")" << std::endl;
    	        std::cout << "[Sample] ";
    	        Helper::printSummary(yt);
    	        std::cout << std::endl;
    	        
		// update weights
		std::vector<double> tWeights(currWeights);
		tWeights.push_back(1.0);
		laiFunctions[ds]->updateWeights(tWeights);
		scoreFunctions[ds]->updateWeights(currWeights);
		
		//std::cout << "WEIGHTS:" << std::endl;
		//for (int i = 0; i < nrComponents; i++) 
		//	std::cout << currWeights[i] << std::endl;
		
		// (1) perform inference
		std::cout << "[Inference] ";
		std::unordered_set<int> yht = unordered_set<int>();
		start = clock();
		switch (config.inference) {
			case GREEDY:
				std::cout << "greedy: ";
				yht = naiveGreedyMaxFixedSize(*laiFunctions[ds], costList, 10, emptyset, 0);
				break;
			case RANDOMIZEDGREEDY:
				std::cout << "randomized greedy: ";
				yht = randomizedGreedyMax(*laiFunctions[ds], 10, emptyset, 0);
				break;
			case SUBSUP:
				std::cout << "subsup: ";
				// TODO: MAKE THIS PROPER
				if (config.initializeWithGreedy)
					yht = naiveGreedyMaxFixedSize(*laiFunctions[ds], costList, 10, emptyset, 0);
				yht = mminSubSup(*rouges[ds], *scoreFunctions[ds], 1.0, 1, 0, yht);
				break;
			case SUPSUB:
				std::cout << "supsub: ";
				// TODO: MAKE THIS PROPER
				if (config.initializeWithGreedy)
					yht = naiveGreedyMaxFixedSize(*laiFunctions[ds], costList, 10, emptyset, 0);
				yht = mminSupSub(*rouges[ds], *scoreFunctions[ds], 1.0, 1, 0, yht);
				break;
			default:
				throw std::runtime_error("Invalid inference.");
		}
		end = clock();
    	        double timeMs = (end - start) / (double)(CLOCKS_PER_SEC / 1000);
    	        Helper::printSummary(yht);
    	        std::cout << " (took " << timeMs << " ms)";
    	        std::cout << std::endl;
    	        std::cout << "Score=" << scoreFunctions[ds]->eval(yht) << " / Loss=" << losses[ds]->eval(yht) << std::endl;
		
		// (2) compute gradient
		std::vector<double> grad(nrComponents);
		for (int i = 0; i < nrComponents; i++) {
			grad[i] = componentsList[ds][i]->eval(yht) - componentsList[ds][i]->eval(yt);
		}
		allGradients.push_back(grad);
		
	    //std::cout << "Gradient:" << std::endl;
	    //for (int c = 0; c < nrComponents; c++) {
		//    std::cout << "  [" << c << "] g = " << grad[c] << std::endl;
	    //} 
		
		//for (int i = 0; i < nrComponents; i++) 
		//	std::cout << "Comp: " << componentsList[ds][i]->eval(yht) << " ### " << componentsList[ds][i]->eval(testSet) << " ### " << componentsList[ds][i]->evalFast(testSet,1)  << std::endl;
		
		// (3) update weights
		optimizer->updateWeights(currWeights, grad);
		
		// (4) project weights
		proj->project(currWeights);
		allWeights.push_back(currWeights);
		
		// update average weights
		for (int c = 0; c <nrComponents; c++) {
			double titer = (double) t;
			avgWeights[c] = avgWeights[c] * (titer / (titer+1.0)) + currWeights[c] / (titer+1.0);
		}
		
		// sanity check
		double sum = 0.0;
		for (int i = 0; i < nrComponents; i++) {
			sum += currWeights[i];
		}
		if (config.projectionType == SIMPLEX && abs(sum - 1.0) > 1e-3) {
			throw std::runtime_error("Projection onto simplex failed.");
		}
		
		std::cout << "[Projection] Sum of weights: " << sum << std::endl;
		
		// output stats, if wanted
		if (t % config.logEvery == 0) {
			scoreFunctions[ds]->updateWeights(currWeights);
			std::unordered_set<int> smax = naiveGreedyMaxFixedSize(*scoreFunctions[ds], costList, 10, emptyset, 0);
			std::cout << "[Test] FIN WEIGHTS: Current score maximizer: ";
			Helper::printSummary(smax);
			std::cout << std::endl;
			std::cout << "[Test] FIN WEIGHTS: => Score = " << rouges[ds]->eval(smax) << " (avg: " << avgRougeHuman[ds] << ")" << std::endl;
			
			// average weights
			scoreFunctions[ds]->updateWeights(avgWeights);
			smax = naiveGreedyMaxFixedSize(*scoreFunctions[ds], costList, 10, emptyset, 0);
			std::cout << "[Test] AVG WEIGHTS: Current score maximizer: ";
			Helper::printSummary(smax);
			std::cout << std::endl;
			std::cout << "[Test] AVG WEIGHTS: => Score = " << rouges[ds]->eval(smax) << " (avg: " << avgRougeHuman[ds] << ")" << std::endl;
		}
		
		t++;
		std::cout << std::endl;
	    }
    }
    
    // ** Final scoring ** 
    std::cout << std::endl << "Final scoring." << std::endl;
    
    // greedy max final
    int ds = config.testDataset - 1;
    scoreFunctions[ds]->updateWeights(currWeights);
    std::unordered_set<int> scoreMaximizerFinalWeights = naiveGreedyMaxFixedSize(*scoreFunctions[ds], costList, 10, emptyset, 0);
    std::cout << "[Test] FIN WEIGHTS: Current score maximizer: ";
    Helper::printSummary(scoreMaximizerFinalWeights);
    std::cout << std::endl;
    std::cout << "[Test] FIN WEIGHTS: => Score = " << rouges[ds]->eval(scoreMaximizerFinalWeights) << " (avg: " << avgRougeHuman[ds] << ")" << std::endl;
    
    scoreFunctions[ds]->updateWeights(avgWeights);
    std::unordered_set<int> scoreMaximizerAverageWeights = naiveGreedyMaxFixedSize(*scoreFunctions[ds], costList, 10, emptyset, 0);
    std::cout << "[Test] FIN WEIGHTS: Current score maximizer: ";
    Helper::printSummary(scoreMaximizerAverageWeights);
    std::cout << std::endl;
    std::cout << "[Test] FIN WEIGHTS: => Score = " << rouges[ds]->eval(scoreMaximizerAverageWeights) << " (avg: " << avgRougeHuman[ds] << ")" << std::endl;
    
    // TODO: store results
    // MATLAB: save(config.resultsFilename,  'stats', 'config', 'desc', 'components', '-v7');
    std::cout << "Saving results to " << config.resultFilename << "." << std::endl;
    ofstream resultsFile;
    resultsFile.open(config.resultFilename);
    
    resultsFile << "Number of components\n";
    resultsFile << nrComponents << "\n";
    resultsFile << "\n";
    
    resultsFile << "Final weights\n";
    for (int c = 0; c < nrComponents; c++)
    	    resultsFile << currWeights[c] << " ";
    resultsFile << "\n";
    resultsFile << "\n";
    
    resultsFile << "Average weights\n";
    for (int c = 0; c < nrComponents; c++)
    	    resultsFile << avgWeights[c] << " ";
    resultsFile << "\n";  
    resultsFile << "\n";
    
    resultsFile << "All weights\n";
    resultsFile << allWeights.size() << "\n";
    for (int i = 0; i < allWeights.size(); i++) {
    	for (int c = 0; c < nrComponents; c++)
		resultsFile << allWeights[i][c] << " ";
	resultsFile << "\n";    
    }
    resultsFile << "\n";
    
    resultsFile << "Greedy Max Final Weights\n";
    for (std::unordered_set<int>::iterator it = scoreMaximizerFinalWeights.begin(); it != scoreMaximizerFinalWeights.end(); it++)
    	    resultsFile << *it << " ";
    resultsFile << "\n";
    resultsFile << "\n";
    
    resultsFile << "Greedy Max Average Weights\n";
    for (std::unordered_set<int>::iterator it = scoreMaximizerAverageWeights.begin(); it != scoreMaximizerAverageWeights.end(); it++)
    	    resultsFile << *it << " ";
    resultsFile << "\n";    
    resultsFile << "\n";
    
    resultsFile << "Config\n";
    ifstream configFile;
    configFile.open(argv[1]);
    std::string line;
    while (getline(configFile, line))
    	    resultsFile << line << "\n";
    configFile.close();
    
    resultsFile.close();

    delete proj;
    delete optimizer;
}
