/*
 *	Speed up with accelerated greedy algorithm for submodular set cover (Minoux 1976, Wolsey 1982).
	Solves the problem \min_{f(X) \geq C} m(X), where f is a monotone submodular function and c is a modular cost function.
	This greedy algorithm has a log(n) approximation guarantee for this problem
	Author: Rishabh Iyer
	Email: rkiyer@uw.edu
	Melodi Lab, University of Washington, Seattle
 *
 */

#define SMALLEST_NUMBER -1e70
#define LARGEST_NUMBER 1e70

#ifndef LAZYGREEDYSC_H
#define LAZYGREEDYSC_H

#ifndef __INCREMENT__
#define __INCREMENT__
class Increment{
	public: 
		Increment(){};
		Increment(double x, int i){ value=x; index=i; }
		bool operator< (const Increment&) const;
	
		int get_index() const { return index; }
		double get_value() const{ return value; }
	private:
		int index;
		double value;
};

bool Increment::operator< (const Increment& right) const
{
	return value < right.value;
}
#endif

inline std::unordered_set<int> lazyGreedySC(submodularFunctions& f, vector<double> costList, double cover, int verbosity = 0, double alpha = 1){
	int nSelected = 0 ; // number of items selected
	std::unordered_set<int> selectedSet;
	bool* hashSelected = new bool[f.getSize()]; // an index table to show whether an item is selected
	for(int i=0;i<f.getSize();i++) hashSelected[i]=false;	
    	double maxV;
    	double preV = 0;
    	double objV = 0;
	double currentCost = 0;
	f.clearpreCompute(); // clear the precomputed statistics, in case it is already not cleared.
	// accelerated greedy algorithm implementation
	priority_queue <Increment> rho;
	// initilize the priority queue
	for (int i = 0; i < f.getSize(); i++) { 
		rho.push(Increment(f.evalGainsaddFast(selectedSet, i)/pow(costList[i],alpha),i)); 
	}
	int sort_cnt = 0;
	while (! rho.empty()) {
		int topid = rho.top().get_index();
		rho.pop();
		maxV = preV + f.evalGainsaddFast(selectedSet, topid);
		double newV = (maxV - preV)/pow(costList[topid],alpha); // return normalized gain 
		if (verbosity >= 5) printf("max gain = %.6e, rho->top() = %.6e\n",newV,rho.top().get_value());
		if (newV < rho.top().get_value()) {
			rho.push(Increment(newV, topid)); // if condition not satisfied, push back and re-sort
			sort_cnt++;
			if (verbosity >= 10) printf("Condition not met, re-sorting queue (%d iterations)\n",sort_cnt);
		}
		else {
			// guaranteed to be optimal because of submodularity
		    	hashSelected[topid] = true;
		    	if (maxV < cover) 
				selectedSet.insert(topid);
			else
				break;
			nSelected++;
			objV += newV;
			currentCost += costList[topid];
			if(verbosity >= 0) printf("Selecting %dth sample, curCost/cover = %.6e/%.6e, preV = %.10e, curV = %.10e\n", topid, currentCost, cover, preV, maxV);
			preV = maxV;
			f.updateStatisticsAdd(selectedSet, topid);
			sort_cnt = 0;
		}    
	}	
	return selectedSet;
}

#endif