/*
 *	Implementation of the Minimum Norm Point Algorithm for Submodular Minimization.
	The problem is: \min_{X} f(X), where f is submodular.
	
	Author: Hui Lin (hlin@tsuri.ee.washington.edu)
	This was modified to fit into the framework of HSSFOTK.
	Melodi Lab, University of Washington, Seattle
 *
 */

#ifndef MINNORMPOINT_H
#define MINNORMPOINT_H


//#define DEBUG
#define INF 1e30
#define NO_INFO 0
#define BASIC_INFO 1
#define MORE_INFO 10
#define HUGE_INFO 100

// Global variables

vector< vector<double> > RMinNormPoint; // the upper triangular matrix described in Wolfe 1976
vector< vector<double> > SMinNormPoint; // the set of bases
totalOrder<double> initialOrderMinNormPoint; // an initial total order that used for warm start
double tolMinNormPoint = 1e-4;
double maxNormInSMinNormPoint = 0;
double epsMinNormPoint = 1e-10;
long int numGreedyCallMinNormPoint = 1;
double lowerBoundDualityGapMinNormPoint = 0;
bool earlyStopEnabledMinNormPoint = true;

inline void affineCombine(vector<double> &x, vector<double> mu, int n){
	for(int i=0;i<n;++i) x[i] = 0.0 ;
	for(int i=0;i<SMinNormPoint.size();++i){
		  for(int j=0;j<n;++j)
			x[j] += SMinNormPoint[i][j]*mu[i];
	}        
}

inline double innerProd(vector<double> a, vector<double> b){
	assert(a.size()==b.size());
	double sum = 0 ;
	for(int i=0;i<a.size();i++) sum += a[i]*b[i];
	return sum ;
}

inline void addVector2S(vector<double>& x){
	double norm = innerProd(x,x);
	//update the R matrix
	// R is an upper triangle matrix such that
	// R^t*R = e*e^t + S^t*S,
	// therefore, whenever S changes, we update R.
	// See section 5, method D in Wolfe 1976 for details
	vector<double> newr ;
	newr.clear();
	double rr=0;
	for(int i=0;i<SMinNormPoint.size();i++){
		double r = 1.0 + innerProd(SMinNormPoint[i],x);
		for(int j=0;j<i;j++){
			r -= newr[j]*RMinNormPoint[i][j];
		}
		r /= RMinNormPoint[i][i] ;
		newr.push_back(r);
		rr += r*r;
	}
	newr.push_back(sqrt(1+norm-rr));
	SMinNormPoint.push_back(x);
	RMinNormPoint.push_back(newr);
	if(norm>maxNormInSMinNormPoint) maxNormInSMinNormPoint = norm ;
}


inline void findIntersectionPoint(vector<double> x0_mu, vector<double>& xstar_mu){

  //assert(x0_mu.size()==xstar_mu.size());

	double beta = 1;
	for(int i=0;i<x0_mu.size();i++){
		if( xstar_mu[i]-x0_mu[i] > 0 ){
			double tmpv = xstar_mu[i]/(xstar_mu[i]-x0_mu[i]) ;
		if(tmpv < beta) beta = tmpv ;
		}
	}
	vector<double> z_mu ;
	z_mu.clear();
	// modify S
	std::vector<int> tobedel ;
	tobedel.clear();
	for(int i=0;i<x0_mu.size();i++){
		double tmpv = (1-beta)*xstar_mu[i]+beta*x0_mu[i] ;
	if(tmpv<epsMinNormPoint) tobedel.push_back(i);
	else z_mu.push_back(tmpv);
	}
	for(int i=0;i<tobedel.size();i++){
		vector< std::vector<double> >::iterator its = SMinNormPoint.begin(); std::advance(its, tobedel[i]-i);
		vector< std::vector<double> >::iterator itr = RMinNormPoint.begin(); std::advance(itr, tobedel[i]-i);
		SMinNormPoint.erase(its);
		RMinNormPoint.erase(itr);
		//update R such that it maintains upper-triangle structure
		for(int j=tobedel[i]-i;j<SMinNormPoint.size();j++){
			double a = RMinNormPoint[j][j] ;
			double b = RMinNormPoint[j][j+1];
			double c = sqrt(a*a+b*b);
			for(int k=j;k<SMinNormPoint.size();k++){
				double d = RMinNormPoint[k][j]; double e = RMinNormPoint[k][j+1] ;
				RMinNormPoint[k][j] = (a*d + b*e)/c ;
				RMinNormPoint[k][j+1] = (-b*d + a*e)/c ;
			}
		}

		for(int j=tobedel[i]-i;j<SMinNormPoint.size();j++){
			RMinNormPoint[j].pop_back();
		}
	}
	xstar_mu = z_mu ;
}



/////////////////////////////////////////////////////
// find the min norm point in the affine hull of S //
/////////////////////////////////////////////////////
inline bool minNormAffineSpan(vector<double> & mu){
	if(SMinNormPoint.size()<=1){  mu.clear(); mu.push_back(1); return true; }
	int m = SMinNormPoint.size();
	vector<double> ubar(m,0.0);
	vector<double> u(m,0.0);
	for(int i=0;i<m;i++){
		ubar[i] = 1.0 ;
		for(int j=0;j<i;j++){
			ubar[i] -= ubar[j]*RMinNormPoint[i][j] ;
		}
		ubar[i] /= RMinNormPoint[i][i] ;
	}

	double sum_u = 0 ;
	for(int i=m-1;i>=0;i--){
		u[i] = ubar[i] ;
		for(int j=m-1;j>i;j--){
			u[i] -= u[j]*RMinNormPoint[j][i];
		}
		u[i] /= RMinNormPoint[i][i] ;
		sum_u += u[i] ;
	}
	mu.clear();
	for(int i=0;i<m;i++) mu.push_back(u[i]/sum_u);
	bool isInterior = true ;
	for(int i=0;i<SMinNormPoint.size();i++){
		if(mu[i] < epsMinNormPoint){ 
			isInterior = false ;
			break;
		}
	}
	return isInterior ;
}

inline bool setOptCheckDuality(vector<double> x, submodularFunctions& f){

    vector<double> myx ;
    unordered_set<int> A;
    double sum = 0;
    A.clear();
    unordered_set<int>::iterator iter = f.getGroundSet().begin();
    for(int i=0;i<x.size();i++, iter++){
      if(x[i] <=0){
        	sum += x[i];
        	A.insert(*iter);
      }
    }
    if(f.eval(A) - sum > lowerBoundDualityGapMinNormPoint) return false;

    return true ;
}

inline bool checkIfPointInS(vector<double> x, int n){
	bool flag = true;
	for(int i = 0; i < SMinNormPoint.size(); i++){
		for(int j = 0; j < n; j++){
			if(fabs(SMinNormPoint[i][j]-x[j]) > epsMinNormPoint){
				flag = false;
				break;
			}
		}
		if(flag) return flag;
	}
	return false;
}

  ////////////////////////////////////////////////////////////////
  // Debug utilities                                            //
  ////////////////////////////////////////////////////////////////

inline void printVector(vector<double> x,const char* name){
	printf("%s=[",name);
	for(int i=0;i<x.size();i++) printf("%.5e ",x[i]);
	printf("]\n");
  }

inline void infoMsg(int verb, int verbosity, const char * const format, ...){
	if(verbosity >= verb){
		va_list ap;
		va_start(ap, format);
		vprintf(format, ap);
		va_end(ap);
	}
  }

inline bool minNormPoint(submodularFunctions& f, unordered_set<int>& minSolSet, unordered_set<int>& maxSolSet, std::vector<double>& xstar, int verbosity = 1){
	unordered_set<int> groundset = f.getGroundSet();
	unordered_set<int> emptySet = unordered_set<int>(); 
	int n = f.getSize(); 
	vector<double> xhat; // minimum weighted base xhat of B(f)
	vector<double> x0; // min-norm point in the affine hull spanned by points in S
	vector<double> x0_mu; //linear coefficients s.t. x0 = \sum_{s_i \in S} mu_i s_i
	xhat.resize(n); 
	x0.resize(n); 
	xstar.resize(n);
	vector<double> xstar_mu;
	totalOrder<double> order(n) ;

	//initialization: xhat is initialized as an extreme point in the base polytope
	if(initialOrderMinNormPoint.size() == groundset.size())
		f.getSubgradient(initialOrderMinNormPoint, emptySet, xhat);
	else
		f.getSubgradient(order, emptySet, xhat);

	numGreedyCallMinNormPoint = 0;
	maxNormInSMinNormPoint = 0;
	SMinNormPoint.clear();
	RMinNormPoint.clear();
	addVector2S(xhat);
	xstar=xhat;
	xstar_mu.push_back(1);

	while(1){
		// step 2: find the minimum weighted base xhat using greedy algorithm
		//xhat.clear();
		affineCombine(xstar,xstar_mu, n);

		// update xhat using the ordering based on xstar, which is equivalent to
		// xhat = min <xstar,x>, x \in B(f), where B(f) is the base polytope of f.
		order.resetOrder(xstar);
		f.getSubgradient(order, emptySet, xhat);
		numGreedyCallMinNormPoint++;
		double sum = 0 ;
		for(int i=0;i<xstar.size();i++) sum += xstar[i] * (xstar[i]-xhat[i]);
		infoMsg(MORE_INFO, verbosity, "termination(%.4e,%.4e),|x|:%.25e,# bases:%d\n",sum, tolMinNormPoint*maxNormInSMinNormPoint,innerProd(xstar,xstar),SMinNormPoint.size());
		if(sum < tolMinNormPoint * maxNormInSMinNormPoint){
			infoMsg(BASIC_INFO, verbosity, "Minimum norm point found!\n"); // xstar is the min norm solution
			break;
		}
		if (earlyStopEnabledMinNormPoint > 0 && setOptCheckDuality(xstar, f)){
    			infoMsg(BASIC_INFO, verbosity, "Early (duality) stopping triggered!!\n");
    			break ;
    		}
		if(checkIfPointInS(xhat, n)){
			infoMsg(BASIC_INFO, verbosity, "point already in S!\n");
			break;
		}
		//add xhat to set S, and update the norm matrix
		addVector2S(xhat);
		xstar_mu.push_back(0);
		if(SMinNormPoint.size() > groundset.size()){
			infoMsg(BASIC_INFO, verbosity, "Trying to augment full-dimensional S!\n");
			break;
		}
		//setp 3:
		while(1){
			if ( minNormAffineSpan(x0_mu)){ // x0 is in the interior of the simplex S
				xstar_mu = x0_mu ;
				break ;
			}
			else{
				findIntersectionPoint(x0_mu,xstar_mu);
			}
		}
	}

	minSolSet.clear();
	maxSolSet.clear();
	bool isUnique = true ;
	for(int i=0;i<xstar.size();i++){
		if(xstar[i] < -epsMinNormPoint ) minSolSet.insert(i);
		if(xstar[i] <= epsMinNormPoint ) maxSolSet.insert(i);
		if(fabs(xstar[i]) < epsMinNormPoint) isUnique = false ;
	}

	return isUnique ;
}


#endif