/*
 *	The naive greedy algorithm for set function maximization. This algorithm has guarantees if the function is monotone submodular, under cardinality/knapsack or matroid constraints. In general, though, it is a useful heuristic, which seems to work well in practice.
	Solves the problem \max_{c(X) \leq B} f(X), where f is a monotone submodular function and c is a modular cost function.
	Special case of this is the problem, \max_{|X| \leq k} f(X).
	Author: Rishabh Iyer
	Melodi Lab, University of Washington, Seattle

	Note: This algorithm is considerably slower than the accelerated greedy, so if you have a reason to believe your function is submodular (or approximately so), 		please use the lazy greedy algorithm.
 *
 */

#ifndef NAIVEGREEDYMAXFIXEDSIZE_H
#define NAIVEGREEDYMAXFIXEDSIZE_H

#define SMALLEST_NUMBER -1e70
#define LARGEST_NUMBER 1e70

#define USEOMP 1

#ifdef USEOMP
#include <omp.h>
#endif

inline std::unordered_set<int> naiveGreedyMaxFixedSize(setFunctions& f, vector<double> costList, double budget, std::unordered_set<int> startSet = std::unordered_set<int>(), int verbosity = 1, double alpha = 1){
	if(verbosity > 0) cout<<"Started the naive greedy algorithm\n"<<flush;
	std::unordered_set<int> sset = startSet;	
    	double currvalbest;
	int curridbest = -1;
	double currentCost = 0;
	int iter = 0;
	std::unordered_set<int> groundSet = f.getGroundSet();
	f.setpreCompute(startSet); // Set the precomputed statistics.
	// naive greedy algorithm implementation
	while (1) {
		curridbest = -1;
		iter++;
		currvalbest = SMALLEST_NUMBER;
		
		#ifdef USEOMP
		#pragma omp parallel for
		for(int i = 0; i < groundSet.size(); i++) {
			std::unordered_set<int>::iterator it = groundSet.begin();
			std::advance(it, i);
			bool ignore = false;
			
			#pragma omp critical 
			{
				if( (sset.find(*it) != sset.end()))
					ignore = true;
			}
			if (ignore)
				continue;
			
			double tval = f.evalGainsaddFast(sset, *it)/pow(costList[*it],alpha);
			#pragma omp critical
			{
				if (tval > currvalbest) {
					currvalbest = tval;
					curridbest = *it;
				}
			}
		}
		#else
		std::unordered_set<int>::iterator it;
		for( it = groundSet.begin(); it != groundSet.end(); it++ ) {
			if( (sset.find(*it) == sset.end())) {
				double tval = f.evalGainsaddFast(sset, *it)/pow(costList[*it],alpha);
				if (tval > currvalbest) {
					currvalbest = tval;
					curridbest = *it;
					//std::cout << "id: " << *it << "  val: " << tval << std::endl;
				}
			}
		}
		#endif
		//std::cout << "Done (" << curridbest << ")." << std::endl;
		
		if (curridbest < 0 || curridbest > groundSet.size() - 1) {
			std::cout << "curridbest: " << curridbest << std::endl;
			throw std::runtime_error("Invalid best item in greedy...");
		}
		if(currentCost + costList[curridbest] <= budget){
			f.updateStatisticsAdd(sset, curridbest);
			currentCost+=costList[curridbest];
			sset.insert(sset.end(), curridbest);
			if(verbosity > 0) cout<<"Adding element "<< curridbest<<" in iteration "<<iter<<" and the objective value is "<<f.eval(sset)<<"\n"<<flush;
		}
		else{
			break;
		}
			    
	}	
	return sset;
}

#endif