/*
 *	Header file for defining the facility location function: f(X) = \sum_{i \in V} \max_{j \in X} s_{ij}.
	Assumes that the kernel S is symmetric.
	Author: Rishabh Iyer.
	Email: rkiyer@uw.edu
	Melodi Lab, University of Washington, Seattle
 *
 */

#ifndef CLUSTEREDDIVERSITY_H
#define CLUSTEREDDIVERSITY_H

#include <math.h>

class clusteredDiversity: public submodularFunctions{
	protected:
	int nrClusters;
	vector<double> rewards;
	vector<int> clusterAssignments;
	double a;
	double normalization;
	
	vector<double> preCompute; // Precomputed statistics of length nrClusters. For a given set X, preCompute[i] = accReward[cluster i]
	int sizepreCompute;// size of the precompute statistics (in this case, 2*n).
	unordered_set<int> preComputeSet; // This points to the preComputed Set for which the statistics p_X is calculated.
	public:
	// Functions
	clusteredDiversity(std::unordered_set<int>& groundSet, std::vector<int> clusterAssignments, std::vector<double> rewards, double a);
	double eval(const unordered_set<int>& sset);
	double evalFast(const unordered_set<int>& sset, bool safe);
	double evalGainsadd(const unordered_set<int>& sset, int item);
	double evalGainsremove(const unordered_set<int>& sset, int item);
	double evalGainsaddFast(const unordered_set<int>& sset, int item, bool safe);
	double evalGainsremoveFast(const unordered_set<int>& sset, int item, bool safe);
	void updateStatisticsAdd(const unordered_set<int>& sset, int item, bool safe);
	void updateStatisticsRemove(const unordered_set<int>& sset, int item, bool safe);
	void setpreCompute(const unordered_set<int>& ssett);
	void clearpreCompute();
};

inline clusteredDiversity::clusteredDiversity(std::unordered_set<int>& groundSet, std::vector<int> clusterAssignments, std::vector<double> rewards, double a): submodularFunctions(groundSet), clusterAssignments(clusterAssignments), rewards(rewards), a(a) {
	
	nrClusters = 0;
	for (int i = 0; i < n; i++) {
    	    if (clusterAssignments[i] > nrClusters)
    	    	    nrClusters = clusterAssignments[i];
        }
        
	sizepreCompute = nrClusters;
	preCompute.assign(sizepreCompute, 0); //precomputed statistics for speeding up greedy
        
    // compute normalization
    double accRewards[nrClusters];
    for (int i = 0; i < n; i++) {
    	    accRewards[clusterAssignments[i]-1] = 0.0;
    }
    for (int i = 0; i < n; i++) {
    	    accRewards[clusterAssignments[i]-1] += rewards[i];
    }
    normalization = 0.0;
    for (int i = 0; i < nrClusters; i++) {
    	    normalization += pow(accRewards[i], a);
    }    
}


inline double clusteredDiversity::eval(const unordered_set<int>& sset){
	double sum = 0;
	std::vector<double> accRewards(nrClusters, 0.0);
	unordered_set<int>::const_iterator it;
	for( it = sset.begin(); it != sset.end(); it++ ){
		accRewards[clusterAssignments[*it]-1] += rewards[*it];
	}
	for (int i = 0; i < nrClusters; i++) {
		sum += pow(accRewards[i], a);
	}
		
	return sum / normalization;
}

inline double clusteredDiversity::evalFast(const unordered_set<int>& sset, bool safe = 0){
	if(safe == 1){
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	double sum = 0.0;
	for(int i=0; i < nrClusters; i++){
		sum += pow(preCompute[i],a);
	}
	return sum / normalization;	
}

inline double clusteredDiversity::evalGainsadd(const unordered_set<int>& sset, int item){
	/*if(sset.find(item) != sset.end()){
		cout<<"Error in using evalGainsadd: the provided item already belongs to the subset\n";
		return 0;
	}
	double gains = 0;
	double* maxval = new double[n];
	unordered_set<int>::const_iterator it;
	for(int i=0;i<n;i++){
		maxval[i] = 0;
		for( it = sset.begin(); it != sset.end(); it++ ){
			if(kernel[i][*it] > maxval[i]){
				maxval[i] = kernel[i][*it];
			}
		}
		if (maxval[i] < kernel[i][item]){
			gains+=(kernel[i][item]-maxval[i]);
		}
	}
	return gains; */
	throw std::runtime_error("To implement... clusteredDiversity::evalGainsadd");
	return 0.0;
}

inline double clusteredDiversity::evalGainsremove(const unordered_set<int>& sset, int item){
	/*if(sset.find(item) == sset.end()){
		cout<<"Error in using evalGainsremove: the provided item does not belong to the subset\n";
		return 0;
	}
	double sum = 0;
	double sumd = 0;
	double maxvali;
	unordered_set<int>::const_iterator it;
	for(int i=0;i<n;i++){
		maxvali = 0;
		for( it = sset.begin(); it != sset.end(); it++ ){
			if(kernel[i][*it] > maxvali){
				maxvali = kernel[i][*it];
			}
		}
		sum+=maxvali;
		if(*it != item){ sumd+=maxvali;}
	}
	return (sum - sumd); */
		throw std::runtime_error("To implement... clusteredDiversity::evalGainsremove");
	return 0.0;
}

inline double clusteredDiversity::evalGainsaddFast(const unordered_set<int>& sset, int item, bool safe = 0){
// Fast evaluation of Adding gains using the precomputed statistics. This is used, for example, in the implementation of the forward greedy algorithm.
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item does not belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) != sset.end()){
			cout<<"Error in using evalGainsaddFast: the provided item already belongs to the subset\n";
			return 0;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	
	double gains=0;
	int cluster = clusterAssignments[item] - 1;
	gains = pow(preCompute[cluster] + rewards[item], a) - pow(preCompute[cluster], a);
	
	return gains / normalization;	
}

inline double clusteredDiversity::evalGainsremoveFast(const unordered_set<int>& sset, int item, bool safe = 0){
// Fast evaluation of Removing gains using the precomputed statistics. This is used, for example, in the implementation of the reverse greedy algorithm.
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) == sset.end()){
			cout<<"Error in using evalGainsremoveFast: the provided item does not belong to the subset\n";
			return 0;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	
	double gains = 0;
	int cluster = clusterAssignments[item] - 1;
	gains = pow(preCompute[cluster], a) - pow(preCompute[cluster] - rewards[item], a);
	return gains;
}

inline void clusteredDiversity::updateStatisticsAdd(const unordered_set<int>& sset, int item, bool safe = 0){
// Update statistics for algorithms sequentially adding elements (for example, the greedy algorithm).
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item does not belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
        if (safe == 1){
		if(sset.find(item) != sset.end()){
			cout<<"Error in using updateStatisticsAdd: the provided item belongs to the sset already\n";
			return;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	
	int cluster = clusterAssignments[item] - 1;
	preCompute[cluster] += rewards[item];
	
	preComputeSet.insert(item);
}

inline void clusteredDiversity::updateStatisticsRemove(const unordered_set<int>& sset, int item, bool safe = 0){
// Update statistics for algorithms sequentially removing elements (for example, the reverse greedy algorithm).
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) == sset.end()){
			std::cout<<"Error in using updateStatisticsRemove: the provided item does not belong to the subset\n";
			return;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}

	int cluster = clusterAssignments[item] - 1;
	preCompute[cluster] -= rewards[item];
	
	preComputeSet.erase(item);
}

inline void clusteredDiversity::clearpreCompute() {
	for (int i = 0; i < sizepreCompute; i++) { 
		preCompute[i] = 0; 
	}
	preComputeSet.clear();
}

inline void clusteredDiversity::setpreCompute(const unordered_set<int>& sset){
	clearpreCompute();
	unordered_set<int>::const_iterator it;
	for( it = sset.begin(); it != sset.end(); it++ ){
		updateStatisticsAdd(sset, *it);
	}
}

#endif