/*
 *	Header file for defining feature based functions (essentially sums of concave over modular functions): f(X) = \sum_{f \in F} w_f g(m_f(X)), where g is a concave function, {m_f}_{f \in F} are a set of feature scores, and f \in F are features.
	Author: Rishabh Iyer.
	Melodi Lab, University of Washington, Seattle
 *
 */

#ifndef COSINESIMILARITY_H
#define COSINESIMILARITY_H

class cosineSimilarity: public submodularFunctions{
private:
	double* sim;
	
protected:
	int nFeatures; // Number of Features |F|
	double *preCompute; // Precomputed statistics. For a set X, p_X(f) = m_f(X). 
	unordered_set<int> preComputeSet; // This points to the preComputed Set for which the statistics p_X is calculated.
	concaveFunction *concave_function;
	
public:
	// Functions
	cosineSimilarity(std::unordered_set<int>& groundSet, std::vector<std::vector<double>> &sim);
	~cosineSimilarity();

	double eval(const unordered_set<int>& sset);
	double evalFast(const unordered_set<int>& sset, bool safe);
	double evalGainsadd(const unordered_set<int>& sset, int item);
	double evalGainsremove(const unordered_set<int>& sset, int item);
	double evalGainsaddFast(const unordered_set<int>& sset, int item, bool safe);
	double evalGainsremoveFast(const unordered_set<int>& sset, int item, bool safe);
	void updateStatisticsAdd(const unordered_set<int>& sset, int item, bool safe);
	void updateStatisticsRemove(const unordered_set<int>& sset, int item, bool safe);
	void setpreCompute(const unordered_set<int>& sset);
	void clearpreCompute();
};

inline cosineSimilarity::cosineSimilarity(std::unordered_set<int>& groundSet, std::vector<std::vector<double>> &sim): submodularFunctions(groundSet) {
	nFeatures = sim.size();
	
	preCompute = new double[nFeatures];
	for (int i = 0; i < nFeatures; i++)
		preCompute[i] = 0.0;
	this->sim = new double[nFeatures*n];
	
	for (int f = 0; f < nFeatures; f++) {
		if (sim[f].size() != n)
			throw std::runtime_error("cosineSimilarity::cosineSimilarity Invalid size of kernel.");
		for (int i = 0; i < n; i++) {
			this->sim[f*n + i] = sim[f][i];
		}
	}
}

inline cosineSimilarity::~cosineSimilarity(){
	delete[] preCompute;
	delete[] sim;
}


inline double cosineSimilarity::eval(const unordered_set<int>& sset){
	// Evaluation of function valuation.
	double sum = 0;
	unordered_set<int>::const_iterator it;	
	int offset = 0;
	std::vector<double> featurescoresset(nFeatures, 0.0);
	for (int j = 0; j < nFeatures; j++){
		for( it = sset.begin(); it != sset.end(); it++ ){
			featurescoresset[j] += sim[offset + *it];  
		}
		offset += n;
	}
	double val;
	for (int ii=0; ii<nFeatures; ii++){
		val = featurescoresset[ii];
		if (val < 1.0) 
			sum += val;
		else
			sum += 1.0;
	}
	return sum;
}

inline double cosineSimilarity::evalFast(const unordered_set<int>& sset, bool safe = 0){
// Evaluation of function valuation.
	if(safe == 1){
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
		
	double sum = 0;
	double val;
	for (int ii=0; ii<nFeatures; ii++) {
		val = preCompute[ii];
		if (val < 1.0)
			sum += val;
		else
			sum += 1.0;
	}
	return sum;
}

inline double cosineSimilarity::evalGainsadd(const unordered_set<int>& sset, int item){
// Evaluation of gains.
	if(sset.find(item) != sset.end()){
		cout<<"Error in using evalGainsadd: the provided item already belongs to the subset\n";
		return 0;
	}
	double gains = 0.0; 
	double temp;
	double diff;
	unordered_set<int>::const_iterator it;
	int offset = 0;
	std::vector<double> featurescoresset(nFeatures, 0.0);
	for (int j = 0; j < nFeatures; j++) { 
		for( it = sset.begin(); it != sset.end(); it++ ){
			featurescoresset[j] += sim[offset + *it];  
		}
		offset += n;
	}
	
	int index = item;
	for (int j = 0; j < nFeatures; j++) {
		temp = featurescoresset[j];
		if (temp >= 1.0)
			continue;
		
		temp += sim[index];
		if (temp < 1.0)  
			gains += temp - featurescoresset[j];
		else
			gains = 1.0 - featurescoresset[j];
		
		index += n;
	}
		
	return gains;
}

inline double cosineSimilarity::evalGainsremove(const unordered_set<int>& sset, int item){
// Evaluation of function valuation.
	/*if(sset.find(item) == sset.end()){
		cout<<"Error in using evalGainsremove: the provided item does not belong to the subset\n";
		return 0;
	}
	double sum = 0;
	double sumd = 0;
	double* featurescoresset = new double[nFeatures];
	double* featurescoressetd = new double[nFeatures];
	unordered_set<int>::const_iterator it;
	for (int i = 0; i < nFeatures; i++) { featurescoresset[i] = 0; featurescoressetd[i] = 0;}
	for( it = sset.begin(); it != sset.end(); it++ ){
		for (int j = 0; j < feats[*it].num_uniq_wrds; j++){
			featurescoresset[feats[*it].featureIndex[j]] += feats[*it].featureVec[j];  
			if(*it != item)  featurescoressetd[feats[*it].featureIndex[j]] += feats[*it].featureVec[j];
		}
	}
	for (int ii=0; ii<nFeatures; ii++){
		sum = sum + featureWeights[ii]*concave_function->eval(featurescoresset[ii]);
		sumd = sumd + featureWeights[ii]*concave_function->eval(featurescoressetd[ii]);	
	}
	return (sum - sumd); */
	throw std::runtime_error("Not implemented. cosineSimilarity::evalGainsremove");
}

inline double cosineSimilarity::evalGainsaddFast(const unordered_set<int>& sset, int item, bool safe = 0){
// Fast evaluation of Adding gains using the precomputed statistics. This is used, for example, in the implementation of the forward greedy algorithm.
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item does not belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) != sset.end()){
			cout<<"Error in using evalGainsaddFast: the provided item already belongs to the subset\n";
			return 0;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}

	double gains = 0.0; 
	double again = 0.0;
	double temp;
	
	int index = item;
	double newVal;
	for (int j = 0; j < nFeatures; j++) {
		temp = preCompute[j];
		if (temp < 1.0)
			;
		else
			temp = 1.0;
		newVal = preCompute[j] + sim[index];
		if (newVal < 1.0) 
			;
		else 
			newVal = 1.0;
		
		gains += newVal - temp;
		
		index += n;
	}
	//std::cout << "gains: " << gains << std::endl;
	
	return gains;
}

inline double cosineSimilarity::evalGainsremoveFast(const unordered_set<int>& sset, int item, bool safe = 0){
// Fast evaluation of Removing gains using the precomputed statistics. This is used, for example, in the implementation of the reverse greedy algorithm.
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) == sset.end()){
			cout<<"Error in using evalGainsremoveFast: the provided item does not belong to the subset\n";
			return 0;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}

	double gains = 0.0; 
	double again = 0.0;
	double temp;
	
	int index = item;
	double newVal;
	for (int j = 0; j < nFeatures; j++) {
		temp = preCompute[j];
		if (temp < 1.0)
			;
		else
			temp = 1.0;
		newVal = preCompute[j] - sim[index];
		if (newVal < 1.0) 
			;
		else 
			newVal = 1.0;
		
		gains += temp - newVal;
		
		index += n;
	}
	//std::cout << "gains: " << gains << std::endl;
	
	return gains;
}

inline void cosineSimilarity::updateStatisticsAdd(const unordered_set<int>& sset, int item, bool safe = 0){
// Update statistics for algorithms sequentially adding elements (for example, the greedy algorithm).
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item does not belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) != sset.end()){
			cout<<"Error in using updateStatisticsAdd: the provided item belongs to the sset already\n";
			return;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	
	int index = item;
	for (int f = 0; f < nFeatures; f++) {
		preCompute[f] += sim[index];		
		index += n;
	}

	preComputeSet.insert(item);
}

inline void cosineSimilarity::updateStatisticsRemove(const unordered_set<int>& sset, int item, bool safe = 0){
// Update statistics for algorithms sequentially removing elements (for example, the reverse greedy algorithm).
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) == sset.end()){
			cout<<"Error in using updateStatisticsRemove: the provided item does not belong to the subset\n";
			return;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	
	int index = item;
	for (int f = 0; f < nFeatures; f++) {
		preCompute[f] -= sim[index];		
		index += n;
	}
        
	preComputeSet.erase(item); 
}

inline void cosineSimilarity::clearpreCompute(){
	for (int i = 0; i < nFeatures; i++) { preCompute[i] = 0; }
	preComputeSet.clear();
}

inline void cosineSimilarity::setpreCompute(const unordered_set<int>& sset){
	clearpreCompute();
	unordered_set<int>::const_iterator it;
	for( it = sset.begin(); it != sset.end(); it++ ){
		updateStatisticsAdd(sset, *it);
	}
}


#endif