/*
 *	Header file for defining feature based functions (essentially sums of concave over modular functions): f(X) = \sum_{f \in F} w_f g(m_f(X)), where g is a concave function, {m_f}_{f \in F} are a set of feature scores, and f \in F are features.
	Author: Rishabh Iyer.
	Melodi Lab, University of Washington, Seattle
 *
 */

#ifndef FEATUREBASEDFUNCTIONS_H
#define FEATUREBASEDFUNCTIONS_H

#ifndef __FEATUREREP__
#define __FEATUREREP__
struct Feature{ //Stores the feature vector for each item in the groundset
     long int index; // index of the item
     int num_uniq_wrds; // number of non-zero enteries in the feature vector
     int* featureIndex; // Indices which are non-zero (generally sparse)
     float* featureVec; // score of the features present.
     int tot_num_wrds;
};
#endif

class concaveFunction;

// Helper Function to read in the Feature based functions.
inline int line2words(char *s, struct Feature * Feature, vector<double>& feature_weights){
	int digitwrd[16000];
	float featureval[16000];
	int pos = 0;
	int a = 0;
	while (sscanf(s,"%d %f %n",&digitwrd[a], &featureval[a], &pos) == 2) {
	 s += pos;
	 a++;
	}
	for (int i=0; i<a; i++){
		feature_weights[digitwrd[i]] += featureval[i];
	}
	Feature->featureIndex = new int[a];
	Feature->featureVec = new float[a];
	memcpy(Feature->featureIndex, digitwrd, sizeof(int)*a);
	memcpy(Feature->featureVec, featureval, sizeof(float)*a);
	return a;
}

class featureBasedFunctions: public submodularFunctions{
private:
	double* featurescoresset;
	
	protected:
	int nFeatures; // Number of Features |F|
	int type; //type: type of concave function, 1: sqrt over modular, 2: inverse function, 3: Log function, 4: Whatever else you want, Default: sqrt over modular
	vector<struct Feature> feats; // structure of the feature vectors for items. Size = n (groundset size)
	vector<double> featureWeights; // Feature Weights (w_f)
	vector<double> preCompute; // Precomputed statistics. For a set X, p_X(f) = m_f(X). 
	unordered_set<int> preComputeSet; // This points to the preComputed Set for which the statistics p_X is calculated.
	int sizepreCompute;// size of the precompute statistics (in this case, |F|).
	concaveFunction *concave_function;
	
	public:
	// Functions
	featureBasedFunctions(unordered_set<int>& groundSet, int type, char* featureFile, concaveFunction *concave_function);
	featureBasedFunctions(unordered_set<int>& groundSet, int type, vector<struct Feature> feats, vector<double> featureWeights, concaveFunction *concave_function);
	~featureBasedFunctions();
	void readFeaturefunction(char* featureFile);
	double eval(const unordered_set<int>& sset);
	double evalFast(const unordered_set<int>& sset, bool safe);
	double evalGainsadd(const unordered_set<int>& sset, int item);
	double evalGainsremove(const unordered_set<int>& sset, int item);
	double evalGainsaddFast(const unordered_set<int>& sset, int item, bool safe);
	double evalGainsremoveFast(const unordered_set<int>& sset, int item, bool safe);
	void updateStatisticsAdd(const unordered_set<int>& sset, int item, bool safe);
	void updateStatisticsRemove(const unordered_set<int>& sset, int item, bool safe);
	void setpreCompute(const unordered_set<int>& sset);
	void clearpreCompute();
};

// TODO: Move to separate include file
class concaveFunction {
public:
	virtual double eval(double z) = 0;
};

class concaveFunctionSquareRoot : public concaveFunction {
public:
	inline double eval(double z) { 
		return sqrt(z);
	}
};

class concaveFunctionPower : public concaveFunction {
private:
	double a = 0.5;
public:
	concaveFunctionPower(double a) : a(a) {} ;
	inline double eval(double z) { 
		return pow(z, a);
	}
};

class concaveFunctionMin : public concaveFunction {
public:
	inline double eval(double z) { 
	      double val = z;
	      if (val < 1.0) // This allows to deal with NaNs 
	          val = z;
	      else
	      	  val = 1.0;
	      return val;
	}
};

inline featureBasedFunctions::featureBasedFunctions(unordered_set<int>& groundSet, int type, char* featureFile, concaveFunction *concave_functionT): submodularFunctions(groundSet), type(type){
	nFeatures = 200000;
	featureWeights.resize(nFeatures);
	readFeaturefunction(featureFile);
	sizepreCompute = nFeatures;
	preCompute.resize(sizepreCompute); 	
	concave_function = concave_functionT;
	
	featurescoresset = new double[nFeatures];
}

inline featureBasedFunctions::featureBasedFunctions(unordered_set<int>& groundSet, int type, vector<struct Feature> feats, vector<double> featureWeightsT, concaveFunction *concave_functionT): submodularFunctions(groundSet), type(type), feats(feats), featureWeights(featureWeightsT){
	nFeatures = 200000;
	featureWeights.resize(nFeatures);
	sizepreCompute = nFeatures;
	preCompute.resize(sizepreCompute);
	concave_function = concave_functionT;
	
	featurescoresset = new double[nFeatures];
}

inline featureBasedFunctions::~featureBasedFunctions(){}

// A specific implementation to read in a feature file.
inline void featureBasedFunctions::readFeaturefunction(char* featureFile){
// read the feature based function
	ifstream iFile;
	FILE *fp = NULL;
	char line2[300000]; //stores information in each line of input
	printf("Reading feature File from %s...\n", featureFile);
	if ((fp = fopen(featureFile, "rt")) == NULL){
	    printf("Error: Cannot open file %s", featureFile);
	    exit(-1);
	}
	long int lineno = 0;
	while ( fgets(line2,sizeof(line2),fp) != NULL){
	    feats.push_back(Feature());     
	    feats[lineno].index = lineno;
	    feats[lineno].num_uniq_wrds = line2words(line2, &feats[lineno], featureWeights);//line2words transforms input with fmt digwords:featurevals into initialization of structure "Feature"    
	    lineno ++;
	}
	fclose(fp);
	if (n != lineno){
	    printf("error: number of points doesn't match with input number of points\n");
	    exit(-1);
	}
	cout<<"done with reading the feature based file\n";
}

/*inline double featureBasedFunctions::concave_function(double K){
	switch(type)
	   {
	   case 1 :
	      return sqrt(K);
	   case 2 :
	      return (1-1/(K+1));
	   case 3 :
	      return log(1 + K);
	   case 4 : {
	      double val = K;
	      if (val < 1.0) // This allows to deal with NaNs 
	          val = K;
	      else
	      	  val = 1.0;
	       return val; }
	   default :
	      return sqrt(K);
	   }
}*/

inline double featureBasedFunctions::eval(const unordered_set<int>& sset){
// Evaluation of function valuation.
	double sum = 0;
	unordered_set<int>::const_iterator it;
	for (int i = 0; i < nFeatures; i++) { featurescoresset[i] = 0; }
	
	int *indices;
	float *featureVec;
	int num_uniq_wrds;
	for( it = sset.begin(); it != sset.end(); it++ ){
		indices = feats[*it].featureIndex;
		featureVec = feats[*it].featureVec;
		num_uniq_wrds = feats[*it].num_uniq_wrds;
		for (int j = 0; j < num_uniq_wrds; j++){
			featurescoresset[indices[j]] += featureVec[j];   
		}
	}
	for (int ii=0; ii<nFeatures; ii++){
		sum += featureWeights[ii]*concave_function->eval(featurescoresset[ii]);
	}
	// TODO: MOVE TO DESTRUCTOR: delete[] featurescoresset;
	return sum;
}

inline double featureBasedFunctions::evalFast(const unordered_set<int>& sset, bool safe = 0){
// Evaluation of function valuation.
	if(safe == 1){
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
		
	double sum = 0;
	for (int ii=0; ii<nFeatures; ii++){
	sum = sum + featureWeights[ii]*concave_function->eval(preCompute[ii]);
	}
	return sum;
}

inline double featureBasedFunctions::evalGainsadd(const unordered_set<int>& sset, int item){
// Evaluation of gains.
	if(sset.find(item) != sset.end()){
		cout<<"Error in using evalGainsadd: the provided item already belongs to the subset\n";
		return 0;
	}
	double gains = 0; 
	double temp;
	double diff;
	double* featurescoresset = new double[nFeatures];
	unordered_set<int>::const_iterator it;
	for (int i = 0; i < nFeatures; i++) { featurescoresset[i] = 0; }
	for( it = sset.begin(); it != sset.end(); it++ ){
	for (int j = 0; j < feats[*it].num_uniq_wrds; j++){
		featurescoresset[feats[*it].featureIndex[j]] += feats[*it].featureVec[j];   
	}
	}
	for (int i=0; i<feats[item].num_uniq_wrds; i++){
	temp = preCompute[feats[item].featureIndex[i]] + feats[item].featureVec[i];
	diff = concave_function->eval(temp) - concave_function->eval(preCompute[feats[item].featureIndex[i]]);
	gains += featureWeights[feats[item].featureIndex[i]] * diff;
	}
	return gains;
}

inline double featureBasedFunctions::evalGainsremove(const unordered_set<int>& sset, int item){
// Evaluation of function valuation.
	if(sset.find(item) == sset.end()){
		cout<<"Error in using evalGainsremove: the provided item does not belong to the subset\n";
		return 0;
	}
	double sum = 0;
	double sumd = 0;
	double* featurescoresset = new double[nFeatures];
	double* featurescoressetd = new double[nFeatures];
	unordered_set<int>::const_iterator it;
	for (int i = 0; i < nFeatures; i++) { featurescoresset[i] = 0; featurescoressetd[i] = 0;}
	for( it = sset.begin(); it != sset.end(); it++ ){
		for (int j = 0; j < feats[*it].num_uniq_wrds; j++){
			featurescoresset[feats[*it].featureIndex[j]] += feats[*it].featureVec[j];  
			if(*it != item)  featurescoressetd[feats[*it].featureIndex[j]] += feats[*it].featureVec[j];
		}
	}
	for (int ii=0; ii<nFeatures; ii++){
		sum = sum + featureWeights[ii]*concave_function->eval(featurescoresset[ii]);
		sumd = sumd + featureWeights[ii]*concave_function->eval(featurescoressetd[ii]);	
	}
	return (sum - sumd);
}

inline double featureBasedFunctions::evalGainsaddFast(const unordered_set<int>& sset, int item, bool safe = 0){
// Fast evaluation of Adding gains using the precomputed statistics. This is used, for example, in the implementation of the forward greedy algorithm.
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item does not belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) != sset.end()){
			cout<<"Error in using evalGainsaddFast: the provided item already belongs to the subset\n";
			return 0;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	double gains = 0; 
	double temp;
	double diff;
	int num_wrds = feats[item].num_uniq_wrds;
	gains = 0;;
	for (int i=0; i<num_wrds; i++){
		temp = preCompute[feats[item].featureIndex[i]] + feats[item].featureVec[i];
		diff = concave_function->eval(temp) - concave_function->eval(preCompute[feats[item].featureIndex[i]]);
		gains += featureWeights[feats[item].featureIndex[i]] * diff;
	}
	return gains;
}

inline double featureBasedFunctions::evalGainsremoveFast(const unordered_set<int>& sset, int item, bool safe = 0){
// Fast evaluation of Removing gains using the precomputed statistics. This is used, for example, in the implementation of the reverse greedy algorithm.
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) == sset.end()){
			cout<<"Error in using evalGainsremoveFast: the provided item does not belong to the subset\n";
			return 0;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	double gains = 0; 
	double temp;
	double diff;
	int num_wrds = feats[item].num_uniq_wrds;
	gains = 0;;
	for (int i=0; i<num_wrds; i++){
		temp = preCompute[feats[item].featureIndex[i]] - feats[item].featureVec[i];
		diff = concave_function->eval(preCompute[feats[item].featureIndex[i]]) - concave_function->eval(temp);
		gains += featureWeights[feats[item].featureIndex[i]] * diff;
	}
	return gains;
}

inline void featureBasedFunctions::updateStatisticsAdd(const unordered_set<int>& sset, int item, bool safe = 0){
// Update statistics for algorithms sequentially adding elements (for example, the greedy algorithm).
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item does not belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) != sset.end()){
			cout<<"Error in using updateStatisticsAdd: the provided item belongs to the sset already\n";
			return;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	for (int i = 0; i < feats[item].num_uniq_wrds; i++){
                preCompute[feats[item].featureIndex[i]] += feats[item].featureVec[i];
        }
	preComputeSet.insert(item);
}

inline void featureBasedFunctions::updateStatisticsRemove(const unordered_set<int>& sset, int item, bool safe = 0){
// Update statistics for algorithms sequentially removing elements (for example, the reverse greedy algorithm).
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) == sset.end()){
			cout<<"Error in using updateStatisticsRemove: the provided item does not belong to the subset\n";
			return;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	for (int i = 0; i < feats[item].num_uniq_wrds; i++){
                preCompute[feats[item].featureIndex[i]] -= feats[item].featureVec[i];
        }
	preComputeSet.erase(item);
}

inline void featureBasedFunctions::clearpreCompute(){
	for (int i = 0; i < sizepreCompute; i++) { preCompute[i] = 0; }
	preComputeSet.clear();
	}

inline void featureBasedFunctions::setpreCompute(const unordered_set<int>& sset){
	clearpreCompute();
	unordered_set<int>::const_iterator it;
	for( it = sset.begin(); it != sset.end(); it++ ){
		updateStatisticsAdd(sset, *it);
	}
}


#endif