/*
 *	Header file for defining the facility location function: f(X) = \sum_{i \in V} \max_{j \in X} s_{ij}.
	Assumes that the kernel S is symmetric.
	Author: Rishabh Iyer.
	Email: rkiyer@uw.edu
	Melodi Lab, University of Washington, Seattle
 *
 */

#ifndef FIDELITYFUNCTIONS_H
#define FIDELITYFUNCTIONS_H

class fidelityFunctions: public submodularFunctions{
private:
	double *bow;
	double threshold;
	int nrWords;
	
	void computeSumBow(const std::unordered_set<int> &sset, std::vector<double> &sumBow);
	
protected:
	unordered_set<int> preComputeSet; // This points to the preComputed Set for which the statistics p_X is calculated.
	
	double* precomputedImageFidelities;
	double* precomputedSumBow;
	
	public:
	// Functions
	fidelityFunctions(std::unordered_set<int>& groundSet, std::vector<std::vector<double>> &bow, double threshold);
	double eval(const unordered_set<int>& sset);
	double evalFast(const unordered_set<int>& sset, bool safe);
	double evalGainsadd(const unordered_set<int>& sset, int item);
	double evalGainsremove(const unordered_set<int>& sset, int item);
	double evalGainsaddFast(const unordered_set<int>& sset, int item, bool safe);
	double evalGainsremoveFast(const unordered_set<int>& sset, int item, bool safe);
	void updateStatisticsAdd(const unordered_set<int>& sset, int item, bool safe);
	void updateStatisticsRemove(const unordered_set<int>& sset, int item, bool safe);
	void setpreCompute(const unordered_set<int>& ssett);
	void clearpreCompute();
};

inline fidelityFunctions::fidelityFunctions(std::unordered_set<int>& groundSet, std::vector<std::vector<double>> &bow, double threshold): submodularFunctions(groundSet), threshold(threshold) {
	// TODO: Set up proper precomputation stuff
	nrWords = bow.size();
	
	precomputedImageFidelities = new double[n];
	precomputedSumBow = new double[nrWords];
	
	this->bow = new double[nrWords*n];
	for (int w = 0; w < nrWords; w++) {
		int offset = w*n;
		for (int i = 0; i < n; i++) {
			this->bow[offset+i] = bow[w][i];
		}
	}
}

void fidelityFunctions::computeSumBow(const std::unordered_set<int> &sset, std::vector<double> &sumBow) {
	std::unordered_set<int>::const_iterator it;
	for (int j = 0; j < nrWords; j++) {
		int offset = j*n;
		for (it = sset.begin(); it != sset.end(); it++) {
			sumBow[j] += bow[offset + *it];
		}
		
	}
}

inline double fidelityFunctions::eval(const std::unordered_set<int>& sset){
	// sumBow = sum(bow(:, S), 2);
	std::vector<double> sumBow(nrWords, 0.0);
	computeSumBow(sset, sumBow);
	
	// sim = min(repmat(sumBow, 1, 100), bow);
	// val = min(sum(sim, 1), alpha*ones(1,100));
	std::vector<double> sum(n);
	for (int j = 0; j < nrWords; j++) {
		double sumBowVal = sumBow[j];
		for (int i = 0; i < n; i++) {
			if (sumBowVal < bow[j*n+i])
				sum[i] += sumBowVal;
			else
				sum[i] += bow[j*n + i];
		}
	}
	
	// val = sum(val) / (100*alpha);
	double val = 0;
	for (int j = 0; j < n; j++) {
		if (sum[j] < threshold)
			val += sum[j];
		else
			val += threshold;
	}
	
	return val;
}

inline double fidelityFunctions::evalFast(const unordered_set<int>& sset, bool safe = 0){
	if(safe == 1) {
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	
	double sum = 0;
	for(int i=0; i < n; i++) {
		sum += precomputedImageFidelities[i];
	}
	return sum; 
}

inline double fidelityFunctions::evalGainsadd(const unordered_set<int>& sset, int item){
	/*if(sset.find(item) != sset.end()){
		cout<<"Error in using evalGainsadd: the provided item already belongs to the subset\n";
		return 0;
	}
	double gains = 0;
	double* maxval = new double[n];
	unordered_set<int>::const_iterator it;
	for(int i=0;i<n;i++){
		maxval[i] = 0;
		for( it = sset.begin(); it != sset.end(); it++ ){
			if(kernel[i][*it] > maxval[i]){
				maxval[i] = kernel[i][*it];
			}
		}
		if (maxval[i] < kernel[i][item]){
			gains+=(kernel[i][item]-maxval[i]);
		}
	}
	return gains;*/
	throw std::runtime_error("Not implemented. fidelityFunctions::evalGainsadd");
}

inline double fidelityFunctions::evalGainsremove(const unordered_set<int>& sset, int item){
	/*if(sset.find(item) == sset.end()){
		cout<<"Error in using evalGainsremove: the provided item does not belong to the subset\n";
		return 0;
	}
	double sum = 0;
	double sumd = 0;
	double maxvali;
	unordered_set<int>::const_iterator it;
	for(int i=0;i<n;i++){
		maxvali = 0;
		for( it = sset.begin(); it != sset.end(); it++ ){
			if(kernel[i][*it] > maxvali){
				maxvali = kernel[i][*it];
			}
		}
		sum+=maxvali;
		if(*it != item){ sumd+=maxvali;}
	}
	return (sum - sumd);*/
	throw std::runtime_error("Not implemented. fidelityFunctions::evalGainsremove");
}

inline double fidelityFunctions::evalGainsaddFast(const unordered_set<int>& sset, int item, bool safe = 0){
// Fast evaluation of Adding gains using the precomputed statistics. This is used, for example, in the implementation of the forward greedy algorithm.
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item does not belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) != sset.end()){
			cout<<"Error in using evalGainsaddFast: the provided item already belongs to the subset\n";
			return 0;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	
	
	double gain = 0.0;
	
	/* for (int i = 0; i < n; i++) {
		if (precomputedImageFidelities[i] >= threshold) {
			continue;
		}
		
		double sum = 0.0;
		for (int j = 0; j < nrWords; j++) {
			if (newSumBow[j] < bow[j][i])
				sum += newSumBow[j];
			else
				sum += bow[j][i];
		}
		
		if (sum < threshold) 
			gain += (sum - precomputedImageFidelities[i]);
		else
			gain += (threshold - precomputedImageFidelities[i]);
	} */
	
	std::vector<double> sum_(n, 0.0);
	double tmp;
	int offset = 0;
	double *index = bow;
	for (int w = 0; w < nrWords; w++) {
		tmp = precomputedSumBow[w] + bow[offset+item];
		for (int i = 0; i < n; i++) {
			if (tmp < *index) 
				sum_[i] += tmp;
			else
				sum_[i] += *index;
			index++;
		}
		offset += n;
	}
	
	for (int i = 0; i < n; i++) {
		tmp = sum_[i];
		if (tmp < threshold) 
			gain += tmp - precomputedImageFidelities[i];
		else
			gain += threshold - precomputedImageFidelities[i];	
	}
	
	return gain; 
	
	/* double tmp;
	int offset;
	double potGain;
	for (int i = 0; i < n; i++) {
		if (precomputedImageFidelities[i] >= threshold)
			continue;
		
		offset = 0;
		for (int w = 0; w < nrWords; w++) {
			potGain = bow[offset+item];
			tmp = precomputedSumBow[w] + potGain;
			
			if (tmp < bow[offset + i])
				gain += potGain;
			else
				gain += bow[offset + i] - precomputedSumBow[w];
			
			offset += n;
		}
	}
	
	return gain; */
	
	
	
	/* double curVal = 0.0;
	for(int i=0; i < n; i++) {
		curVal += precomputedImageFidelities[i];
	}
	
	double newVal = 0.0;
	for (int i = 0; i < n; i++) {
		if (precomputedImageFidelities[i] >= threshold) {
			newVal += threshold;
			continue;
		}
		
		double sum = 0.0;
		for (int j = 0; j < nrWords; j++) {
			if (precomputedSumBow[j] + bow[j][item] < bow[j][i])
				sum += precomputedSumBow[j] + bow[j][item];
			else
				sum += bow[j][i];
		}
		
		if (sum < threshold) 
			newVal += sum;
		else
			newVal += threshold;
	}
	
	return (newVal - curVal); */
}

inline double fidelityFunctions::evalGainsremoveFast(const unordered_set<int>& sset, int item, bool safe = 0){
// Fast evaluation of Removing gains using the precomputed statistics. This is used, for example, in the implementation of the reverse greedy algorithm.
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) == sset.end()){
			cout<<"Error in using evalGainsremoveFast: the provided item does not belong to the subset\n";
			return 0;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	
	double gain = 0.0;
	std::vector<double> sum_(n, 0.0);
	double tmp;
	int offset = 0;
	double *index = bow;
	for (int w = 0; w < nrWords; w++) {
		tmp = precomputedSumBow[w] - bow[offset+item];
		for (int i = 0; i < n; i++) {
			if (tmp < *index) 
				sum_[i] += tmp;
			else
				sum_[i] += *index;
			index++;
		}
		offset += n;
	}
	
	for (int i = 0; i < n; i++) {
		tmp = sum_[i];
		if (tmp < threshold) 
			gain += tmp - precomputedImageFidelities[i];
		else
			gain += threshold - precomputedImageFidelities[i];	
	}
	
	return gain; 
}

inline void fidelityFunctions::updateStatisticsAdd(const unordered_set<int>& sset, int item, bool safe = 0){
// Update statistics for algorithms sequentially adding elements (for example, the greedy algorithm).
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item does not belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) != sset.end()){
			cout<<"Error in using updateStatisticsAdd: the provided item belongs to the sset already\n";
			return;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}


	/*for (int i = 0; i < n; i++) {
		if (precomputedImageFidelities[i] >= threshold)
			continue;
		
		double sum = 0.0;
		for (int j = 0; j < nrWords; j++) {
			if (precomputedSumBow[j] < bow[j][i])
				sum += precomputedSumBow[j];
			else
				sum += bow[j][i];
			if (sum >= threshold) 
				break;
		}
		
		if (sum < threshold) 
			precomputedImageFidelities[i] = sum;
		else
			precomputedImageFidelities[i] = threshold;
	} */
	
	std::vector<double> sum_(n, 0.0);
	double tmp;
	int offset = 0;
	for (int w = 0; w < nrWords; w++) {
		precomputedSumBow[w] += bow[offset+item];
		tmp = precomputedSumBow[w];
		for (int i = 0; i < n; i++) {
			if (tmp < bow[offset+i]) 
				sum_[i] += tmp;
			else
				sum_[i] += bow[offset+i];
		}
		offset += n;
	}
	
	double v;
	for (int i = 0; i < n; i++) {
		v = sum_[i];
		if (v > threshold) 
			v = threshold;
		
		precomputedImageFidelities[i] = v;
	}
	
	preComputeSet.insert(item);
}

inline void fidelityFunctions::updateStatisticsRemove(const unordered_set<int>& sset, int item, bool safe = 0){
// Update statistics for algorithms sequentially removing elements (for example, the reverse greedy algorithm).
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) == sset.end()){
			std::cout<<"Error in using updateStatisticsRemove: the provided item does not belong to the subset\n";
			return;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}

	std::vector<double> sum_(n, 0.0);
	double tmp;
	int offset = 0;
	for (int w = 0; w < nrWords; w++) {
		precomputedSumBow[w] -= bow[offset+item];
		tmp = precomputedSumBow[w];
		for (int i = 0; i < n; i++) {
			if (tmp < bow[offset+i]) 
				sum_[i] += tmp;
			else
				sum_[i] += bow[offset+i];
		}
		offset += n;
	}
	
	double v;
	for (int i = 0; i < n; i++) {
		v = sum_[i];
		if (v > threshold) 
			v = threshold;
		
		precomputedImageFidelities[i] = v;
	}
	
	preComputeSet.erase(item); 
}

inline void fidelityFunctions::clearpreCompute(){
	for (int i = 0; i < n; i++) 
		precomputedImageFidelities[i] = 0.0;
	for (int w = 0; w < nrWords; w++)
		precomputedSumBow[w] = 0.0;
	preComputeSet.clear();
}

inline void fidelityFunctions::setpreCompute(const unordered_set<int>& sset){
	clearpreCompute();
	unordered_set<int>::const_iterator it;
	for( it = sset.begin(); it != sset.end(); it++ ){
		updateStatisticsAdd(sset, *it);
	}
}

#endif