/*
 *	Header file for defining feature based functions (essentially sums of concave over modular functions): f(X) = \sum_{f \in F} w_f g(m_f(X)), where g is a concave function, {m_f}_{f \in F} are a set of feature scores, and f \in F are features.
	Author: Rishabh Iyer.
	Melodi Lab, University of Washington, Seattle
 *
 */

#ifndef HINGELOSS_H
#define HINGELOSS_H

class hingeLoss: public submodularFunctions{
private:
	double center;
	
public:
	// Functions
	hingeLoss(std::unordered_set<int>& groundSet, int center);
	~hingeLoss();

	double eval(const unordered_set<int>& sset);
	double evalFast(const unordered_set<int>& sset, bool safe);
	double evalGainsadd(const unordered_set<int>& sset, int item);
	double evalGainsremove(const unordered_set<int>& sset, int item);
	double evalGainsaddFast(const unordered_set<int>& sset, int item, bool safe);
	double evalGainsremoveFast(const unordered_set<int>& sset, int item, bool safe);
	void updateStatisticsAdd(const unordered_set<int>& sset, int item, bool safe);
	void updateStatisticsRemove(const unordered_set<int>& sset, int item, bool safe);
	void setpreCompute(const unordered_set<int>& sset);
	void clearpreCompute();
};

inline hingeLoss::hingeLoss(std::unordered_set<int>& groundSet, int center): submodularFunctions(groundSet), center((double) center) {
}

inline hingeLoss::~hingeLoss(){
}


inline double hingeLoss::eval(const unordered_set<int>& sset){
	// Evaluation of function valuation.	
	double size = (double) sset.size();
	return -(size - center)*(size-center);
}

inline double hingeLoss::evalFast(const unordered_set<int>& sset, bool safe = 0){
// Evaluation of function valuation.
	double size = (double) sset.size();
	return -(size - center)*(size-center);
}

inline double hingeLoss::evalGainsadd(const unordered_set<int>& sset, int item){
// Evaluation of gains.
	double size = (double) sset.size();
	double old = -(size - center)*(size-center);
	size++;
	double newv = -(size - center)*(size-center);
	return newv - old;
}

inline double hingeLoss::evalGainsremove(const unordered_set<int>& sset, int item){
// Evaluation of function valuation.
	double size = (double) sset.size();
	double old = -(size - center)*(size-center);
	size--;
	double newv = -(size - center)*(size-center);
	return newv - old;
}

inline double hingeLoss::evalGainsaddFast(const unordered_set<int>& sset, int item, bool safe = 0){
// Fast evaluation of Adding gains using the precomputed statistics. This is used, for example, in the implementation of the forward greedy algorithm.
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item does not belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	double size = (double) sset.size();
	double old = -(size - center)*(size-center);
	size++;
	double newv = -(size - center)*(size-center);
	return newv - old;
}

inline double hingeLoss::evalGainsremoveFast(const unordered_set<int>& sset, int item, bool safe = 0){
// Fast evaluation of Removing gains using the precomputed statistics. This is used, for example, in the implementation of the reverse greedy algorithm.
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	double size = (double) sset.size();
	double old = -(size - center)*(size-center);
	size--;
	double newv = -(size - center)*(size-center);
	return newv - old;
}

inline void hingeLoss::updateStatisticsAdd(const unordered_set<int>& sset, int item, bool safe = 0){
// Update statistics for algorithms sequentially adding elements (for example, the greedy algorithm).
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item does not belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.

}

inline void hingeLoss::updateStatisticsRemove(const unordered_set<int>& sset, int item, bool safe = 0){
// Update statistics for algorithms sequentially removing elements (for example, the reverse greedy algorithm).
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.

}

inline void hingeLoss::clearpreCompute(){
}

inline void hingeLoss::setpreCompute(const unordered_set<int>& sset){
}


#endif