/*
 *	Header file for defining feature based functions (essentially sums of concave over modular functions): f(X) = \sum_{f \in F} w_f g(m_f(X)), where g is a concave function, {m_f}_{f \in F} are a set of feature scores, and f \in F are features.
	Author: Rishabh Iyer.
	Melodi Lab, University of Washington, Seattle
 *
 */

#ifndef ROUGE_FUNCTIONS_H
#define ROUGE_FUNCTIONS_H

#include <algorithm>

#ifndef __FEATUREREP__
#define __FEATUREREP__
struct Feature{ //Stores the feature vector for each item in the groundset
     long int index; // index of the item
     int num_uniq_wrds; // number of non-zero enteries in the feature vector
     int* featureIndex; // Indices which are non-zero (generally sparse)
     float* featureVec; // score of the features present.
     int tot_num_wrds;
};
#endif

class concaveFunction;

class rougeFunctions: public submodularFunctions{
private:
	int nFeatures;
	double *bow;
	double* featurescoresset;
	
protected:
	vector<double> preCompute; // Precomputed statistics. For a set X, p_X(f) = m_f(X). 
	unordered_set<int> preComputeSet; // This points to the preComputed Set for which the statistics p_X is calculated.
	int sizepreCompute;// size of the precompute statistics (in this case, |F|).
	std::vector<concaveFunction*> concave_function;
	double normalization;
	
	public:
	// Functions
	rougeFunctions(unordered_set<int>& groundSet, vector<vector<double>> &bow, vector<unordered_set<int>> &summaries);
	~rougeFunctions();
	void computeConcaveFunctions(int n, vector<vector<double>> &bow, vector<unordered_set<int>> &summaries);
	double eval(const unordered_set<int>& sset);
	double evalFast(const unordered_set<int>& sset, bool safe);
	double evalGainsadd(const unordered_set<int>& sset, int item);
	double evalGainsremove(const unordered_set<int>& sset, int item);
	double evalGainsaddFast(const unordered_set<int>& sset, int item, bool safe);
	double evalGainsremoveFast(const unordered_set<int>& sset, int item, bool safe);
	void updateStatisticsAdd(const unordered_set<int>& sset, int item, bool safe);
	void updateStatisticsRemove(const unordered_set<int>& sset, int item, bool safe);
	void setpreCompute(const unordered_set<int>& sset);
	void clearpreCompute();
};


inline rougeFunctions::rougeFunctions(unordered_set<int>& groundSet, vector<vector<double>> &bow, vector<unordered_set<int>> &summaries): submodularFunctions(groundSet) {
	nFeatures = bow.size();
	computeConcaveFunctions(groundSet.size(), bow, summaries);
	sizepreCompute = nFeatures;
	preCompute.resize(sizepreCompute); 

	// copy bag-of-words	
	this->bow = new double[nFeatures*n];
	double *index = this->bow;
	for (int f = 0; f < nFeatures; f++) {
		for (int i = 0; i < n; i++) {
			*index = bow[f][i];
			index++;
		}
	}
	
	featurescoresset = new double[nFeatures];
}

inline rougeFunctions::~rougeFunctions() {
	delete[] bow;
	delete[] featurescoresset;
}

class concaveFunctionPiecewiseLinear : public concaveFunction {
private:
	std::vector<double> thresholds;
	std::vector<double> offsets;
	std::vector<int> counts;
	int n;
public:
	concaveFunctionPiecewiseLinear(std::vector<double> thresholds, std::vector<double> offsets, std::vector<int> counts) : thresholds(thresholds), offsets(offsets), counts(counts) { n = counts.size(); } ;
	inline double eval(double z) {
		int idx = 0;
		for (int k = 0; k < n; k++) {
			if (z < thresholds[k]) {
				idx = k - 1;
				break;
			}
		}
		double val = offsets[idx] + (z - thresholds[idx]) * counts[idx];
		return val;
	}
	
};

inline void rougeFunctions::computeConcaveFunctions(int n, vector<vector<double>> &bow, vector<unordered_set<int>> &summaries) {
	std::cout << "Computing concave functions..." << std::endl;
	double tol = 1e-12;
	
	// compute bow summaries
	int nrWords = bow.size();
	int nrSummaries = summaries.size();
	std::vector<std::vector<double>> bowSummaries(nrWords);
	for (int i = 0; i < nrWords; i++) {
		bowSummaries[i].resize(nrSummaries);
	}
	for (int s = 0; s < summaries.size(); s++) {
		unordered_set<int>::iterator it;
		for (it = summaries[s].begin(); it != summaries[s].end(); it++) {
			for (int w = 0; w < nrWords; w++) {
				bowSummaries[w][s] += bow[w][*it];
			}
		}
	}
	
	// compute normalization
	double normalization = 0.0;
	for (int i = 0; i < nrWords; i++) {
		for (int j = 0; j < nrSummaries; j++) {
			normalization += bowSummaries[i][j];
		}
	}
	this->normalization = normalization;
	
	// compute thresholds and counts
	std::vector<std::vector<double>> thresholds(nrWords);
	std::vector<std::vector<double>> offsets(nrWords);
	std::vector<std::vector<int>> counts(nrWords);
	
	concave_function.resize(nrWords);
	for (int w = 0; w < nrWords; w++) {
		std::vector<double> sThresholds(bowSummaries[w]);
		std::sort(sThresholds.begin(), sThresholds.end());
		
		std::list<double> tThresholds;
		std::list<double> tOffsets;
		std::list<int> tCounts;
		tOffsets.push_back(0.0);
		double lastVal = 0.0;
		int count = 0;
		for (int s = 0; s < nrSummaries; s++) {
			double curVal = sThresholds[s];
			if (abs(curVal - lastVal) < tol) {
				
			} else {
				tCounts.push_back(nrSummaries - count);
				tThresholds.push_back(lastVal);
				std::list<double>::iterator it = tOffsets.begin();
				std::advance(it, (tOffsets.size() - 1));
				tOffsets.push_back(*it + (nrSummaries - count) * (curVal - lastVal));
				lastVal = curVal;
				// count = 1;
			}
			count++;
		}
		tCounts.push_back(nrSummaries - count);
		tThresholds.push_back(lastVal);
		
		// for sanity
		tCounts.push_back(0);
		tThresholds.push_back(std::numeric_limits<double>::max());
		tOffsets.push_back(std::numeric_limits<double>::max());
		
		// convert to array
		thresholds[w].resize(tThresholds.size());
		offsets[w].resize(tOffsets.size());
		counts[w].resize(tCounts.size());
		for (int i = 0; i < tThresholds.size(); i++) {
			std::list<double>::iterator it1 = tThresholds.begin();
			std::list<double>::iterator it2 = tOffsets.begin();
			std::list<int>::iterator it3 = tCounts.begin();
			std::advance(it1, i);
			std::advance(it2, i);
			std::advance(it3, i);
			thresholds[w][i] = *it1;
			offsets[w][i] = *it2;
			counts[w][i] = *it3;
		}
		
		concaveFunction *cfct = new concaveFunctionPiecewiseLinear(thresholds[w], offsets[w], counts[w]);
		concave_function[w] = cfct;
	}
	
	std::cout << "COMPUTED." << std::endl;
}

inline double rougeFunctions::eval(const unordered_set<int>& sset){
	// Evaluation of function valuation.
	double sum = 0;
	unordered_set<int>::const_iterator it;

	int offset = 0;
	for (int j = 0; j < nFeatures; j++) {
		featurescoresset[j] = 0.0;
		for( it = sset.begin(); it != sset.end(); it++ ){
			featurescoresset[j] += bow[offset + *it];   
		}
		offset += n;
	}
	
	for (int ii=0; ii<nFeatures; ii++) {
		sum += concave_function[ii]->eval(featurescoresset[ii]);
	}
	
	return sum;
}

inline double rougeFunctions::evalFast(const unordered_set<int>& sset, bool safe = 0){
	// Evaluation of function valuation.
	if(safe == 1){
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
		
	double sum = 0;
	for (int ii=0; ii < nFeatures; ii++){
		sum += concave_function[ii]->eval(preCompute[ii]);
	}
	return sum;
}

inline double rougeFunctions::evalGainsadd(const unordered_set<int>& sset, int item){
	// Evaluation of gains.
	if(sset.find(item) != sset.end()){
		cout<<"Error in using evalGainsadd: the provided item already belongs to the subset\n";
		return 0;
	}
	double gains = 0; 
	double temp;
	double diff;
	unordered_set<int>::const_iterator it;

	int offset = 0;
	for (int j = 0; j < nFeatures; j++) {
		featurescoresset[j] = 0.0;
		for( it = sset.begin(); it != sset.end(); it++ ){
			featurescoresset[j] += bow[offset + *it];   
		}
		offset += n;
	}
	
	int index = item;
	for (int i = 0; i < nFeatures; i++) {
		temp = featurescoresset[i] + bow[index];
		
		diff = concave_function[i]->eval(temp) - concave_function[i]->eval(featurescoresset[i]);
		gains += diff;
		
		index += item;
	}
	
	return gains;
}

inline double rougeFunctions::evalGainsremove(const unordered_set<int>& sset, int item){
// Evaluation of function valuation.
	/*if(sset.find(item) == sset.end()){
		cout<<"Error in using evalGainsremove: the provided item does not belong to the subset\n";
		return 0;
	}
	double sum = 0;
	double sumd = 0;
	double* featurescoresset = new double[nFeatures];
	double* featurescoressetd = new double[nFeatures];
	unordered_set<int>::const_iterator it;
	for (int i = 0; i < nFeatures; i++) { featurescoresset[i] = 0; featurescoressetd[i] = 0;}
	for( it = sset.begin(); it != sset.end(); it++ ){
		for (int j = 0; j < feats[*it].num_uniq_wrds; j++){
			featurescoresset[feats[*it].featureIndex[j]] += feats[*it].featureVec[j];  
			if(*it != item)  featurescoressetd[feats[*it].featureIndex[j]] += feats[*it].featureVec[j];
		}
	}
	for (int ii=0; ii<nFeatures; ii++){
		sum = sum + concave_function[ii]->eval(featurescoresset[ii]);
		sumd = sumd + concave_function[ii]->eval(featurescoressetd[ii]);	
	}
	return (sum - sumd); */
	throw std::runtime_error("Not implemented. rougeFunctions::evalGainsremove");
}

inline double rougeFunctions::evalGainsaddFast(const unordered_set<int>& sset, int item, bool safe = 0){
// Fast evaluation of Adding gains using the precomputed statistics. This is used, for example, in the implementation of the forward greedy algorithm.
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item does not belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) != sset.end()){
			cout<<"Error in using evalGainsaddFast: the provided item already belongs to the subset\n";
			return 0;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	
	double gains = 0; 
	double temp;
	double diff;

	int index = item;
	for (int i = 0; i < nFeatures; i++) {
		temp = preCompute[i] + bow[index];
		
		diff = concave_function[i]->eval(temp) - concave_function[i]->eval(preCompute[i]);
		gains += diff;
		
		index += item;
	}
	
	return gains;
}

inline double rougeFunctions::evalGainsremoveFast(const unordered_set<int>& sset, int item, bool safe = 0){
// Fast evaluation of Removing gains using the precomputed statistics. This is used, for example, in the implementation of the reverse greedy algorithm.
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) == sset.end()){
			cout<<"Error in using evalGainsremoveFast: the provided item does not belong to the subset\n";
			return 0;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	
	double gains = 0; 
	double temp;
	double diff;

	int index = item;
	for (int i = 0; i < nFeatures; i++) {
		temp = preCompute[i] - bow[index];
		
		diff = concave_function[i]->eval(preCompute[i]) - concave_function[i]->eval(temp);
		gains += diff;
		
		index += item;
	}
	
	return gains;
}

inline void rougeFunctions::updateStatisticsAdd(const unordered_set<int>& sset, int item, bool safe = 0){
// Update statistics for algorithms sequentially adding elements (for example, the greedy algorithm).
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item does not belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) != sset.end()){
			cout<<"Error in using updateStatisticsAdd: the provided item belongs to the sset already\n";
			return;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	
	int index = item;
	for (int i = 0; i < nFeatures; i++) {
		preCompute[i] += bow[index];		
		index += item;
	}
        
	preComputeSet.insert(item);
}

inline void rougeFunctions::updateStatisticsRemove(const unordered_set<int>& sset, int item, bool safe = 0){
// Update statistics for algorithms sequentially removing elements (for example, the reverse greedy algorithm).
// The safety parameter 'safe' controls the safety and speed of this procedure. If not sure how to use it, set it to it's default value, viz. 1
// If safe = 0, (for the sake of speed) we do not check if item belong to the subset. We also do not check if the preComputed statistics actually points to the set sset. You would need to check both these.
	if (safe == 1){
		if(sset.find(item) == sset.end()){
			cout<<"Error in using updateStatisticsRemove: the provided item does not belong to the subset\n";
			return;
		}
		if(sset != preComputeSet)
			setpreCompute(sset);
	}
	
	int index = item;
	for (int i = 0; i < nFeatures; i++) {
		preCompute[i] -= bow[index];		
		index += item;
	}
        
	preComputeSet.erase(item);
}

inline void rougeFunctions::clearpreCompute(){
	for (int i = 0; i < sizepreCompute; i++) { preCompute[i] = 0; }
	preComputeSet.clear();
	}

inline void rougeFunctions::setpreCompute(const unordered_set<int>& sset){
	clearpreCompute();
	unordered_set<int>::const_iterator it;
	for( it = sset.begin(); it != sset.end(); it++ ){
		updateStatisticsAdd(sset, *it);
	}
}


#endif