/*
 *	A tutorial for SOTK.
 *	Author: Rishabh Iyer
 *	Email: rkiyer@uw.edu
 *	Melodi Lab, University of Washington, Seattle
 *
 *	Here, we run medium scale experiments using the TIMIT corpus, for speech recognition. 
 *	The size of the dataset is n = 4620.
 *	Our algorithms are tested on variants of datasubset selection problem for speech recognition
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <sys/stat.h>
#include <sys/types.h>
#include <set>
#include <unordered_set>
#include <algorithm>

#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "math.h"
#include "time.h"
#include <stdarg.h>
#include <assert.h>
#include <queue>
using namespace std;

// Utilities
#include "Utils/error.h"
#include "Utils/sort.h"
#include "Utils/utils.h"
#include "Utils/fileHandling.h"
#include "Utils/totalOrder.h"

//Submodular Functions
#include "Functions/setFunctions.h"
#include "Functions/submodularFunctions.h"
#include "Functions/modularFunctions.h"
#include "Functions/featureBasedFunctions.h"
#include "Functions/facilityLocation.h"
#include "Functions/graphCutFunctions.h"
#include "Functions/saturatedCoverage.h"
#include "Functions/complexityFunctions.h"
#include "Functions/confusabilityFunctions.h"
#include "Functions/combineSubmodularFunctions.h"
#include "Functions/scaleSubmodularFunctions.h"
#include "Functions/combineSubmodularModularFunctions.h"
#include "Functions/differenceSubmodularFunctions.h"
#include "Functions/complementSubmodularFunctions.h"
#include "Functions/truncateSubmodularFunctions.h"
#include "Functions/clusteredSubmodularFunctions.h"


//Algorithms
#include "Algorithms/lazyGreedyMax.h"
#include "Algorithms/lazyGreedySC.h"
#include "Algorithms/revlazyGreedyMax.h"
#include "Algorithms/bidirGreedyMax.h"
#include "Algorithms/localSearchMax.h"
#include "Algorithms/minNormPoint.h"
#include "Algorithms/naiveGreedyMax.h"
#include "Algorithms/mminCC.h"
#include "Algorithms/mminSCSC.h"
#include "Algorithms/mminSCSK.h"
#include "Algorithms/mminUC.h"
#include "Algorithms/mminSubSup.h"
#include "Algorithms/mminModMod.h"
#include "Algorithms/mminSupSub.h"
#define SMALLEST_NUMBER -1e70
#define LARGEST_NUMBER 1e70


int main(int argc, char** argv){
	int n; // GroundSet size
	unordered_set<int> groundSet = unordered_set<int>(); // GroundSet.
	unordered_set<int> emptyset = unordered_set<int>();
	int numWords;	
	cout<<"***************************************************************************\n";
	cout<<"Running Small Scale Experiments with TIMIT\n";
	cout<<"***************************************************************************\n";
	char* similarityFileTIMIT = "TestData/TIMITSimilarity";
	char* featureFileTIMIT = "TestData/TIMITFeature";
	char* transcriptFileTIMIT = "TestData/TIMITTranscript";
	n = 4620;
	numWords = 48;
	time_t start, end;
	double timetaken;
	for (int i = 0; i < n; ++i) 
		groundSet.insert(i);
	vector<double> costList(n, 1);	
	std::unordered_set<int> optSet;
	/*

			******************** Monotone Submodular Maximization *******************

	*/
	cout<<"\n";
	cout<<"First we do monotone submodular maximization using the Greedy Algorithms!\n";
	cout<<"\n";
	/*
		Facility Location Function
	*/
	cout<<"***************************************************************************\n";
	cout<<"Let's try the facility Location Function: Choose 5% of the data\n";
	double budget = 0.05*n;
	facilityLocation fL(groundSet, similarityFileTIMIT);
	// Lazy Greedy
	start = time(0);
	optSet = lazyGreedyMax(fL, costList, budget, emptyset, 0);
	end = time(0);
	timetaken = difftime(end, start);
	cout<<"Maximum obtained with the Lazy Greedy with Facility Location is "<< fL.evalFast(optSet)<<" and the time taken is "<<timetaken<<" seconds.\n";

	// Naive Greedy
	start = time(0);
	optSet = naiveGreedyMax(fL, costList, budget, emptyset, 0);
	end = time(0);
	timetaken = difftime(end, start);
	cout<<"Maximum obtained with the Naive Greedy with Facility Location is "<< fL.evalFast(optSet)<<" and the time taken is "<<timetaken<<" seconds.\n";
	
	// Random Subset
	srand(time(NULL));
	unordered_set<int> randomSet;
	for (int tt = 0; tt < budget; tt++){
		int randel = rand() % n;
		randomSet.insert(randomSet.end(), randel);
	}
	cout<<"The value of a random set with Facility Location of the same size is "<< fL.eval(randomSet)<<"\n";
	cout<<"\n";
	/*
		Saturated Coverage Function
	*/
	cout<<"***************************************************************************\n";
	cout<<"Let's try the Saturated Coverage Function: Choose 5% of the data, and alpha = 0.1.\n";
	budget = 0.05*n;
	saturatedCoverage sC(groundSet, similarityFileTIMIT, 0.1);
	// Lazy Greedy
	start = time(0);
	optSet = lazyGreedyMax(sC, costList, budget, emptyset, 0);
	end = time(0);
	timetaken = difftime(end, start);
	cout<<"Maximum obtained with the Lazy Greedy with Saturated Coverage is "<< sC.evalFast(optSet)<<" and the time taken is "<<timetaken<<" seconds.\n";

	// Naive Greedy
	start = time(0);
	optSet = naiveGreedyMax(sC, costList, budget, emptyset, 0);
	end = time(0);
	timetaken = difftime(end, start);
	cout<<"Maximum obtained with the Naive Greedy with Saturated Coverage is "<< sC.evalFast(optSet)<<" and the time taken is "<<timetaken<<" seconds.\n";

	// Random Subset
	srand(time(NULL));
	randomSet.clear();
	for (int tt = 0; tt < budget; tt++){
		int randel = rand() % n;
		randomSet.insert(randomSet.end(), randel);
	}
	cout<<"The value of a random set with Saturated Coverage of the same size is "<< sC.eval(randomSet)<<"\n";
	cout<<"\n";

	/*
		Feature Based Function
	*/
	cout<<"***************************************************************************\n";
	cout<<"Let's try the Feature Based Functions: Choose 5% of the data.\n";
	budget = 0.05*n;
	featureBasedFunctions ff(groundSet, 1, featureFileTIMIT);
	// Lazy Greedy
	start = time(0);
	optSet = lazyGreedyMax(ff, costList, budget, emptyset, 0);
	end = time(0);
	timetaken = difftime(end, start);
	cout<<"Maximum obtained with the Lazy Greedy with Feature Based Functions is "<< ff.evalFast(optSet)<<" and the time taken is "<<timetaken<<" seconds.\n";

	// Naive Greedy
	start = time(0);
	optSet = naiveGreedyMax(sC, costList, budget, emptyset, 0);
	end = time(0);
	timetaken = difftime(end, start);
	cout<<"Maximum obtained with the Naive Greedy with Feature Based Functions is "<< ff.evalFast(optSet)<<" and the time taken is "<<timetaken<<" seconds.\n";

	// Random Subset
	srand(time(NULL));
	randomSet.clear();
	for (int tt = 0; tt < budget; tt++){
		int randel = rand() % n;
		randomSet.insert(randomSet.end(), randel);
	}
	cout<<"The value of a random set with Saturated Coverage of the same size is "<< ff.eval(randomSet)<<"\n";
	cout<<"\n";
	
	/*
		Combining all the above functions
	*/
	cout<<"***************************************************************************\n";
	cout<<"Let's try combining all the above Functions: Choose 5% of the data.\n";
	budget = 0.05*n;
	int k = 3; // number of combining functions.
	vector<submodularFunctions*> fvec = vector<submodularFunctions*>();
	vector<double> lambdaVec(k, 1);
	scaleSubmodularFunctions ffscaled(groundSet, &ff, 1/ff.eval(groundSet));
	fvec.push_back(&ffscaled);
	scaleSubmodularFunctions sCscaled(groundSet, &sC, 1/sC.eval(groundSet));
	fvec.push_back(&sCscaled);
	scaleSubmodularFunctions fLscaled(groundSet, &fL, 1/fL.eval(groundSet));
	fvec.push_back(&fLscaled);
	combineSubmodularFunctions fComb(groundSet, fvec, lambdaVec);
	// Lazy Greedy
	start = time(0);
	optSet = lazyGreedyMax(fComb, costList, budget, emptyset, 0);
	end = time(0);
	timetaken = difftime(end, start);
	cout<<"Maximum obtained with the Lazy Greedy with combination is "<< fComb.evalFast(optSet)<<" and the time taken is "<<timetaken<<" seconds.\n";

	// Naive Greedy
	start = time(0);
	optSet = naiveGreedyMax(fComb, costList, budget, emptyset, 0);
	end = time(0);
	timetaken = difftime(end, start);
	cout<<"Maximum obtained with the Naive Greedy with combination is "<< fComb.evalFast(optSet)<<" and the time taken is "<<timetaken<<" seconds.\n";

	// Random Subset
	srand(time(NULL));
	randomSet.clear();
	for (int tt = 0; tt < budget; tt++){
		int randel = rand() % n;
		randomSet.insert(randomSet.end(), randel);
	}
	cout<<"The value of a random set with Saturated Coverage of the same size is "<< fComb.eval(randomSet)<<"\n";
	cout<<"\n";
	/*

			******************** Non-Monotone Submodular Maximization *******************
	*/
	cout<<"\n";
	cout<<"***************************************************************************\n";
	cout<<"Now we do non-monotone submodular maximization!\n";
	cout<<"\n";
	/*
		Asymmetric Graph Cut + Similarity Penalty 
	*/
	cout<<"***************************************************************************\n";
	double lambda = 1;
	cout<<"Let's try the graph cut function with similarity penalty. Set lambda = "<<lambda<<".\n";

	graphCutFunctions gC (groundSet, similarityFileTIMIT, lambda);
	graphCutFunctions gCcopy (groundSet, similarityFileTIMIT, lambda); // make a copy for the bidirectional greedy algorithms.

	start = time(0);
	optSet = localSearchMax(gC, emptyset, 0);
	end = time(0);
	timetaken = difftime(end, start);
	cout<<"Maximum obtained with the Local Search Algorithm with graphCut Penalty is "<< gC.eval(optSet)<<" and the time taken is "<<timetaken<<" seconds.\n";

	start = time(0);
	totalOrder<int> initOrder(n);
	optSet = bidirGreedyMax(gC, gCcopy, initOrder, 1, 0);
	end = time(0);
	timetaken = difftime(end, start);
	cout<<"Maximum obtained with the Deterministic Bi-directional Greedy Algorithm with graphCut Penalty is "<< gC.eval(optSet)<<" and the time taken is "<<timetaken<<" seconds.\n";
	
	start = time(0);
	optSet = bidirGreedyMax(gC, gCcopy, initOrder, 1, 1);
	end = time(0);
	timetaken = difftime(end, start);
	cout<<"Maximum obtained with the Randomized Bi-directional Greedy Algorithm with graphCut Penalty is "<< gC.eval(optSet)<<" and the time taken is "<<timetaken<<" seconds.\n";

	srand(time(NULL));
	randomSet.clear();
	for (int i = 0; i < n; i++){
		if( (rand() % 2) == 0)
			randomSet.insert(randomSet.end(), i);
	}
	cout<<"The value of a random set with graphCut Penalty "<< gC.eval(randomSet)<<"\n";
	cout<<"\n";

/*

			******************** Difference between Submodular Functions *******************
*/


}
