function [bow,FDESC,dictionary,features] = get_BOW4(imagescolor, images, options)
% get_bow   Computes a bag-of-words representation of an images
%  TODO: NEEDS UPDATE
% Returns a bag of visual word representation for every image. Currently,
% supported featureTypes are 'sift', 'denseSift', 'phow', ...
%
% options.clustering can be either 'random', 'kmeans' or 'given' (if given
% is selected options.centers must contain the cluster centers)
%
%   options.encoder = {'nn', 'stIP', 'stIPmax', 'stIPsum'}
%   options.encoderThreshold = ... (in case of stIP*)
%
%   options.clustering = {'kmeans', 'random', given'}
%   options.centers = ... (in case of given)

%% TODO: Return meaningful stuff in FDESC and dictionary for all methods of choice

%% settings
nOptArg = 3;
if nargin < nOptArg || isempty(options) || ~isfield(options, 'clustering'),
    options.clustering = 'random';
    warning('Clustering algorithm not specified. Using default clustering "%s".\n', options.clustering);
end

if nargin < nOptArg || isempty(options) || ~isfield(options, 'nrVisualWords'),
    options.nrVisualWords = 500;
    warning('You did not set the number of visual words, hence it is set to the default of %d.', options.nrVisualWords);
end

if nargin < nOptArg || isempty(options) || ~isfield(options, 'featureType'),
    options.featureType = 'phow';
    warning('You did not set the type of features, hence it is set to the default of "%s".', options.featureType);
end

if nargin < nOptArg || isempty(options) || ~isfield(options, 'verbose'),
    options.verbose = true;
    warning('get_BOW: Verbosity not deactivated, so you are in verbose mode right now.\n');
end

if nargin < nOptArg || isempty(options) || ~isfield(options, 'encoderOptions'),
    options.encoder = 'nn';
    warning('EncoderOptions not specified. Using default encoder "%s".\n', options.encoder);
end

if nargin < nOptArg || isempty(options) || ~isfield(options, 'normalizeBOW'),
    options.normalizeBOW = 'none';
    warning('Normalization for BOW not specified. Using default normalization "%s".\n', options.normalizeBOW);
end

if nargin < nOptArg || isempty(options) || ~isfield(options, 'normalizeFeatures'),
    options.normalizeFeatures = 'l1';
    warning('Normalization for features not specified. Using default normalization "%s".\n', options.normalizeFeatures);
end

if nargin < nOptArg || isempty(options) || ~isfield(options, 'buildPyramid'),
    options.buildPyramid = false;
end

if nargin < nOptArg || isempty(options) || ~isfield(options, 'dataset'),
    options.dataset = 'X';
    warning('Dataset not specified. No way to store histograms in a meaningful way.');
end

if nargin < nOptArg || isempty(options) || ~isfield(options, 'appendColorHistogram'),
    options.appendColorHistogram = false;
end

if strcmp(options.featureType, 'color') || strcmp(options.featureType, 'colorSQ')
    optionsHist = struct();
    optionsHist.bins = [8,8,8];
end

if strcmp(options.featureType, 'colorHistogram'),
    if ~strcmp(options.clustering, 'kmeans') || ~strcmp(options.normalizeFeatures, 'l1'), % || ~strcmp(options.encoderOptions.encoder, 'nn') 
        error('ColorHistogram must be paired with kmeans, nn, and l1-normalization to work properly.');
    end
end

if nargin < nOptArg || isempty(options) || ~isfield(options, 'summaries'),
    imagePriors = 1/100*ones(100,1); % posterior for drawing features from images
else
    imagePriors = zeros(100,1);
    summaries = sort(options.summaries,2);
    for i = 1:100,
        imagePriors(i) = sum(any(summaries == i, 2)) / size(summaries, 1);
    end
    imagePriors = imagePriors / max(imagePriors);
end

%% settings for feature computation
switch options.featureType,
    case 'dsift'
        optionsVF = {'Step', 8, 'Size', 16, 'fast'};
    case 'phow'
        optionsVF = {'Step', 4, 'Sizes', [4,8,16]}; % just use default options
        % optionsVF = {'Sizes', [4,6,8,10], 'Step', 4, 'Color', 'gray'};
    otherwise
        optionsVF = {};
end

%% general stuff
nrImages = size(images, 3);

desc = [];
nrFeatureVectors = 0;

%% compute feature descriptors for all images
if options.verbose,
    fprintf(1, 'Extracting features using "%s".\n', options.featureType);
end
features = cell(nrImages,1);
desc = [];
descImg = {};
for img = 1:nrImages,
    if options.verbose,
        fprintf(1, '.');
    end
    switch options.featureType,
        case 'sift' % SIFT (using standard interest point detector)
            [F, features{img}] = vl_sift(single(images(:,:,img)), optionsVF{:});
            features{img} = single(features{img});
            descImg{img} = F(1:2,:);
            desc = horzcat(desc, F);
        case 'dsift' % dense SIFT
            [F, features{img}] = vl_dsift(single(images(:,:,img)), optionsVF{:});
            features{img} = single(features{img});
            descImg{img} = F(1:2,:);
            desc = horzcat(desc, F);
        case 'phowColor' % PHOW
            optionsVF = {'Color', 'rgb'};
            [F, features{img}] = vl_phow(single(images(:,:,img)), optionsVF{:});
            features{img} = single(features{img});
            descImg{img} = F(1:2,:);
            desc = horzcat(desc, F);
            clear('F');
        case 'phow' % PHOW
            [F, features{img}] = vl_phow(single(images(:,:,img)), optionsVF{:});
            features{img} = single(features{img});
            descImg{img} = F(1:2,:);
            desc = horzcat(desc, F);
            clear('F');
        case 'loadFromFile' % Loads precomputed SIFT descriptors generated by the software from Koen van de Sande
            if ~isfield(options, 'directory'),
                error('You must set the directory option.');
            end
            directory = options.directory;
            if ~exist(directory, 'dir'),
                error('Specified directory does not exist.');
            end
            
            % load descriptors from file            
            filename = sprintf('%s/%d.desc', directory, img);
            [features{img}, descriptors] = readDescriptorsFromFile(filename);
            descImg{img} = descriptors(1:2,:);
            desc = horzcat(desc, descriptors);
        case 'colorHistogram' % learning according to http://hal.inria.fr/docs/00/61/45/23/PDF/acmm11_boc_wengert-douze-jegou_hal.pdf
            % get frequent colors for clustering
            features{img} = single(get_frequentColors(imagescolor(:,:,:,img)));
            desc = horzcat(desc, 1:size(features{img},2));
            features{img} = features{img}';
        case {'colorDirect'} % color histogram
            FDESC = [];
            dictionary = [];
            bow = [];
            for img = 1:nrImages,
                bowImg = [];
                patchSize = 10;
                patchOffset = 10;
                nr = (100-patchOffset) / patchOffset;
                optionsHist = struct();
                optionsHist.bins = [8,8,4];
                optionsHist.histType = 'SQ';
                
                nrm = max(imagescolor(:)) * 1.01;

                xmin = 1;
                feat = [];
                for i = 1:nr,
                    xmax = xmin + patchSize - 1;
                    ymin = 1;
                    for l = 1:nr,
                        ymax = ymin + patchSize - 1;

                        patch = single(imagescolor(xmin:xmax,ymin:ymax,:,img)) / nrm; % / 1.001;% / 255.1;
                        h = get_colorHistogramPatch(patch, optionsHist);
                        h = h / sum(h);
                        feat = horzcat(feat, h);
                        desc = horzcat(desc, [xmin;xmax;ymin;ymax]);

                        ymin = ymin + patchOffset;
                    end
                    xmin = xmin + patchOffset;
                end
                
                features{img} = feat;
            end
            
            % select a subset of 50 descriptors per image for efficiency
            % reasons
            for img = 1:nrImages,
                perm = randperm(size(features{img},2));
                features{img} = features{img}(:,perm(1:50));
            end
                        
            return;
        case {'color','colorSQ'} % color histogram
            % TODO: Check and cleanup
            % warning('Extracting color histograms -- experimental.');
            if false, %exist(sprintf('tmp/data_%d_%d.mat',options.dataset, img), 'file'),
                load(sprintf('/tmp/data_%d_%d.mat',options.dataset, img));
                % warning('Loading precomputed color histograms...');
            else
                patchSize = 20;
                patchOffset = 5;
                nr = (100-(patchSize-patchOffset)) / patchOffset;
                if strcmp(options.featureType, 'colorSQ'),
                    optionsHist.histType = 'SQ';
                else
                    optionsHist.histType = 'VQ';
                end

                xmin = 1;
                feat = [];
                for i = 1:nr,
                    xmax = xmin + patchSize - 1;
                    ymin = 1;
                    for l = 1:nr,
                        ymax = ymin + patchSize - 1;

                        patch = imagescolor(xmin:xmax,ymin:ymax,:,img); % / 1.001;% / 255.1;
                        h = get_colorHistogramPatch(patch, optionsHist);
                        h = h / sum(h);
                        feat = horzcat(feat, single(h));
                        desc = horzcat(desc, [xmin;xmax;ymin;ymax]);

                        ymin = ymin + patchOffset;
                    end
                    xmin = xmin + patchOffset;
                end

                fname = sprintf('tmp/data_%d_%d.mat', options.dataset, img);
                save(fname, 'feat', 'desc');
                fprintf(1, 'Saving histogram to "%s".\n', fname);
            end
            
            % extract feature coordinate points
            patchSize = 10;
            patchOffset = 5;
            nr = (100-patchOffset) / patchOffset;

            nrm = max(imagescolor(:)) * 1.01;

            xmin = 1;
            ldesc = [];
            for i = 1:nr,
                xmax = xmin + patchSize - 1;
                ymin = 1;
                for l = 1:nr,
                    ymax = ymin + patchSize - 1;

                    ldesc = horzcat(ldesc, [(xmax+xmin)/2; (ymax+ymin)/2]);

                    ymin = ymin + patchOffset;
                end
                xmin = xmin + patchOffset;
            end
            
            % pool features
            
            descImg{img} = ldesc;
            
            features{img} = single(feat);
        case 'hog'
            cellsize = 4;
            hog = vl_hog(single(images(:,:,img)), cellsize);
            t = size(hog);
            thog = hog;
            hog = [];
            for x = 1:size(thog,1),
                for y = 1:size(thog,2),
                    f = thog(x,y,:);
                    hog = horzcat(hog, f(:));
                end
            end
            % hog = reshape(hog, prod(t(1:2)), t(end));
            features{img} = single(hog);
            desc = rand(2, size(hog,2));
        case 'surf' % surf features
            % TODO: Test
            Options.upright=true;
            Options.tresh=0.00005;
            Options.init_sample = 1;
            Ipts=OpenSurf(single(images(:,:,img)),Options);
            features{img} = single(horzcat(Ipts.descriptor));
            desc = horzcat(desc, Ipts.descriptor); % TODO: Do something useful
        otherwise
            error('Unsupported feature type.');
    end
    nrFeatureVectors = nrFeatureVectors + size(features{img}, 2);
end
if options.verbose,
    fprintf(1, '\n');
end
if options.verbose,
    fprintf(1, 'Computed %d feature descriptors.\n', nrFeatureVectors);
end

%% normalize features
if options.verbose,
    fprintf(1, 'Normalizing extracted features using "%s".\n', options.normalizeFeatures);
end
for img = 1:nrImages,
    switch options.normalizeFeatures
        case 'l1'
            nrm = norm(features{img}, 1);
            if nrm < 1e-6,
                nrm = 1;
            end
            features{img} = bsxfun(@times, features{img}, 1./nrm);
        case 'l2'
            nrm = norm(features{img}, 2);
            if nrm < 1e-6,
                nrm = 1;
            end
            features{img} = bsxfun(@times, features{img}, 1./nrm);
        case 'none'
            
        otherwise
            error('Invalid method for normalizing the extracted features selected.');
    end
end

%% concatenate color histogram
if options.appendColorHistogram,
    optionsHistConcat = struct();
    optionsHistConcat.histType = 'SQ';
    optionsHistConcat.bins = [16,16,16];
    if options.verbose,
        fprintf(1, 'Appending color histogram: ');
    end
    for img = 1:100,
        width = 10;
        if options.verbose,
            fprintf(1, '.');
        end
        concat = zeros(sum(optionsHistConcat.bins),size(descImg{img},2),'single');
        for l = 1:size(descImg{img},2),
            x = descImg{img}(1,l);
            y = descImg{img}(2,l);
            xmin = max(1,x-width);
            xmax = min(100,x+width);
            ymax = min(100,y+width);
            ymin = max(1,y-width);
            patch = imagescolor(xmin:xmax,ymin:ymax,:,img);
            concat(:,l) = get_colorHistogramPatch(patch, optionsHistConcat);
            concat(:,l) = concat(:,l) / norm(concat(:,l), 1);
        end
        features{img} = vertcat(features{img}, concat);
    end
    fprintf(1, '\n');
    clear('optionsHistConcat');
end

%% concatenate feature vectors for easier use
featuresM = zeros(size(features{1}, 1), nrFeatureVectors, 'single');
idx = 1;
for img = 1:nrImages,
    nrDesc = size(features{img}, 2);
    featuresM(:, idx:(idx+nrDesc-1)) = features{img};
    idx = idx + nrDesc;
end

%% Select centers for encoding
if options.verbose,
    fprintf(1, 'Selecting centers (dictionary) for encoding using "%s".\n', options.clustering);
end
dictionary = zeros(size(featuresM,1), options.nrVisualWords);
fvecids = [];
switch options.clustering,
    case 'kmeans'
        fprintf(1, 'Performing k-means (this may take a while).\n');
        dictionary = vl_kmeans(featuresM, options.nrVisualWords, 'algorithm', 'elkan', 'MaxNumIterations', 100);
        fvecids = 1:options.nrVisualWords; % not true, just for making the code run
    case 'random'
        perm = randperm(size(featuresM,2));
        dictionary = featuresM(:,perm(1:options.nrVisualWords));
        fvecids = horzcat(fvecids, perm(1:options.nrVisualWords));
    case 'gmm'
        fprintf(1, 'Training GMM (this may take a while).\n');
        t = randperm(size(featuresM,2));
        [means, covariances, priors] = vl_gmm(featuresM(:,t(1:1e5)), options.nrVisualWords);
    case 'randomDict'
        dictionary = rand(size(featuresM,1), options.nrVisualWords, 'single');
        for k = 1:options.nrVisualWords,
            dictionary(:,k) = dictionary(:,k) / norm(dictionary(:,k),1);
        end
        % TODO: Do something useful
        perm = randperm(size(featuresM,2));
        fvecids = horzcat(fvecids, perm(1:options.nrVisualWords));
    case 'randomLogSpace' % TODO: sample correctly
        % select angle
        angle = rand(options.nrVisualWords,1)*2*pi;
        
        % select distance
        maxdist = norm([50,50]);
        z = Inf(options.nrVisualWords,1);
        while any(z > maxdist),
            z(z > maxdist) = abs(randn(sum(z > maxdist),1))*maxdist/2;
        end
        
        x = 50 + cos(angle) .* z;
        y = 50 + sin(angle) .* z;
        fvecids = []; %NaN(options.nrVisualWords, 1);
        k = 1;
        while k <= options.nrVisualWords,
            % select img
            img = ceil(rand(1)*100);
            
            % find closest descriptor
            tmp = descImg{img} - repmat([x(k);y(k)], 1, size(descImg{img},2));
            tmp = sum(tmp.^2, 1);
            [~,idx] = min(tmp);
            
            % compute similarity to other centers and reject if too similar
%             if k > 1,
%                 tidx = (img-1)*size(features{img},2) + idx;
%                 sim = sum(abs(featuresM(:,fvecids) - repmat(featuresM(:,tidx), 1, length(fvecids))), 1) / size(featuresM,1);
%                 fprintf(1, 'Max similarity: %f\n', max(sim));
%                 if max(sim) > 0.095,
%                     fprintf(1, 'Reject\n');
%                     k = k+1;
%                     continue;
%                 end
%             end
            
            if any(fvecids == (img-1)*size(features{img},2) + idx),
                % fprintf(1, 'Skipping...\n');
                x(k) = rand(1)*100;
                y(k) = rand(1)*100;
                k = k +1;
                continue;
            end
            
            fvecids(end+1) = (img-1)*size(features{img},2) + idx;
            
            t = descImg{img};
            xs = t(1,idx);
            ys = t(2,idx);
            
            % fprintf(1, 'Selecting feature %d at %d/%d (should be close to %2.2f/%2.2f) (distance to center: %2.2f).\n', fvecids(k), xs, ys, x(k), y(k), norm([xs-50, ys-50]));

            k = k + 1;
        end
        dictionary = featuresM(:,fvecids);
    case 'randomPrior'
        % select features according to image priors (may not be that good idea)
        warning('Using image frequencies as priors for selecting features.');
        for l = 1:options.nrVisualWords,
            a = rand(1);
            i = find(imagePriors >= a);
            perm = randperm(length(i));
            img = perm(1); % image to select a feature from
            nrFeaturesInImage = size(features{img}, 2);
            perm = randperm(nrFeaturesInImage);
            dictionary(:,l) = features{img}(:,perm(1));
            % fprintf(1, '[%4d] Selecting features from image %3d: #%d\n', l, img, perm(1));
            fvecids = horzcat(fvecids, (nrFeaturesInImage)*(img-1) + perm(1));
        end
    case 'given'
        if nargin < nOptArg || isempty(options) || ~isfield(options, 'centers'),
            error('Centers must be specified.');
        end
        dictionary = options.centers;
        fvecids = 1:options.nrVisualWords; % not true, just for making the code run
    otherwise
        error('Invalid method for selecting centers chosen.');
end
    
% set FDESC
if ~strcmp(options.clustering, 'gmm'),
    FDESC = descImg;
%     FDESC = cell(options.nrVisualWords, 1);
%     for i =  1:length(fvecids),
%     data.sift = featuresM(:,fvecids(i));
%     data.desc = desc(:,fvecids(i));
%     data.img = ceil(fvecids(i) / size(features{1}, 2));
%     FDESC{i} = data;
%     end
    FDESC = descImg;
    %FDESC = 'not set';   
end

% shall we determine an overall color histogram?
if strcmp(options.featureType, 'colorHistogram'),
    warning('Computing color histogram for whole images.');
    % recompute features
    for img = 1:nrImages,
        features{img} = single(reshape(imagescolor(:,:,:,img), 300*300, 3))';
        x = repmat(1:100, 100, 1)';
        y = repmat(1:100, 100, 1);
        descImg{img} = [x(:)',y(:)'];
        clear('x', 'y');
    end
end

%% Compute bag-of-words, that is perform encoding
if ~options.buildPyramid,
    bow = NaN(size(dictionary,2), nrImages);
else
    nrLayers = 1;
    bow = NaN(sum((2.^(0:nrLayers)).^2)*size(dictionary,2), nrImages);
end

% precompute thresholds
if options.verbose,
    fprintf(1, 'Precomputing thresholds...\n');
end
switch options.encoderOptions.encoder,
    case 'stIP'
        encTh = zeros(size(dictionary,2), 1);
        for k = 1:size(dictionary, 2),
            tmp = dictionary(:,k)'*featuresM;
            f = sort(tmp,2);
            encTh(k) = f(1,ceil(length(f)*options.encoderOptions.encoderThreshold));
        end
        %encTh'
end

if options.verbose,
    fprintf(1, 'Encoding using "%s".\n', options.encoderOptions.encoder);
    fprintf(1, 'Encoding: ');
end

for img = 1:nrImages,
    if options.verbose,
        fprintf(1, '.');
    end
    % apply noise to dictionary
    %dictionary = dictionary + randn(size(dictionary)) * mean(dictionary(:)) * 0.1;
    %dictionary(dictionary < 0) = 0;
    
    if length(descImg) >= img,
        options.encoderOptions.descriptors = descImg{img};
    else
        options.encoderOptions.descriptors = [];
    end
    switch options.encoderOptions.encoder,
        case {'nn', 'nnSoft', 'nnEMD'}
            if strcmp(options.featureType, 'colorSQ'),
                options.encoderOptions.bins = optionsHist.bins;
                options.encoderOptions.quantizerType = 'SQ';
            end
            if options.buildPyramid,
                [~,encmap] = encode(features{img}, dictionary, options.encoderOptions.encoder, options.encoderOptions);
                aidx = 1;
                for layer = 0:nrLayers,
                    patchSize = 100 / (2^layer);
                    for i = 1:(2^layer),
                        for j = 1:(2^layer),
                            xmin = (i-1)*patchSize + 1;
                            xmax = xmin + patchSize - 1;
                            ymin = (j-1)*patchSize + 1;
                            ymax = ymin + patchSize - 1;
                            
                            desc = descImg{img};
                            idx1 = (desc(1,:) >= xmin) & (desc(1,:) <= xmax);
                            idx2 = (desc(2,:) >= ymin) & (desc(2,:) <= ymax);
                            idx = logical(idx1 & idx2);
                            
                            eidx = aidx + options.nrVisualWords-1;
                            bow(aidx:eidx,img) = 1/(2^(nrLayers-layer)) * hist(encmap(idx), 1:options.nrVisualWords);
                            aidx = aidx + options.nrVisualWords;
                        end
                    end
                end
            else
                bow(:,img) = encode(features{img}, dictionary, options.encoderOptions.encoder, options.encoderOptions);
            end
        case 'stIP'
            warning('Temporary test implementation. Please remove me.');
            % encTh = 0.95*ones(size(encTh));
            % compute true threshold feature wise such only th-percent are active
            tmp = dictionary'*features{img};
            tmp = tmp - repmat(encTh, 1, size(tmp,2));
            tmp(tmp < 0) = 0;
            f = sum(tmp,1);
            f(f < 1e-6) = 1;
            % f = ones(size(f));
            %tmp = tmp ./ repmat(f, size(tmp,1), 1); % summed activiation = 1 per feature
            bow(:,img) = sum(tmp,2);
           
            
            % tmp = max(tmp, zeros(size(tmp)));
%             nr = size(features{img});
%             bow(k,img) = sum(tmp(:,((img-1)*nr+1):(img*nr)), 2);
            clear('tmp');
        case 'none'
            
        case 'gmm'
            l = zeros(options.nrVisualWords, size(features{img}, 2));
            for i = 1:options.nrVisualWords,
                l(i,:) = mvnpdf(features{img}', means(:,i)', diag(covariances(:,i)))*priors(i);
            end
            [~,d] = max(l, [], 1);
            tmp = accumarray(d', ones(size(d))');
            if length(tmp) < options.nrVisualWords,
                tmp = vertcat(tmp, zeros(options.nrVisualWords-length(tmp), 1));
            end
            bow(:,img) = tmp;
        case 'none'
            
        otherwise
            bow(:,img) = encode(features{img}, dictionary, options.encoderOptions.encoder, options.encoderOptions);
    end
end
if options.verbose,
    fprintf(1, '\n');
end

% if strcmp(options.featureType, 'colorHistogram'),
%     warning('Applying power rule.');
%     % recompute features
%     bow = sqrt(bow);
% end

%% Normalize bag-of-words
if options.verbose,
    fprintf(1, 'Normalizing bag-of-words using "%s".\n', options.normalizeBOW);
end
switch options.normalizeBOW,
    case 'l1'
        for k = 1:size(bow, 2),
            nrm = norm(bow(:,k), 1);
            if nrm < 1e-6,
                nrm = 1;
            end
            bow(:,k) = bow(:,k) / nrm;
        end
    case 'maxActivation'
        for k = 1:size(bow, 2),
            bow(:,k) = bow(:,k) / (size(features{img}, 2) * options.encoderOptions.encoderThreshold);
        end        
%     case 'numberOfWords'
%         for k = 1:size(bow, 2),
%             nrm = sum(bow(:,k) > 0, 1);
%             if nrm < 1e-6,
%                 nrm = 1;
%             end
%             bow(:,k) = bow(:,k) / nrm;
%         end
    case 'none'
        % do nothing
    otherwise
        error('Invalid normalization method.');
end

end

function [features, descriptors] = readDescriptorsFromFile(filename)
    if ~exist(filename, 'file'),
        error('Specified file "%s" does not exist.', filename);
    end
    
    fileContent = fileread(filename);
    fileLines = textscan(fileContent, '%s', 'Delimiter', '\n');
    fileLines = fileLines{1};
    
    % read file type
    a = fileLines{1};
    if ~strcmp(strtrim(a), 'KOEN1'),
        fclose(fh);
        error('Invalid file specified.');
    end
    
    % read descriptor dimension
    descDimension = str2double(fileLines{2});
    
    % read number of descriptors
    nrDescriptors = str2double(fileLines{3});
    
    % read in features and descriptors
    features = NaN(descDimension, nrDescriptors, 'single');
    descriptors = NaN(5, nrDescriptors, 'single');
    for i = 1:nrDescriptors,
        a = fileLines{3+i};
        idx = strfind(a, ';');
        descString = a(1:idx);
        featString = a((idx+1):end);
        
        % parse feature
        idx = 1;
        tmp = textscan(featString, '%d');
        features(:,i) = single(cell2mat(tmp));
        
        % parse descriptor
        [a,b] = strtok(descString); % a = '<CIRCLE', b = ' 91 93 0.8 0 0>;');
        b = b(1:(end-2));
        tmp = textscan(b, '%d');
        descriptors(:,i) = single(cell2mat(tmp));
    end
end
