%% implements the image collection summarization algorithm according to
%% Chunlei Yang, Jialie Shen, Jinye Peng, Jianping Fan: 
%%   Image collection summarization via dictionary learning for sparse representation. Pattern Recognition 46(3): 948-961 (2013)

%% reset
clc;
clear all;

%% settings
maxIterations = 40; % from paper (40)
beta = 0.05; % diversity weight, from paper
nrInitialTries = 3; % number of different initial dictionaries to try, from paper
maxConsRej = 20; % maximum number of consecutive rejects, from paper
maxTries = 30; % (40 according to paper) maximum number of dictionaries selected in the basis update stage. from paper
tempDecreaseRate = 0.9; % rate for temperature decrease, from paper
nrDictionaryEntries = 10; % from paper
nrVisualWords = 100; % from paper

tempInitial = 1; % initial temeperature (value not stated in paper)
tempStop = 0.1; % stopping temperature (value not stated in paper)
sigma = 1; % variance for distance measure (value not stated in paper)
lambda = 1e-3; % (value not stated in paper)
alphaTest = 1; % for equation (5) (value not stated in paper)
dataset = 2; 

%% preprocessing (compute sift features and select centers by k-means, 
%% compute codewords -> output is codebook (each column is the codeword
%% for one picture, each row corresponds to a visual feature)
%codebook = abs(rand(1000, 100));
%[imagescolor, images] = get_images(2, false);
%codebook = get_BOW4(imagescolor, images, nrVisualWords, 'phow');

codebook = get_BOW_evaluation(dataset);

% normalize codebook
for i = 1:size(codebook, 2),
    codebook(:,i) = codebook(:,i) / norm(codebook(:,i));
end

% compute similiarity matrix
sim = zeros(size(codebook, 2));
for i = 1:size(codebook, 2),
    for l = 1:size(codebook, 2),
        sim(i,l) = exp(-norm(codebook(:,i) - codebook(:,l))^2 / (2*sigma^2));
    end
end

% sort similarity matrix
[sortedSim, simPerm] = sort(sim, 1, 'descend');

% compute correlations
correlations = zeros(size(codebook, 2));
for i = 1:size(codebook, 2),
    for l = 1:size(codebook, 2),
        % TODO: Is this what the authors meant?
        correlations(i,l) = (codebook(:,i) - mean(codebook(:,i)))' * (codebook(:,l) - mean(codebook(:,l))) / (std(codebook(i,:)) * std(codebook(l,:)));
    end
end

%% main loop
consRej = 0; % number of consecutive rejects
temp = tempInitial; % initial temperature
iter = 0;

% create inital random dictionary
D = {};
perm = randperm(size(codebook, 2));
D{1} = perm(1:nrDictionaryEntries);
R = {};
[obj, A] = sparseCoding(codebook, codebook(:,D{1}), lambda);
R{1} = obj + beta*computeCorrelation(D{1}, correlations);
Ropt = R{1};
Dopt = D{1};

fprintf(1, '*** Initial dictionary: objective = %f.\n\n\n\n', Ropt);

while (temp > tempStop) && (consRej < maxConsRej) && (iter < maxIterations),
    % update iteration counter
    iter = iter + 1;
    fprintf(1, '>> Iteration %d\n', iter);
    fprintf(1, 'There are currently %d candidate dictionaries.\n', length(D));
    
    % variables for this loop
    Dnew = {};
    Rnew = {};
    
    % (1) Update temperature
    % Note: equation (4), there is a k_0 mentioned whose value is not
    %       stated -> I set it such that first temperature is 1
    temp = tempInitial / log((exp(1)-1) + iter);
    fprintf(1, 'Current temperature: %f\n', temp);
    
    % (2) update dictionaries and make exploration
    searchRange = ceil(exp((temp - tempInitial)/tempInitial) * size(codebook, 2));
    fprintf(1, 'Current search range: %d\n', searchRange);
    for d = 1:length(D), % cycle through candidate dictionaries
        if mod(d, 10) == 0,
            %d
            %fprintf(1, 'Accepted: %d/%d\n', length(Rnew), maxTries * d);
        end
        for l = 1:maxTries,
            % select at most maxTries bases to try
            candidateDictionary = D{d};

            % update bases
            for k = 1:nrDictionaryEntries,
                perm = randperm(searchRange);
                if perm(1) ~= 1,
                    perm = perm(1);
                else
                    perm = perm(2);
                end
                perm = simPerm(perm, candidateDictionary(k));
                candidateDictionary(k) = perm;
            end

            % (4) check if this dictionary gets accepted
            Rtemp = sparseCoding(codebook, codebook(:,candidateDictionary), lambda) + beta*computeCorrelation(candidateDictionary, correlations);
            if Rtemp < R{d}, % accept right away, as objective decreased
                % fprintf(1, 'Dictionary accepted right away.\n');
                Dnew{end+1} = candidateDictionary;
                Rnew{end+1} = Rtemp;
            else
                u = rand(1);
                % fprintf(1, 'Comparing random %f to %f.\n', u, exp(-(Rtemp - R{d})/(temp*alphaTest)));
                if exp(-(Rtemp - R{d})/(temp*alphaTest)) >= u,
                    %fprintf(1, 'Dictionary accepted by chance.\n');
                    Dnew{end+1} = candidateDictionary;
                    Rnew{end+1} = Rtemp;
                end
                % fprintf(1, 'XXX dictionary not accepted.\n');
            end
        end
    end
    
    % (9) is there a new best dictionary?
    if min(cell2mat(Rnew)) < Ropt,
        [~,id] = min(cell2mat(Rnew)) ;
        Ropt = Rnew{id};
        Dopt = Dnew{id};
        fprintf(1, '*** New optimal dictionary.\n');
    else
        maxConsRej = maxConsRej + 1;
    end
    
    fprintf(1, '*** Optimal dictionary: objective = %f.\n', Ropt);
    [Rt,At,err] = sparseCoding(codebook, codebook(:,Dopt), lambda);
    corrPenalty = beta*computeCorrelation(Dopt, correlations);
    Rt + corrPenalty
    fprintf(1, 'Reconstruction error: %f\n', err);
    fprintf(1, 'Sparsity penality: %f\n', lambda*sum(sum(abs(A))));
    fprintf(1, 'Correlation penalty: %f\n', corrPenalty);
    
    % replace old quantities with new ones
    D = Dnew;
    R = Rnew;
    
    if length(Dnew) == 0,
        fprintf(1, 'Stopping because there are no more candidate dictionaries.\n');
        break;
    end
    
    fprintf(1, '\n\n\n');
end


%% perform iterative basis selection
converged = false;
while ~converged,
    converged = false;
    
    % update bases
    [Rtemp, A] = sparseCoding(codebook, codebook(:,Dopt), lambda) + beta*computeCorrelation(Dopt, correlations);
    for k = 1:nrDictionaryEntries,
        
    end
    
    % test if there is an increase in the objective
    
end

% compute approximation error

% show selected summary