function [BoW, sim] = get_BOW2(imagescolor, images, numberOfWords, featureType, options)
% Returns a bag of visual word representation for every image. Visual
% features are selected as codes whenever they do not appear in too many
% images and they are note correlated too strong to the previous selection.

%% settings
if nargin < 5,
    if strcmp(featureType, 'dsift'),
        options = {'Step', 2, 'Size', 4};
    elseif strcmp(featureType, 'phow'),
        % options = {'Step', 2};
        options = {'Sizes', [2,4,8,16], 'Step',2, 'Color', 'gray'};
    else
        options = {};
    end
end

%% general stuff
nrImages = size(images, 3);
dictSize = numberOfWords;

features = {};
frames = {};
nrFeatureVectors = 0;

%% compute sift descriptors for all images
for img = 1:nrImages,
    if strcmp(featureType, 'dsift'),
        [F, features{img}] = vl_dsift(single(images(:,:,img)), options{:}); % vl_sift(single(images(:,:,img))); % vl_dsift(single(images(:,:,img)), 'Step', 4);
        features{img} = single(features{img});
    elseif strcmp(featureType, 'sift'),
        [F, features{img}] = vl_sift(single(images(:,:,img)), options{:}); % vl_sift(single(images(:,:,img))); % vl_dsift(single(images(:,:,img)), 'Step', 4);
        features{img} = single(features{img});
    elseif strcmp(featureType, 'phow'),
        [F, features{img}] = vl_phow(single(images(:,:,img)), options{:}); % vl_sift(single(images(:,:,img))); % vl_dsift(single(images(:,:,img)), 'Step', 4);
        %options = {};
        %[F, features{img}] = vl_sift(single(images(:,:,img)), options{:}); % vl_sift(single(images(:,:,img))); % vl_dsift(single(images(:,:,img)), 'Step', 4);
        features{img} = single(features{img});
        frames{img} = F;
    elseif strcmp(featureType, 'surf'),
        Options.upright=true;
        Options.tresh=0.00005;
        Options.init_sample = 1;
        Ipts=OpenSurf(single(images(:,:,img)),Options);
        features{img} = horzcat(Ipts.descriptor);
    else
        error('Unsupported feature type.');
    end
    nrFeatureVectors = nrFeatureVectors + size(features{img}, 2);
end

fprintf(1, 'Computed %d feature descriptors.\n', nrFeatureVectors);

%% concatenate feature vectors and perform dictionary computation 
%% by k-means clustering
featuresM = zeros(size(features{1}, 1), nrFeatureVectors);
idx = 1;
for img = 1:nrImages,
    nrDesc = size(features{img}, 2);
    featuresM(:, idx:(idx+nrDesc-1)) = features{img};
    idx = idx + nrDesc;
end
featuresM = single(featuresM);

fprintf(1, 'Performing k-means...\n');
%clust = KMedoids(single(featuresM)', dictSize, {});

% dictionary = vl_kmeans(single(featuresM), dictSize, 'verbose', 'algorithm', 'elkan', 'MaxNumIterations', 100);

% shuffle feature vectors
perm = randperm(size(featuresM,2));
featuresM = featuresM(:,perm);
dictionary = [];
accepted = [];
idx = 1;
bowContainer = [];
nrMaxActiveAccrossExamples = 40;
acc = [];
for cwIdx = 1:numberOfWords,
    accept = false;
    mincount = 20+mod(cwIdx, 5) * 10;
    mincount
    maxcount = mincount + 10 - 1
    while ~accept,
        dictionary = featuresM(:,idx);
        
        if ~isempty(acc),
            if any((acc(2,:) <= maxcount) &  (acc(2,:) >= mincount)),
                t = (acc(2,:) <= maxcount) &  (acc(2,:) >= mincount);
                l = find(t == 1);
                l = l(1);
                accidx = acc(1,l);
                count = acc(2,l);
                fprintf(1, 'Fast accepting %d (count=%d).\n', acc(1,l), acc(2,l));
                acc(:,t) = [];
                break;
            end
        end
        
        % correlation threshold
        coOk = 0;
        co = dictionary' * featuresM;
        coth = mean(co);
        
        % compute BoW for this dictionary
        tdist = vl_alldist(single(dictionary), featuresM);
        me = mean(tdist);
        mi = min(tdist);
        th = mean([me, mi]);
        bow = zeros(nrImages, 1);
        count = 0;
        for img = 1:nrImages,
            dist = vl_alldist(single(dictionary), single(features{img}));
            bow(img) = sum(dist < th);
            if bow(img) > 0,
                count = count + 1;
            end
        end
        
        % count
        if count < nrMaxActiveAccrossExamples,
            %tmp = horzcat(bowContainer, bow);
%             if cwIdx > 1,
%                 tmp = horzcat(bowContainer, bow);
%                 tmp(tmp > 0) = 1;
%                 a = mean(tmp, 2); % average sparsity along number of images
%                 mean(a)
%                 th = 1 / log(1/2*cwIdx)
%                 if mean(a) <= th,
%                     coOk = 1;
%                 end
%             else
%                 coOk = 1;
%             end
            coOk = 1;
        end
        
        if mincount <= count && count <= maxcount, %count < nrMaxActiveAccrossExamples && coOk,
            accept = true;
        else
            acc = horzcat(acc, [idx; count]);
        end
        
        accidx = idx;
        idx = idx + 1;
    end
    
    addIdx = accidx;
    fprintf(1, '[%3d] Accepted: %d (count=%d)\n', cwIdx, addIdx, count);
    accepted = horzcat(accepted, addIdx);
    bowContainer = horzcat(bowContainer, bow);
end
accepted

dictionary = featuresM(:,accepted);

fprintf(1, 'Computing BoWs...\n');
bow = []; %zeros(dictSize, nrImages);
kdtree = vl_kdtreebuild(dictionary);
%meandist = vl_alldist(single(dictionary), single(featuresM));
%meandist = mean(meandist, 2);

model = struct();
model.numSpatialX = [2 4];
model.numSpatialY = [2 4];
numWords = dictSize;
bow = zeros(numberOfWords, nrImages);
tdist = vl_alldist(single(dictionary), featuresM);
me = mean(tdist, 2);
mi = min(tdist, [], 2);
th = mean([me, mi], 2);
% threshholds
for img = 1:nrImages,
    dist = vl_alldist(single(dictionary), single(features{img})); % rows: distances to dictionary vectors; cols: distances to features
    %[~, binsa] = min(dist, [], 1);
    %[dist,bt] = sort(dist,1);
    for i = 1:size(dist,2), % first dim: dict size, second dim: number of interest points
        for j = 1:size(dist, 1),
            if dist(j,i) < th(j),
                bow(j, img) = bow(j, img) + 1;
            end
        %if dist(1,i) < 0.5*meandist(bt(i)), % 2*dist(2,i),
        %   bow(bt(1,i), img) = bow(bt(1,i), img) + 1;
        %end
        end
    end
end

BoW = bow;
for k = 1:size(BoW, 2),
    BoW(:,k) = BoW(:,k);% / norm(BoW(:,k), 1); %size(features{k},2);% norm(BoW(:,k), 1);
end

% options = zeros(1,14);
% options(1) = 1; % display
% options(2) = 1;
% options(3) = 0.1; % precision
% options(5) = 1; % initialization
% options(14) = 100; % maximum iterations
% 
% centers = zeros(dictSize, size(sift, 1));
% 
% % run kmeans
% fprintf('\nRunning k-means\n');
% dictionary = sp_kmeans(centers, sift', options);

% %% compute BOWs
% % perform vector quantization
% disp('Performing vector quantization...');
% sift = zeros(dictSize, nrImages);
% dictionary = dictionary';
% for i = 1:nrImages,
%     for j = 1:size(sift_all{i}, 2),
%         tmp = zeros(dictSize, 1);
%         for k = 1:size(dictionary, 1),
%             tmp(k) = norm(dictionary(k,:) - single(sift_all{i}(:,j))');
%         end
%         %[~,idx] = min(tmp);
%         %sift(idx, i) = sift(idx, i) + 1;
%         [a,b] = sort(tmp);
%         idx = b(1);
%         sift(idx, i) = sift(idx, i) + 1;
%     end
% end

for i = 1:size(bow, 2),
    nf = 1;
    if norm(bow(:,i)) > 0,
        nf = norm(bow(:,i));
    end
    bow(:,i) = bow(:,i) / nf;
end

%% Get the similarity matrix
sim = NaN(nrImages, nrImages);
for i = 1 : nrImages,
    for j = 1 : nrImages,
        sim(i, j) = bow(:,i)'*bow(:,j);
    end
end

if any(any(isnan(sim))),
    error('There are NaNs in the similarity matrix.');
end

imagesc(sim);

end