%% Learns the objective as discussed in the meeting on October, 1st

% given a set of features (color histograms, bag of features), learn 
% weights that assign importance to each of the words automatically. We
% do this by solving:
%
%   max_w sum_{k=1}^K f_w(A_k),
%
% where 
%
%   f_w(A) := sum_{u \in U} w_u \sqrt{ \sum_{a \in A} c_u(a) },
%
% U is the set of available words, D = {A_1, ..., A_K} is the set of 
% human summaries, c_u(a) counts the occurences of word u in image a (maybe
% using some kind of TF-IDF scaling), and w is the weight vector to be 
% tuned (non-negative).

clc;
clear all;

%% settings
dataset = 2;

% other settings
useTFIDF = true;

lambda = 10; % regularizer
nrAverage = 1000;
nrImages = 100;
summaryLength = 10;



for dataset = 1:14,
    %% load summaries and compute scores
    [summaries, nrSummaries] = load_summaries(dataset);
    
    %% compute random summaries
    % rng(123456);
    rand('seed', 123456);
    randomSummaries = randomSummary(nrImages, summaryLength, nrAverage);    
    
    %% load dataset
    [imagesrgb, imagesgray, imageshsv] = get_images(dataset, false);

    %% get bag-of-words
    [bow, bowDescriptorStr] = get_BOW(dataset);
    fprintf(1, 'Descriptor of bag-of-words: %s\n', bowDescriptorStr);

    %% compute tfidf
    tfidf = get_TFIDF(bow);
    
    %% apply tfidf
    if useTFIDF,
        bow = applyTFIDF(bow, tfidf);
    end

    %% stack obtained features or do something else that is useful
    
    %% check for features active only in a single image
    tmp = sum(bow > 0, 2) <= 1;
    if any(tmp),
        fprintf(1, 'There are %d words that are active in only one (or even none) image.\n', sum(tmp));
    end          
    
    %% compute bag-of-words of humand and random summaries
    summariesAsBOW = computeBOWSummary(summaries, bow);

     %% greedily select summary
    options.bow = bow;
    options.summariesAsBow = summariesAsBOW;  
    
    %% Run algorithm to get rid of bad human summaries   
    for addStd = 0:2, % prune up to two std deviations away from mean
        dirname = sprintf('%s/html/images/SummaryPruning/set%d/std%d', getPreDir(), dataset, addStd);
        if exist(dirname, 'dir'),
            warning('Directory "%s" already exists, skipping.', dirname);
            continue;
        end
        
        tSummaries = summaries;
        converged = false;
        iter = 0;
        while true,
            % compute bag-of-words of human and random summaries
            summariesAsBOW = zeros(size(tSummaries,1), size(bow, 1));
            for s = 1:size(tSummaries,1),
                summariesAsBOW(s,:) = sum(bow(:,tSummaries(s,:)),2);
            end
            options.summariesAsBow = summariesAsBOW;

            [randScore, randScores] = summaryEvaluation(randomSummaries, tSummaries, 'rouge', options);
            [sumScore, scores] = summaryEvaluation(tSummaries, tSummaries, 'rouge', options);
            
            %% compute submodular min
            [~,greedyMinScore,greedyMinSelected,~] = mmin(@(x) summaryEvaluation(x, tSummaries, 'rouge', options), 100, [], 'MUB', 10);
            [greedyMaxSelected, greedyMaxScore, ~] = greedySummary(@(x) summaryEvaluation(x, tSummaries, 'rouge', options));
            
            %% remove summaries
            toRemove = find(scores < mean(randScores) + addStd*std(randScores))'; % remove all summaries worse than the mean random ones
            fprintf(1, 'Removing elements...');
            toRemove
            removedSummaries = tSummaries(toRemove,:);
            tSummaries(toRemove,:) = [];


            %% plot distribution of summaries
            if iter == 0,
                xmin = greedyMinScore - 5;%min(min(scores), min(randScores)) - 0.1;
                xmax = greedyMaxScore + 5; %max(max(scores), max(randScores)) + 0.1;
                bins = linspace(xmin-1e-3, xmax+1e-3, 40);
                finebins = linspace(xmin-1e-3, xmax+1e-3, 100);
                %ymax = 0.3; %max(max(histc(scores, bins)), max(histc(randScores, bins)));
            end

            clf
            hold all;
            a1 = histc(scores, bins);
            a1 = a1(1:(end-1));
            a1 = a1 / (bins(2)-bins(1)) / sum(a1);
            m = mean(scores);
            s = std(scores);
            b1 = normpdf(finebins, m, s);

            a2 = histc(randScores, bins);
            a2 = a2(1:(end-1));
            a2 = a2 / (bins(2)-bins(1)) / sum(a2);
            m = mean(randScores);
            s = std(randScores);
            b2 = normpdf(finebins, m, s);

            ymax = max([max(a1), max(a2), max(b1), max(b2)])*1.1;

            hb = bar(bins(1:(end-1)) + (bins(2) -bins(1))/2,a1,'r');
            set(get(hb, 'Children'), 'FaceAlpha', 0.2);
            xlim([xmin, xmax]);
            ylim([0, ymax]);
            h = plot(finebins, b1, 'Linewidth', 2);
            set(h, 'Color', 'r');

            hb = bar(bins(1:(end-1)) + (bins(2) -bins(1))/2,a2,'b');
            set(get(hb, 'Children'), 'FaceAlpha', 0.2);
            xlim([xmin, xmax]);
            ylim([0, ymax]);
            h = plot(finebins, b2, 'Linewidth', 2);
            set(h, 'Color', 'b');
            xlim([xmin, xmax]);
            ylim([0, ymax]);
            
            % mark random min/max, human min/max
            tmp = min(randScores);
            h1 = line([tmp, tmp], [0, ymax]);
            tmp = max(randScores);
            h2 = line([tmp, tmp], [0, ymax]);
            set(h1, 'Color', 'b');
            set(h2, 'Color', 'b');
            set(h1, 'LineStyle', ':');
            set(h2, 'LineStyle', ':');
            
            tmp = min(scores);
            h1 = line([tmp, tmp], [0, ymax]);
            tmp = max(scores);
            h2 = line([tmp, tmp], [0, ymax]);
            set(h1, 'Color', 'r');
            set(h2, 'Color', 'r');
            set(h1, 'LineStyle', ':');
            set(h2, 'LineStyle', ':');
            
            % draw greedyMin/greeyMax
            h1 = line([greedyMinScore, greedyMinScore], [0, ymax]);
            h2 = line([greedyMaxScore, greedyMaxScore], [0, ymax]);
            set(h1, 'Color', 'k');
            set(h2, 'Color', 'k');
            
            scoresStr = sprintf('Human summaries/Random summaries (overlap): %f/%f (%f)\n', mean(scores), mean(randScores), computeGaussianOverlap(mean(scores), mean(randScores), std(scores), std(randScores)));
            title(sprintf('Iteration %d: %s\n', iter, scoresStr));
            
            %% TODO: Save best greedy summary/worst greedy summary
            % save to file
            fnameImage = sprintf('%s/html/images/SummaryPruning/set%d/std%d/iteration_%d.png', getPreDir(), dataset, addStd, iter);
            saveImage(fnameImage, 800, 400);
            fnameSummaries = sprintf('%s/html/images/SummaryPruning/set%d/std%d/iteration_%d.mat', getPreDir(), dataset, addStd, iter);
            save(fnameSummaries, 'removedSummaries', 'greedyMinScore', 'greedyMinSelected', 'greedyMaxSelected', 'greedyMaxScore', 'randScores', 'scores', 'randomSummaries');
            
            iter = iter + 1;

            fprintf(1, scoresStr);    

            if isempty(toRemove),
                fprintf(1, 'We are converged.\n');
                break;
            end
        end
    end
    
%     %% compute threshhold map
%     perm = randperm(100);
%     for l = 1:10,
%         sset = perm(1:l);
%         tbow = sum(bow(:,sset),2);
%         map = repmat(tbow(:)', size(summariesAsBOW,1), 1) > summariesAsBOW;
%         imagesc(map)
%         pause(3)
%     end
%     
%     %% compute soft threshhold map
%     perm = randperm(100);
%     for l = 1:10,
%         sset = perm(1:l);
%         tbow = sum(bow(:,sset),2);
%         map = min(repmat(tbow(:)', size(summariesAsBOW,1), 1), summariesAsBOW) ./ summariesAsBOW;
%         imagesc(map)
%         colormap(hot);
%         colorbar;
%         pause(3)
%     end
%     
%     %% compute accumulating threshhold map
%     map = repmat(tbow(:)', size(summariesAsBOW,1), 1) > summariesAsBOW;
%     omap = zeros(size(map));
%     for l = 1:200,
%         sset = randomSummary(100, 10);
%         tbow = sum(bow(:,sset),2);
%         map = repmat(tbow(:)', size(summariesAsBOW,1), 1) > summariesAsBOW;
%         omap = omap | map;
%         imagesc(omap)
%         pause(1)
%     end
    
    %% saturation of one summary against all others
%     clf
%     for i = 1:size(summariesAsBOW,1),
%         tbow = summariesAsBOW(i,:);
%         map = min(repmat(tbow(:)', size(summariesAsBOW,1), 1), summariesAsBOW) ./ summariesAsBOW;
%         imagesc(map);
%         title(sprintf('Saturation Map of Summary %d', i));
%         colormap(hot);
%         colorbar;
%         
%         % save
%         fname = sprintf('html/images/SaturationMaps/set%d/summary%d.png', dataset, i);
%         saveImage(fname, 1000, 400);
%     end
end
