%% Learns the objective as discussed in the meeting on October, 1st

% given a set of features (color histograms, bag of features), learn 
% weights that assign importance to each of the words automatically. We
% do this by solving:
%
%   max_w sum_{k=1}^K f_w(A_k),
%
% where 
%
%   f_w(A) := sum_{u \in U} w_u \sqrt{ \sum_{a \in A} c_u(a) },
%
% U is the set of available words, D = {A_1, ..., A_K} is the set of 
% human summaries, c_u(a) counts the occurences of word u in image a (maybe
% using some kind of TF-IDF scaling), and w is the weight vector to be 
% tuned (non-negative).

clc;
clear all;

%% settings
dataset = 2;

% encoder settings
% encoder = 'nnSoft';
encoder = 'nnSIFT';
encoderThreshold = 0.80;
normalizeBOW = 'l1';

% clustering settings
clustering = 'random';
nrVisualWordsRange = 10:10:1000;

% other settings
useTFIDF = false;

lambda = 10; % regularizer
nrAverage = 1000;
nrImages = 100;
summaryLength = 10;


%% load dataset
[imagescolor, images] = get_images(dataset, false);

%% load summaries and compute scores
summaryfile = sprintf('../Data/set%d/summary.txt', dataset);
summary = load(summaryfile);
nrSummaries = size(summary, 1);

pairs = nchoosek(1:100,2);
for i = 1:size(pairs,1),
    pair = pairs(i,:);
    
    %% compute sift descr
    % pair =[1,76]
    options = {'Sizes', [4,6,8,10], 'Step', 4};
    [f1,s1] = vl_phow(single(images(:,:,pair(1))), options{:});
    [f2,s2] = vl_phow(single(images(:,:,pair(2))), options{:});
    
    % match
    [m,sc] = vl_ubcmatch(s1,s2,1.5);
    if size(m,2) == 0,
        fprintf(1, 'No matches.');
        pair
        continue;
    end
    
    % consider tentative matches, and compute color histograms around
    % these. Reject matches below a certain threshold
    rejected = zeros(size(m,2),1);
    range = 4;
    [rr,cc] = meshgrid(1:100);
    bins = linspace(0,1,16);
    for l = 1:size(m,2),
        src = f1(:,m(1,l));
        dest = f2(:,m(2,l));        
        x1 = src(1);
        y1 = src(2);
        x2 = dest(1);
        y2 = dest(2);
        
        mask1 = sqrt((rr-x1).^2 + (cc-y1).^2) <= range;
        mask2 = sqrt((rr-x2).^2 + (cc-y2).^2) <= range;
        
        t1 = rgb2hsv(imagescolor(:,:,:,pair(1)));
        imgr = t1(:,:,1);
        imgg = t1(:,:,2);
        imgb = t1(:,:,3);
        slice1 = imgr(logical(mask1));
        slice2 = imgg(logical(mask1));
        slice3 = imgb(logical(mask1));
        slice_img1 = [slice1, slice2, slice3];

        t2 = rgb2hsv(imagescolor(:,:,:,pair(2)));
        imgr = t2(:,:,1);
        imgg = t2(:,:,2);
        imgb = t2(:,:,3);
        slice1 = imgr(logical(mask2));
        slice2 = imgg(logical(mask2));
        slice3 = imgb(logical(mask2));
        slice_img2 = [slice1, slice2, slice3];

        h1 = [histc(slice_img1(:,1), bins), histc(slice_img1(:,2), bins), histc(slice_img1(:,3), bins)] / (sum(sum(mask1))*3);
        h2 = [histc(slice_img2(:,1), bins), histc(slice_img2(:,2), bins), histc(slice_img2(:,3), bins)] / (sum(sum(mask2))*3);
        
        % TODO: Replace by EMD
        if sum(sum(min(h1,h2))) < 0.5,
            disp('Rejection based on color histogram.');
            rejected(l) = 1;
        end
    end
    
    % reject almost single patches
    
    % reject matches that are not from central focus to out of focus
    for l = 1:size(m,2),
        src = f1(:,m(1,l));
        dest = f2(:,m(2,l));        
        if norm(src(1:2) - dest(1:2)) > 100/2, % half the image
            rejected(l) = 1;
            disp('Rejection based on distance of points');
        end
    end
    
    
    % display
    jimg = horzcat(imagescolor(:,:,:,pair(1)), imagescolor(:,:,:,pair(2)));
    clf
    image(jimg);
    for l = 1:size(m,2),
        src = f1(:,m(1,l));
        dest = f2(:,m(2,l));
        h = line([src(1) 100+dest(1)], [src(2), dest(2)]);
        if rejected(l),
            set(h, 'color', 'r');
        else
            set(h, 'color', 'g');
        end
        %
        title('red: rejected matches; green: accepted matches');
    end
    
    %%
    pause
end

