function [] = outputFunctionValues(configFile, configComponentsFile)

% run the config
run(configFile);

%% setup
addpath('Java/MatlabInterface');
addpath('SubmodularComponents');

% OCTAVE: 
addpath(fullfile(pwd, 'external/sfo'));
if vl_isoctave(),
    page_output_immediately(1);
    page_screen_output(0);
    pkg load java
end

javaaddpath('Java/SubmodularFunctions/bin');

rand('seed', 123456');
nrDatasets = 14;
  
p = mfilename('fullpath');
fprintf(1, '=== I am %s ===\n', p);  

%% >> check if all required parameters are set >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
if ~exist('config', 'var'),
   error('No config set.');
end

check_config(config);
%% << check parameters <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


%% >> loading data >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
fprintf(1, 'Loading data...\n');
summaries = cell(nrDatasets, 1);
trainingData = cell(nrDatasets, 1);
testData = cell(nrDatasets, 1);
nrTrain = zeros(nrDatasets, 1);
nrTest = zeros(nrDatasets, 1);
for ds = 1:nrDatasets,
    % [imagesrgb, imagesgray, imageshsv, imageslab] = get_images(dataset, false);
    summaries{ds} = pruneSummaries(ds, get_pruning_info(ds, config.overlap));
    
    % shuffle summaries and extract training and test set
    nrSamples = size(summaries{ds}, 1);
    %nrTrain(ds) = ceil(nrSamples*perTrain);
    %nrTest(ds) = nrSamples - nrTrain(ds);
    nrTrain(ds) = nrSamples;
    nrTest(ds) = nrSamples;

    % summaries{ds} = summaries{ds}(randperm(nrSamples), :);
    % sort summaries by score
    bowLoss = get_BOW_evaluation(ds);
    bowSummaryLoss = computeBOWSummary(summaries{ds}, bowLoss);
    optionsEvaluationLoss = struct();
    optionsEvaluationLoss.bow = bowLoss;
    
    if config.sortSummariesByScore,
        optionsEvaluationLoss.summariesAsBow = computeBOWSummary(summaries{ds}, optionsEvaluationLoss.bow);
        [~,scores] = summaryEvaluation(summaries{ds}, summaries{ds}, 'rouge', optionsEvaluationLoss);
        [a,ordering] = sort(scores, 1, 'descend');
    else
        % randomly shuffle summaries
        ordering = randperm(size(summaries{ds},1));
    end
    summaries{ds} = summaries{ds}(ordering,:);    

    trainingData{ds} = summaries{ds};
    testData{ds} = summaries{ds};
    %trainingData{ds} = summaries{ds}(1:nrTrain(ds), :);
    %testData{ds} = summaries{ds}((nrTrain(ds)+1):end, :);
end
randomSummaries = randomSummary(100, 10, 1000);

fprintf(1, 'There are %d samples for training and %d samples for testing.\n', sum(nrTrain), sum(nrTest));
%% << Loading data <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

%% setup loss functions
tic;
fprintf(1, 'Setting up loss functions... ');
losses = cell(nrDatasets,1);
lossesJava = cell(nrDatasets, 1);
for ds = 1:nrDatasets,
    switch lower(config.lossFunction)
        case 'oneminusrouge' % OneMinusRouge
            bowLoss = get_BOW_evaluation(ds);
            bowSummaryLoss = computeBOWSummary(trainingData{ds}, bowLoss);
            
            optionsEvaluationLoss = struct();
            optionsEvaluationLoss.bow = bowLoss;
            optionsEvaluationLoss.summariesAsBow = bowSummaryLoss;

            trData = trainingData{ds};
            loss = @(x) 1 - summaryEvaluation(x, trData, 'rouge', optionsEvaluationLoss);
            losses{ds} = loss;
            
            % java
            lossesJava{ds} = OneMinusRougeJ(MatrixJ(trData), MatrixJ(bowLoss), MatrixJ(bowSummaryLoss'));
        case 'surrogaterougeloss' % SurrogateRougeLoss
            bowLoss = get_BOW_evaluation(ds);
            trData = trainingData{ds};
            bowSummaryLoss = computeBOWSummary(trData, bowLoss);
            
            loss = @(x) rougeLoss(x, bowLoss, bowSummaryLoss, ds);
            losses{ds} = loss;
        otherwise
          error('Invalid loss function specified.');
    end
    clear('trData');
end
clear('loss');
ttt = toc;
fprintf(1, ' DONE (took %f seconds)\n', ttt);

%% >> Similarities >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
% compute similarities
fprintf(1, 'Loading similarities of bag-of-words...\n');
sims = {};
simsDesc = {};
simsIdx = 1;
sim = []

tic;
fprintf(1, 'Loading further similarities... ');
% load similarities computed from overfeat
for ds = 1:nrDatasets,
    fname = sprintf('%s/features/overfeat_sim_ds_%d.mat', getPreDir(), ds);
    load(fname);
    sims{ds,simsIdx} = MatrixJ(sim);
    simsDesc{ds,simsIdx} = sprintf('overft');
end
ttt = toc;
fprintf(1, ' DONE (took %f seconds)\n', ttt);
simsIdx = simsIdx + 1;

%% << Similiarties <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

%% SETUP SHELLS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fcts = cell(nrDatasets, 1);
fctsJava = cell(nrDatasets,1);
desc = cell(1, 10);
nrSims = size(sims, 2);
idx = 1;

if ~exist(configComponentsFile, 'file'),
    error('Configuration file for components not found.');
end

components = readComponentsFile(configComponentsFile);
hash = DataHash(components);
cacheFile = strcat('cache/', hash, '.mat');
cacheFile

fprintf(1, 'Setting up components...\n');
cacheIndex = cell(nrDatasets,0);
cache = {};
% load cache
if exist(cacheFile, 'file') 
    fprintf(1, 'Loading cache file: %s\n', cacheFile);
    load(cacheFile);
end

idx = 1;
for compIdx = 1:length(components)
if isfield(components(compIdx), 'resource')
   % TODO: Make better description
   desc{compIdx} = sprintf('%s: %s', components(compIdx).type, components(compIdx).resources(1).type);
else
   desc{compIdx} = components(compIdx).type;
end
for ds = 1:nrDatasets,
    fprintf(1, '[Component %3d] %s\n', compIdx, components(compIdx).type);
    
    % command to construct component
    createStr = sprintf('comp = %sJ(', components(compIdx).type);
    
    % load resources
    nrResources = length(components(compIdx).resources);
    for resIdx = 1:nrResources,
    	if length(fieldnames(components(compIdx).resources)) == 0,
    	    break;
    	end
    	resource = components(compIdx).resources(resIdx);
    	res_desc_full = strcat(resource.type, ':', resource.url);
    	cacheHit = cellfun(@(x) strcmpi(x, res_desc_full), cacheIndex(ds,:));
    	
    	if any(cacheHit)
    	   % fprintf(1, '    Found "%s" in cache (index %d)\n', res_desc_full, find(cacheHit));
    	   chidx = find(cacheHit);
    	else
    	   fprintf(1, '    Loading resource "%s"\n', res_desc_full);
    	   tmp = load_resource(resource.type, resource.url, ds);
    	   chidx = length(cacheIndex)+1;
    	   cacheIndex{ds,chidx} = res_desc_full;
    	   
    	   if strcmpi(resource.type, 'value') || strcmpi(resource.type, 'QualityMeasure') || strcmpi(resource.type, 'Cluster')
    	   	cache{ds,chidx} = tmp;
    	   else
    	   	cache{ds,chidx} = MatrixJ(tmp);
    	   end
    	end
    	
    	createStr = strcat(createStr, sprintf('cache{%d,%d}', ds, chidx));
    	if resIdx < nrResources
    	    createStr = strcat(createStr, ',');
    	end
    end
    createStr = strcat(createStr, ');');
    
    % crate
    % fprintf(1, '  => Creating component by "%s".\n', createStr);
    try
        eval(createStr);
    catch 
        fprintf(1, 'Creation of Java object failed.\n');
        keyboard
    end
    fctsJava{ds,idx} = comp;
end
idx = idx + 1;
end



%% >> setup evaluation functions and precompute scores >>>>>>>>>>>>>>>>>>>>>>>>>
tic;
evalFcts = cell(nrDatasets,1);
fprintf(1, 'Setting up evaluation functions... ');
for ds = 1:nrDatasets,
    bowLoss = get_BOW_evaluation(ds);
    bowSummaryLoss = computeBOWSummary(testData{ds}, bowLoss);
    
    optionsEvaluationLossTest = struct();
    optionsEvaluationLossTest.bow = bowLoss;
    optionsEvaluationLossTest.summariesAsBow = computeBOWSummary(testData{ds}, bowLoss);
    
    evalFct = @(x) summaryEvaluation(x, testData{ds}, 'rouge', optionsEvaluationLossTest);
    evalFcts{ds} = evalFct;
end
clear('optionsEvaluationLossTest');
ttt = toc;
fprintf(1, 'DONE (took %.2f seconds)\n', ttt);

%% precompute average scores for displaying purposes
fprintf(1, 'Precomputing human/random scores on test data... ');
resFileName = fullfile(getPreDir(), '/tmp/evalFcts.mat');
if exist(resFileName, 'file'),
    load(resFileName);
    fprintf(1, 'S');
else
    preScores = zeros(nrDatasets,2);
    for ds = 1:nrDatasets,
        fprintf(1, '*');
        [preScores(ds,1), trScores] = evalFcts{ds}(trainingData{ds});
        preScores(ds,2) = evalFcts{ds}(randomSummaries);
        
        % min/max
	rougeMax = sfo_greedy_lazy(sfo_fn_wrapper(evalFcts{ds}), 1:100, 10);
	preScores(ds,4) = evalFcts{ds}(rougeMax);

	[~,~,rougeMin,~] = mmin(evalFcts{ds}, 100, [], 'MUB', 10);
	preScores(ds,3) = evalFcts{ds}(rougeMin);
    
        % min/max on training
        preScores(ds,5) = min(trScores);
        preScores(ds,6) = max(trScores);sumScores
		
        preScores(ds,7) = mean(trScores);
        preScores(ds,8) = std(trScores);
    end
    
    save(resFileName, 'preScores');
end
fprintf(1, ' DONE\n');
%% << setup evaluation functions and precompute scores <<<<<<<<<<<<<<<<<<<<<<<<<
testSet = 1:10;
for ds = 1:1, 
    for fctIdx = 1:size(fctsJava, 2),
        val = fctsJava{ds,fctIdx}.eval(testSet);
        minval = fctsJava{ds,fctIdx}.eval([]);
        maxval = fctsJava{ds,fctIdx}.eval(1:100);
        fprintf(1, '[%2d,%3d] val=%f [min=%f, max=%f] %s\n', ds, fctIdx, val, minval, maxval, desc{fctIdx});
    end
end

disp('rouge');
for ds = 1:14,
    fprintf(1, 'Dataset %d\n', ds);
    lossesJava{ds}.eval(testSet)
    % 1-lossesJava{ds}.eval([])
    % 1-lossesJava{ds}.eval(1:100)
end
preScores(:,1)

%% >> Sanity check: are all components normalized >>>>>>>>>>>>>>>>>>>>>>>>>
fprintf(1, 'Checking if all components are properly normalized...\n');
for ds = 1:14,
    for fctIdx = 1:size(fctsJava, 2),
        valL = fctsJava{ds,fctIdx}.eval([]);
        valU = fctsJava{ds,fctIdx}.eval(1:100);
        if valL < 0 || valU > 1 || valL > 1 || valU < 0,
            error('Functions not properly normalized.');
        end
    end
end

fprintf(1, 'Checking if all losses are properly normalized...\n');
for ds = 1:14,
        valL = lossesJava{ds}.eval([]);
        valU = lossesJava{ds}.eval(1:100);
        if valL < 0 || valU > 1 || valL > 1 || valU < 0,
            error('Functions nor properly normalized.');
        end
end
%% << Sanity check: are all components normalized <<<<<<<<<<<<<<<<<<<<<<<<<


end

function [components] = readComponentsFile(configComponentsFile)

components = struct();
fid = fopen(configComponentsFile, 'r');
tline = fgetl(fid);
compIdx = 1;
while ischar(tline)
    if all(isspace(tline)), % skip empty lines
       tline = fgetl(fid);
       continue;
    end

    tline = strtrim(tline);
    if length(tline) < 10 || ~strcmpi('Component:', tline(1:10)),
        error('Invalid configuration: "%s"\n', tline);
    end
    
    tline = strrep(tline, 'Component:', '');
    tline = strtrim(tline);
    
    components(compIdx).type = tline;
    
    % Read in resources (if any)
    resIdx = 1;
    components(compIdx).resources = struct();
    tline = fgetl(fid);
    tline = strtrim(tline);
    while ischar(tline) && length(tline) >= 9 && strcmpi('Resource:', tline(1:9)),
        tline = strtrim(tline);
    	tline = strrep(tline, 'Resource:', '');
        tline = strtrim(tline);
        
        [tok,remainder] = strtok(tline, ':');
        res_type = tok;
        res_url = strtrim(remainder(2:end)); % skip ':'
        components(compIdx).resources(resIdx).type = res_type;
        components(compIdx).resources(resIdx).url = res_url;
        
        resIdx = resIdx + 1;
        
        tline = strtrim(fgetl(fid));
    end
    compIdx = compIdx + 1;
end
fclose(fid);

end

function [] = check_config(config)

if ~isfield(config, 'testDatasets') || isempty(config.testDatasets),
    error('You must set the dataset using the variable ''testDatasets''.');
end
if ~isfield(config, 'resultsFilename') || isempty(config.resultsFilename),
    error('You must set the filename for storing the results using ''resultsFilename''.');
end
if ~isfield(config, 'overlap'),
    error('Desired overlap must be specified.');
end
if ~isfield(config, 'lossFunction'),
    error('Loss function not set. Use ''lossFunction = xyz;'' to set it.\n');
end
if ~isfield(config, 'optimizer'),
    error('Optimizer options not set.');
end
if ~isfield(config.optimizer, 'inference'),
    error('LAI inference not set.');
end
if ~isfield(config, 'trainDatasets'),
    error('Training datasets not set.');
end

fprintf(1, 'Config seems ok.\n');

end
