function X = sfo_modmodimproveduc(F, G, V, mode, Xinit)
% The modular-modular procedure for maximizing a difference between
% submodular functions. It performs a modular maximization at every step 
% for maximizing mF - mG, where mG is the modular upper bound of G and mF is the modular upper bound of F.
% This is an improved version of the code which includes a hueristic in
% choosing the right permutation.

% mode = 1 : Random subgradient
% mode = 2: Efficient subgradient (default)
if nargin == 3
    mode = 2;
    Xinit = [];
end
if nargin == 4
    Xinit = [];
end
X = Xinit;
TOL = 1e-6;
bestVal = -inf;
% pi = V(randperm(length(V)));
[vecVminusj vecj] = find_vectors_fast(G, V);
H = sfo_fn_lincomb({F, G}, [1, -1]);

while(1)  
    Ga = sfo_fn_wrapper(@(A) mod_uppera(G, A, X, V, vecj) - mod_uppera(G, [], X, V, vecj));
    Gb = sfo_fn_wrapper(@(A) mod_upperb(G, A, X, V, vecVminusj) - mod_upperb(G, [], X, V, vecVminusj));
    if mode == 1
        pi = get_rand_perm(V, X);
    else
        pi = get_right_perm(H, V, X);
    end
    Fw = sfo_ssp_modular_approx(F,pi);
    Fl = sfo_fn_wrapper(@(A) sum(Fw(sfo_unique_fast(A))));
    FMa = sfo_fn_lincomb({Fl,Ga},[1,-1]);
    FMb = sfo_fn_lincomb({Fl,Gb},[1,-1]);
    Aa = sfo_greedy_lazy(FMa, V, inf);
    Ab = sfo_greedy_lazy(FMb, V, inf);
    if (F(Aa) - G(Aa) >= F(Ab) - G(Ab))
        X = Aa;
    else 
        X = Ab;
    end
    currVal = F(X) - G(X);
    disp(sprintf('The value at the current iteration is: %f', currVal));
    if currVal > bestVal + TOL
        bestVal = currVal;
    else
        break;
    end
end

end

function pi = get_right_perm(H, V, X)
    % The function to get the permutations based on a hueristic 
    % of arranging the elements based on the current gains.
    D = sfo_setdiff_fast(V,X);
    diff = [];
    diffcord = [];
    for j = 1 : length(X)
        diff(j) = H(setdiff(X, X(j))) - H(X);
        diffcord(j) = X(j);
    end
    %[value k] = max(diff);
    %Xk = setdiff(X, X(k));
    addel = [];
    addelcord = [];
    for j = 1 : length(D)
        addel(j) = H([X, D(j)]) - H(X);
        addelcord(j) = D(j);
    end
    [~, diffsortorder] = sort(diff, 'ascend');
    [~, addelsortorder] = sort(addel, 'descend');
    pi = [diffcord(diffsortorder), addelcord(addelsortorder)];
end

function pi = get_rand_perm(V, X)
    D = sfo_setdiff_fast(V,X);
    D = D(randperm(length(D)));
    pi = [X D];
end

function H = sfo_ssp_modular_approx(G,pi)
H = zeros(1,max(pi(:)));
W = [];
oldVal = G(W);
for i = 1:length(pi)
    newVal = G([W pi(i)]);
    H(pi(i)) = newVal-oldVal;
    oldVal = newVal;
    W = [W pi(i)];
end
end

function [val] = mod_uppera(F, A, B, V, vecj)
% Produces the modular upper bound-1 of a submodular function F, with
% respect to a set B. 

n = length(V);
val = F(B);
AminusB = sfo_setdiff_fast(A, B);
BminusA = sfo_setdiff_fast(B, A);
val = sum(vecj(AminusB));
currset = F(B);
for i = 1 : length(BminusA)
    j = BminusA(i);
    Bminusj = setdiff(B, j);
    val = val + F(Bminusj) - currset;
end
end

function [val] = mod_upperb(F, A, B, V, vecVminusj)
% Produces the modular upper bound-1I of a submodular function F, with
% respect to a set B. 

n = length(V);
val = 0;
AminusB = sfo_setdiff_fast(A, B);
BminusA = sfo_setdiff_fast(B, A);
currset = F(B);
val = -sum(vecVminusj(BminusA));
for i = 1 : length(AminusB)
    j = AminusB(i);
    val = val + F([B, j]) - currset;
end
end

function [vecVminusj vecj] = find_vectors_fast(F, V)
n = length(V);
    for j = 1 : n
        Vminusj = sfo_setdiff_fast(V, j);
        vecVminusj(j) = F(V) - F(Vminusj);
        vecj(j) = F(j) - F([]);
    end
end

