function [ score, scores ] = summaryEvaluation(testSummary, summaries, evaluationType, options)
% summaryEvaluation  Evaluates a summary.
%   This function evaluates the summary "testSummaries" against the
%   summaries "summaries". The function is called as
%
%     [ score ] = summaryEvaluation(testSummary, summaries, evaluationType, options)
%
%   where evaluationType can be either
%
%     * combinatorial (combinatorial n-gram),
%     * rouge (classical rouge score),
%     * weightedRouge (weighted rouge score).
% 
%   When "combinatorial" is selected as evaluationType, the value options.k
%   corresponds to "order" of n-grams considered. When "rouge" or
%   "weightedRouge" is selected, the field options.bow must correspond to 
%   bag-of-words representations of the summaries. Further, in the case of
%   "weightedRouge", the field options.w is the weight-vector. For
%   efficiency, the precomputed words in a summary can be supplied by
%   setting the field "summariesAsBow" in the "option" structure.

%% Options
nOptArg = 4;
if nargin < nOptArg || isempty(options),
    error('Options must be specified.');
end

if ~strcmp(evaluationType, 'combinatorial') && ~strcmp(evaluationType, 'rouge') && ~strcmp(evaluationType, 'weightedRouge')  && ~strcmp(evaluationType, 'largestSubset'),
    error('Invalid evaluation type specified.');
end

if strcmp(evaluationType, 'combinatorial') && ~isfield(options, 'k'),
    error('The field "options.k" must be set.');
end

if (strcmp(evaluationType, 'rouge') || strcmp(evaluationType, 'weightedRouge'))  && ~isfield(options, 'bow'),
    error('The field "options.bow" must be set.');
end

if (strcmp(evaluationType, 'rouge') || strcmp(evaluationType, 'weightedRouge'))  && ~isfield(options, 'summariesAsBow'),
    warning('For efficiency reasons, the bag-of-words representation of the summaries should be provided.');
end

if strcmp(evaluationType, 'weightedRouge')  && ~isfield(options, 'w'),
    error('The field "options.w" must be set.');
end

%% set often used quantities
nrSummaries = size(summaries, 1);

if (strcmp(evaluationType, 'rouge') || strcmp(evaluationType, 'weightedRouge')),
    % precomupte bag-of-words for summaries
    nrWords = size(options.bow, 1);
    if ~isfield(options, 'summariesAsBow'),
        summariesAsBow = NaN(nrSummaries, nrWords);
        for s = 1:nrSummaries,
            summariesAsBow(s,:) = sum(options.bow(:,summaries(s,:)), 2);
        end
    else
        summariesAsBow = options.summariesAsBow;
    end
    
    testSummaryAsBow = sum(options.bow(:,testSummary), 2);
end

%% sanity check
if (strcmp(evaluationType, 'rouge') || strcmp(evaluationType, 'weightedRouge')),
    if size(summaries, 1) ~= size(summariesAsBow, 1),
        error('Size of summaries and summariesAsBow must match.');
    end
end

if strcmp(evaluationType, 'weightedRouge'),
    if ~isfield(options, 'w') || size(summariesAsBow,2) ~= size(options.w,2) || ~isvector(options.w),
        error('Invalid weight vector, or weight vector not specified.');
    end
end

%% if testSummary contains multiple summaries, score them all
if size(testSummary, 1) > 1,
    nrTestSummaries = size(testSummary, 1);
    scores = NaN(nrTestSummaries, 1);
    for k = 1:nrTestSummaries,
        scores(k) = summaryEvaluation(testSummary(k,:), summaries, evaluationType, options);
    end
    score = mean(scores);
    return
elseif size(testSummary, 1) == 0,
    score = 0;
    scores = [];
    return;
end

%% compute score
switch evaluationType
    case 'combinatorial'
        % sort summaries for faster evaluation
        summaries = sort(summaries, 2);
        num = 0;
        den = 0;
        for k = 1:min(options.k, length(testSummary)), % TODO: Check if this is meaningful
            comb = nchoosek(testSummary, k);
            comb = sort(comb, 2); % seems to be sorted anyway, but the documentation does not state that this is guaranteed
            for j = 1:nrSummaries,
                currsummary = summaries(j, :);
                tmp = ismembc(comb, currsummary);
                num = num + sum(all(tmp,2));
            end
            den = den + nrSummaries*nchoosek(length(testSummary), k);
        end
    case 'largestSubset'
        summaries = sort(summaries, 2);
        den = 10 * size(summaries,1);
        num = sum(sum(ismembc(summaries, sort(testSummary))));
    case 'rouge'
        num = 0; 
%         for fi = 1:nrWords, % cycle through all visual words
%             for j = 1:nrSummaries, % cycle thorugh all summaries
%                 num = num + min(testSummaryAsBow(fi), summariesAsBow(j, fi));
%             end
%         end
        a = min(repmat(testSummaryAsBow(:)', nrSummaries, 1), summariesAsBow);
        num = sum(a(:));
        den = sum(summariesAsBow(:));
    case 'weightedRouge'
        den = 0;
        a = min(repmat(testSummaryAsBow(:)', nrSummaries, 1), summariesAsBow);
        num = sum(sum(a,1) .* options.w);
        den = den + sum(sum(summariesAsBow,1) .* options.w);
    otherwise
        error('Evaluation type not supported (yet?).')
end    
score = num/den;
scores = score;
    
end